
(* what is the purpose/are the purposes of AI?
   => how can this reflect in a simpler implementation (assertions we can take? )
*)


module AbstractInterpretation


open System
open System.Collections.Generic

open Ast
open X64instr

open InstrIO
open MoreInstrIO

open AstIO
open MoreAstIO


open NXyn
open NXyn.Aux
open NXyn.Aux.Global

open NXyn.Aux.IntMisc
open NXyn.Aux.OtherMisc
open NXyn.Aux.SetMisc


open Basics

open LocalScheduling

open OptSimple


(** AI starts *)




type SSArreg = R64 * int

let WriteSSArreg_to ((reg:R64),n) (w:IO.TextWriter) = fprintf w "%s[%d]" R64name.[reg.Int]  n

type LinFactor = Var   of SSArreg
               | Konst of String
               | ParamQ of int (* function parameter #<x>, 1-based   1=rdi, 2=rsi, 3=rdx, 4=rcx, 5=r8, 6=r9, 7=rsp[8], 8=rsp[16], 9=rsp[24], ... *)


               
type LinEx = Int64 * Map<LinFactor,Int64>

let make_LinEx(x:LinFactor) = (0L, Map[x,1L])


let (|LinEx_Int64|_|) ((k,m):LinEx) = if m.IsEmpty then Some(k) else None

let WriteLinFactor_to (x:LinFactor) (w:IO.TextWriter) = match x with Var(r) -> WriteSSArreg_to r w
                                                                   | Konst(s) -> w.Write(s)
                                                                   | ParamQ(i) -> fprintf w "param(%d)" i
                                                                   
let WriteLinEx_to ((k,map):LinEx) w = fprintf w "linex([%t])" (fun w -> if k <> 0L then fprintf w "%d; " k
                                                                        for KeyValue(key,n) in map do fprintf w "%d * %t; " n (WriteLinFactor_to key))

type iregs = array<int * Set<LinEx>> // use map? (something that allows for sparcity)

let R_occursIn_LinEx (r:R64) ((_,m):LinEx) = m |> Map.exists (fun v n -> match v with Var(t,_) -> t = r | _ -> false)

let eaOption_of_linex (k,map) = match Map.toList(map) with [(Var(r1,_),1L)] -> Some(M.B_D(r1,k))
                                                         | [(Var(r1,_),1L); (Var(r2,_),n2)] -> Some(M.B_SI_D(r1,r2,n2,k))
                                                         | [(Var(r1,_),n1); (Var(r2,_),1L)] -> Some(M.B_SI_D(r2,r1,n1,k))
                                                         | _ -> None
            
let ea_variations' (iregfile:iregs) ea =
    let addFactorTimesN scale factor n dict = dict |> Map.add factor (n*scale + (defaultArg (Map.tryFind factor dict) 0L))

    match ea with
        | EA64(IexprInt64(k), None, None) -> [|(k,Map.empty)|]

        | EA64(IexprInt64(k), Some({Base=Q(b)}), Some({Index=Q(i); Scale=iScale})) when i >= 42 || b >= 42 -> [||]

        | EA64(IexprInt64(k), Some({Base=Q(b)}), None) ->
            let (b_idx0, b_content0) = iregfile.[b]

            let b_ref = (k,Map [Var(Q(b),b_idx0),1L])

            let combine ((k1,x):LinEx) = (k + k1,
                                          Map.empty
                                          |> Map.foldBack (addFactorTimesN 1L) x
                                          |> Map.filter (fun _ n -> n <> 0L))
            let set = new HashSet<LinEx>()
            b_ref |> set.Add |> ignore
            b_content0 |> Set.iter (fun b' -> combine b' |> set.Add |> ignore)
            set |> ArrayMisc.array_of_collection
                
        | EA64(IexprInt64(k), None, Some({Index=Q(i); Scale=iScale})) ->
            let (i_idx0, i_content0) = iregfile.[i]

            let i_ref = (k,Map [Var(Q(i),i_idx0), iScale])

            let combine ((k2,y):LinEx) = (k + k2*iScale,
                                          Map.empty
                                          |> Map.foldBack (addFactorTimesN iScale) y
                                          |> Map.filter (fun _ n -> n <> 0L))

            let set = new HashSet<LinEx>()
            i_ref |> set.Add |> ignore
            i_content0 |> Set.iter (fun i' -> combine i' |> set.Add |> ignore)
            set |> ArrayMisc.array_of_collection
            
        | EA64(IexprInt64(k), Some({Base=Q(b)}), Some({Index=Q(i); Scale=iScale})) ->
            let combine ((k1,x):LinEx) ((k2,y):LinEx) = (k + k1 + k2*iScale,
                                                         Map.empty
                                                         |> Map.foldBack (addFactorTimesN 1L) x
                                                         |> Map.foldBack (addFactorTimesN iScale) y
                                                         |> Map.filter (fun _ n -> n <> 0L))
            (* b_idx0, b_content0  => bIX, bContent *)
            let (b_idx0, b_content0) = iregfile.[b]
            let (i_idx0, i_content0) = iregfile.[i]

            let b_ref = (0L,Map [Var(Q(b),b_idx0), 1L])
            let i_ref = (0L,Map [Var(Q(i),i_idx0), 1L])

            let set = new HashSet<LinEx>()
            combine b_ref i_ref |> set.Add |> ignore
            b_content0 |> Set.iter (fun b' -> combine b' i_ref |> set.Add |> ignore)
            i_content0 |> Set.iter (fun i' -> combine b_ref i' |> set.Add |> ignore
                                              b_content0 |> Set.iter (fun b' -> combine b' i' |> set.Add |> ignore))
            set |> ArrayMisc.array_of_collection

        | _ -> [||]

let ea_variations (iregfile:iregs) ea =
    let ea = match ea with EA64(k, Some({Base=Q(b)}), Some({Index=Q(i); Scale=s})) when b = i -> EA64(k,None,Some({Index=Q(i);Scale=s+1L}))
                         | EA64(k, Some({Base=Q(b)}), Some({Index=Q(i); Scale=0L}))           -> EA64(k,Some({Base=Q(b)}),None)
                         | _ -> ea

    let variations = ea_variations' iregfile ea


    if variations = [||] then Set.empty // variations
    else if variations.Length > 1000 then failwithf "variations: too big %s" !Global.InFileName
    else Set.ofArray(variations)

(*
    let lens = Array.map (fun(_,m:Map<_,_>) -> m |> Map.toArray |> Array.map fst |> Array.filter (function Var(_,n) -> n = 1 | _ -> false) |> Array.length) variations
         let mn = Array.min lens
         let mx = Array.max lens
         if !dodump then eprintf "mn = %d, mx = %d: " mn mx

         if mn <> mx then
             let vsN = Array.filter (fun(_,m:Map<_,_>) -> mx = (m |> Map.toArray |> Array.map fst |> Array.filter (function Var(x,n) -> x = R64.rsp && n = 1 | Konst _ -> true | _ -> false) |> Array.length)) variations
            
             if !dodump then eprintfn "totallen = %d => %d" variations.Length vsN.Length
             vsN
         else
             if !dodump then eprintfn "totallen = %d" variations.Length
             variations
*)             
//         let (yes,no) = variations |> Array.partition (fun(_,m:Map<_,_>) -> m.Count = mn)
//         yes

(*    
    if variations = [||] then variations
    else let lens = Array.map (fun(_,m:Map<_,_>) -> m.Count) variations
         let mn = Array.min lens
         let mx = Array.max lens
         let (yes,no) = variations |> Array.partition (fun(_,m:Map<_,_>) -> m.Count = mn)
         yes
*)



(***)

(* stack: byte-addressable memory that is private to this thread *)
type stackItem = Si of int (* part *) * int (* size *) * Set<LinEx> 

type stackT = System.Collections.Generic.Dictionary<Int32, stackItem> (* representation? set? *)

let interp__Leaq (memory:stackT) (iregfile:iregs) (s:M) (Q(d)) = let (dIX,_) = iregfile.[d]
                                                                 iregfile.[d] <- (dIX+1, ea_variations iregfile s)

let didmemclear = ref false

(* DOC stack push/pop: need to be rewritten before AI *)
let rec interp_ (memory:stackT) (iregfile:iregs) (i:Instr) =
  let (|ReaInt64|_|) (Q(s)) = iregfile.[s] (* |> List.head DEL *) |> snd |> set_tryPick (function(LinEx_Int64(k)) -> Some(Q(s),k) | _ -> None)

  let memCLEAR() = didmemclear := true; memory.Clear()

//  if !dodump then eprintfn "%A (memSIZ=%d, rfSIZ=%d)" i memory.Count (Array.sum (Array.map (fun(c) -> snd |> Set.count) iregfile))
  match i with
    (* instructions that do not alter the AI state *)
    | C(_) | I0(I_I0.Retq) -> ()
    | R8sR8s(_) | R16sR16s(_) | R32sR32s(_) | R64sR64s(_) -> ()
    | R64ds((I_R64ds.Addq(LitInt32(0)) | I_R64ds.Subq(LitInt32(0)) | I_R64ds.Xorq(LitInt32(0)) | I_R64ds.Orq(LitInt32(0))),   _) -> ()
    | R64ds((I_R64ds.Rolq(LitInt32(0)) | I_R64ds.Rorq(LitInt32(0)) | I_R64ds.Rclq(LitInt32(0)) | I_R64ds.Rcrq(LitInt32(0))),  _) -> ()
    | R64ds((I_R64ds.Shlq(LitInt32(0)) | I_R64ds.Shrq(LitInt32(0)) | I_R64ds.Sarq(LitInt32(0)) | I_R64ds.Imulq(LitInt32(1))), _) -> ()

    | I0(I_I0.Ret(LitInt64(k))) -> interp__Leaq memory iregfile (M.B_D(R64.rsp,k)) R64.rsp
                                   interp_ memory iregfile (I0(I_I0.Retq))

    // HACK, extend?
    | M64sR64ds(I_M64sR64ds.Addq, {S=s; DS=ds}) -> interp_ memory iregfile (M64sR64d(I_M64sR64d.Movq, {S=s;D=Q(16)}))
                                                   interp_ memory iregfile (R64sR64ds(I_R64sR64ds.Addq, {S=Q(16); DS=ds}))
    | M64sR64ds(I_M64sR64ds.Subq, {S=s; DS=ds}) -> interp_ memory iregfile (M64sR64d(I_M64sR64d.Movq, {S=s;D=Q(16)}))
                                                   interp_ memory iregfile (R64sR64ds(I_R64sR64ds.Subq, {S=Q(16); DS=ds}))


    (* stack push/pop *)
    | R64d(I_R64d.Popq,  {D=d}) -> interp_ memory iregfile (M64sR64d(I_M64sR64d.Movq, {S=M64.S(M.B(R64.rsp)); D=d}))
                                   interp__Leaq memory iregfile (M.B_D(R64.rsp, 8L)) R64.rsp

    | R64s(I_R64s.Pushq, {S=s}) -> interp__Leaq memory iregfile (M.B_D(R64.rsp,-8L)) R64.rsp
                                   interp_ memory iregfile (R64sM64d(I_R64sM64d.Movq, {S=s; D=M64.S(M.B(R64.rsp))}))

    (* instructions operating on 32-bit integer registers *)
    | R32d(I_R32d.Movl(LitInt64(k)), {D = L(d)}) when k >= 0L -> interp__Leaq memory iregfile (M.D(k))  (Q(d))
    | R32ds(I_R32ds.Andl(LitInt64(0L)), {DS = L(ds)})         -> interp__Leaq memory iregfile (M.D(0L)) (Q(ds))
    | R32sR32ds((I_R32sR32ds.Subl | I_R32sR32ds.Xorl), {DS=(L(d) as ds); S=s}) when ds = s -> interp__Leaq memory iregfile (M.D(0L)) (Q(d))
    
    | R64d(I_R64d.Movq(LitInt64(k)),      {D=d}) -> interp__Leaq memory iregfile (M.D(k)) d
    | R64d(I_R64d.Movabsq(LitInt64(k)),   {D=d}) -> interp__Leaq memory iregfile (M.D(k)) d
    | R64d(I_R64d.Leaq(EA64(_) as s_ea),  {D=d}) -> interp__Leaq memory iregfile s_ea     d

    | R64ds(I_R64ds.Addq(LitInt64(k)),  {DS=ds}) -> interp__Leaq memory iregfile (M.B_D(ds,k))        ds
    | R64ds(I_R64ds.Andq(LitInt64(0L)), {DS=ds}) -> interp__Leaq memory iregfile (M.D(0L))            ds 
    | R64ds(I_R64ds.Decq,               {DS=ds}) -> interp__Leaq memory iregfile (M.B_D(ds,-1L))      ds
    | R64ds(I_R64ds.Imulq(LitInt64(k)), {DS=ds}) -> interp__Leaq memory iregfile (M.SI(ds,k))         ds
    | R64ds(I_R64ds.Incq,               {DS=ds}) -> interp__Leaq memory iregfile (M.B_D(ds, 1L))      ds
    | R64ds(I_R64ds.Negq,               {DS=ds}) -> interp__Leaq memory iregfile (M.SI(ds,-1L))       ds
    | R64ds(I_R64ds.Notq,               {DS=ds}) -> interp__Leaq memory iregfile (M.SI_D(ds,-1L,-1L)) ds
    | R64ds(I_R64ds.Shlq(LitInt32(sh)), {DS=ds}) -> interp__Leaq memory iregfile (M.SI(ds,1L <<< sh)) ds
    | R64ds(I_R64ds.Subq(LitInt64(k)),  {DS=ds}) -> interp__Leaq memory iregfile (M.B_D(ds,-k))       ds
     
    | R64sR64d(I_R64sR64d.Imulq(LitInt64(k)), {S=s; D=d}) -> interp__Leaq memory iregfile (M.SI(s,k)) d
    | R64sR64d(I_R64sR64d.Movq,               {S=s; D=d}) -> interp__Leaq memory iregfile (M.B(s))    d

    | R64sR64ds(I_R64sR64ds.Addq, {DS=ds; S=s})               -> interp__Leaq memory iregfile (M.B_SI(ds,s,1L))  ds
    | R64sR64ds(I_R64sR64ds.Imulq, {S=ReaInt64 (_,k); DS=ds}) -> interp__Leaq memory iregfile (M.SI(ds,k))       ds
    | R64sR64ds(I_R64sR64ds.Imulq, {S=s; DS=ReaInt64 (ds,k)}) -> interp__Leaq memory iregfile (M.SI(s,k))        ds 
    | R64sR64ds(I_R64sR64ds.Xorq, {DS=ds; S=s}) when ds = s   -> interp__Leaq memory iregfile (M.D(0L))          ds
    | R64sR64ds(I_R64sR64ds.Subq, {DS=ds; S=s})               -> interp__Leaq memory iregfile (M.B_SI(ds,s,-1L)) ds
    
(* do similarly:
      XsMd.Movaps/Movapd | movups/movupd/movss/movsd/movlps/movhps/movlpd/movhpd 
      YsMd..../
          
      use attribute: Copy? Load? Store?
      CopyToMem? CopyFromMem?
      similarly
*)
(* Attributes: IntLoad(4), IntStore(8), VecLoad(16), VecStore(4)
     x
     StartFrom

     size?
     extents?
     Sext
     Dext
     DSext??
*)

    (* STORE *)
    | R64sM64d(I_R64sM64d.Movq,{S=Q(s); D=M64.S(m)}) ->
        (* TBD does it potentially alias?   if so start new state about which we know NOTHING *)
        let (s_idx0,s_content) = iregfile.[s]
        let m_vars = ea_variations iregfile m
        
//        match m_vars |> set_tryFind (function (k,map) -> map.Count = 1 && (match Map.tryFind (ParamQ(0)) map with | Some(x) -> x = 1L | None -> false)) with
        match m_vars |> Set.toArray |> Array.tryFind (function (k,map) -> map.Count = 1 && (match Map.tryFind (Var(R64.rsp,1)) map with | Some(x) -> x = 1L | None -> false)) with
            | Some((k,_) as m') ->
                if k < 0L then
                    let k = int k
                    (* sizeof(Q) = hard 8 *)

                    let s_content = Set.add (make_LinEx (Var(Q(s),s_idx0))) s_content

                    
                    let size = 8
                    for i = 0 to size-1 do memory.[int(k)+i] <- Si(i, size, s_content)
                    
                    if !dodump then printfn "MEM1> memory %A (%A)" memory m'
                else
                    if !dodump then printfn "MEM2> stack outside %A (%A)" memory m'
                
            | _ ->
                if m_vars |> Set.exists (R_occursIn_LinEx R64.rsp) then
                    if !dodump then printfn "MEM4 > clear"
                    memCLEAR()
                else
                    if !dodump then printfn "MEM3> memunknown %A" m

    (* LOAD *)
    | M64sR64d(I_M64sR64d.Movq, {S=M64.S(m); D=Q(d)}) ->
        let d_content = iregfile.[d]
        let (d_idx0, d_content0) = d_content
        let d_idx = d_idx0 + 1

        (* BOILER *)
        let m_vars = ea_variations iregfile m
//        match m_vars |> set_tryFind (function (k,map) -> map.Count = 1 && (match Map.tryFind (ParamQ(0)) map with Some(x) -> x = 1L | None -> false)) with
        match m_vars |> set_tryFind (function (k,map) -> map.Count = 1 && (match Map.tryFind (Var(R64.rsp,1)) map with Some(x) -> x = 1L | None -> false)) with
            | Some((k,_) as m') ->
                let k = int k
                let xs = [|0..7|] |> Array.map (fun(i) -> if !dodump then eprintfn "   memory tryget: stack[%d]" (k+i)
                                                          let (found,value) = memory.TryGetValue(k+i)
                                                          if not(found) then None
                                                          else match value with Si(part,siz,content) -> if siz = 8 && part = i then Some(content) else None
                                                                              | _ -> None)
//                eprintfn "LOAD %A" xs
                if xs.[0] <> None && xs |> Array.forall ((=) xs.[0]) then
                    if !dodump then eprintfn "  LOAD EQ"

                    let m_content = xs.[0] |> Option.get
                    iregfile.[d] <- (d_idx,m_content) // ::d_content)
                else
                    if !dodump then eprintfn "  LOAD NEQ"
                    let set = Set.singleton (make_LinEx (Var(Q(d),d_idx)))
                    iregfile.[d] <- (d_idx,set) // ::d_content) // Set.empty
            | _ ->
                if !dodump then eprintfn "LOAD unknown"
                let set = Set.singleton (make_LinEx (Var(Q(d),d_idx)))
                iregfile.[d] <- (d_idx,set) // DEL::d_content) // Set.empty

    | _ -> (* TOUCH all Rds and Rd *)
//        eprintfn "doAI: unknown %A" i    
        (* TOUCH ALL Ms_of_Instr, so that neededstores are preserved *)
        (* handle like LOAD (above) *)

        for ds in (FixedRds_of_Instr(i) @ FixedRd_of_Instr(i) @ Rds_of_Instr(i) @ Rd_of_Instr(i)) do
          let ds = (ds:R).Int
          let ds_content = iregfile.[ds]
          let (ds_idx0, ds_content0) = ds_content
          let ds_idx = ds_idx0 + 1
          
          let set = Set.singleton (make_LinEx (Var(Q(ds),ds_idx)))
          let ds_newcontent = (ds_idx, set) // Set.empty)

          
          iregfile.[ds] <- ds_newcontent //::ds_content) (* cut that? *)

        if Instr_uses_FixedMem(i) then
            eprintfn "FIXEDmem => flush"
            memCLEAR()
        else
          for (m_size,m) in (Mds_of_Instr i @ Md_of_Instr i) do  (* for M8/M16/M32/M64/M128/M256 *)
            let m_vars = ea_variations iregfile m
//            match m_vars |> set_tryFind (function (k,map) -> map.Count = 1 && (match Map.tryFind (ParamQ(0)) map with | Some(x) -> x = 1L | None -> false)) with
            match m_vars |> set_tryFind (function (k,map) -> map.Count = 1 && (match Map.tryFind (Var(R64.rsp,1)) map with | Some(x) -> x = 1L | None -> false)) with
                | Some((k,_) as m') ->
                  if k < 0L then
                    let k = int k
                    for i = 0 to m_size-1 do
                        memory.Remove(int(k)+i) |> ignore

                    if !dodump then printfn "MEM1> memory %A (%A)" memory m'
                  else
                    if !dodump then printfn "MEM2> stack outside %A (%A)" memory m'
                
                | _ ->
                  if m_vars |> Set.exists (R_occursIn_LinEx R64.rsp) then
                    if !dodump then printfn "MEM4 > clear"
                    memCLEAR()
                  else
                      if !dodump then printfn "MEM3> memunknown %A" m 

        if !dodump then printfn "unknown instr %A" i

(***)

let dump_iregfile (iregfile:iregs) = if !dodump then
                                         for i = 0 to 15 do
                                             printf "[%s]" R64name.[i]
                                             let (n,equiv) = iregfile.[i]
//                                             for  in es do
                                             printf "; %s[%d]" R64name.[i] n

                                             for x in equiv do
                                                 printf " = " 
                                                 WriteLinEx_to x stdout
                                             printf "\n"
(***)

let ai_map_instr1a (memory:stackT) (iregfile:iregs) (i:Instr) =
    match i with R64sR64d(I_R64sR64d.Movq, {S=Q(s); D=Q(d)}) -> let (_, d_equiv) = iregfile.[d] //DEL |> List.head
                                                                let (_, s_equiv) = iregfile.[s] //DEL|> List.head
                                                                                                                        
                                                                if sets_intersect s_equiv d_equiv then // |> Set.exists (fun x -> Set.exists ((=)x) s_equiv) then
                                                                    //eprintfn "NOP %A" i
                                                                    Instr.Nop
                                                                else i
               | i -> i


let ai_map_instr1b (memory:stackT) (iregfile:iregs) (i:Instr) =
    match i with R64sR64d(I_R64sR64d.Movq, {S=Q(s); D=Q(d)}) -> let (d_n, d_equiv) = iregfile.[d] //DEL |> List.head
                                                                let (s_n, s_equiv) = iregfile.[s] //DEL|> List.head
                                                                match ea_variations iregfile (M.B_SI((Q(s)), (Q(d)), -1L)) |> set_tryFind (fun(k,m) -> Map.isEmpty m) with
                                                                    | Some((k,_)) -> // eprintfn "ADDQ"
                                                                                     R64ds(I_R64ds.Addq(Lit.Int64(k)), {DS=Q(d)})
                                                                    | _ -> i
               | i -> i



let ai_map_instr1c (memory:stackT) (iregfile:iregs) (i:Instr) =
    
  let memCLEAR() = didmemclear := true; memory.Clear()

  match i with
    | M64sR64d(I_M64sR64d.Movq, {S=M64.S(s); D=Q(d)}) ->
        (* BOILER *)
        let m_vars = ea_variations iregfile s
        
        match m_vars |> set_tryFind (function (k,map) -> map.Count = 1 && (match Map.tryFind (Var(R64.rsp,1)) map with Some(x) -> x = 1L | None -> false)) with
//        match m_vars |> set_tryFind (function (k,map) -> map.Count = 1 && (match Map.tryFind (ParamQ(0)) map with Some(x) -> x = 1L | None -> false)) with
            | Some((k,_) as m') ->
                let k = int k
                let xs = [|0..7|] |> Array.map (fun(i) -> if !dodump then eprintfn "   memory tryget: stack[%d]" (k+i)
                                                          let (found,value) = memory.TryGetValue(k+i)
                                                          if not(found) then None
                                                          else match value with Si(part,siz,content) -> if siz = 8 && part = i then Some(content) else None
                                                                              | _ -> None)
//                eprintfn "LOAD %A" xs
                if xs.[0] <> None && xs |> Array.forall ((=) xs.[0]) then
                    if !dodump then eprintfn "  LOAD EQ"
                    let m_content = xs.[0] |> Option.get

                    match [|0..15|] |> Array.tryFind (fun ix -> let (_,ix_equiv) = iregfile.[ix] //DEL|> List.head
                                                                sets_intersect ix_equiv m_content) // DELix_equiv |> Set.exists (fun(ix_val) -> m_content |> Set.exists ((=)ix_val)))
                     with | Some(s) -> if s=d then Instr.Nop
                                       else R64sR64d(I_R64sR64d.Movq, {S=Q(s); D=Q(d)})
                          | None -> //eprintfn "NO INFO1 %A (%A) mcontent = %A, content(rax) = %A" i (Q(d)) m_content iregfile.[0] 
                                    i
                else //eprintfn "NO INFO2 %A" i
                     if !dodump then eprintfn "  LOAD NEQ"
                     i
//                    iregfile.[d] <- ((d_idx,[||])::d_content)
            | _ ->
//                eprintfn "LOAD unknown %A" i
                if !dodump then eprintfn "LOAD unknown"
                i
//                iregfile.[d] <- ((d_idx,[||])::d_content)

    (* STORE *)
    | R64sM64d(I_R64sM64d.Movq,{S=Q(s); D=M64.S(m)}) ->
        (* TBD does it potentially alias?   if so start new state about which we know NOTHING *)
        let (s_idx0,s_content) = iregfile.[s]
        let m_vars = ea_variations iregfile m
        
//        match m_vars |> set_tryFind (function (k,map) -> map.Count = 1 && (match Map.tryFind (ParamQ(0)) map with | Some(x) -> x = 1L | None -> false)) with
        match m_vars |> set_tryFind (function (k,map) -> map.Count = 1 && (match Map.tryFind (Var(R64.rsp,1)) map with | Some(x) -> x = 1L | None -> false)) with
            | Some((k,_) as m') ->
                if k < 0L then
                    let k = int k
                    if !dodump then printfn "MEM1> memory %A (%A)" memory m'

//                    let size = 8
//                    for i = 0 to size-1 do memory.[int(k)+i] <- Si((Q(s), s_idx0), i, size, s_content)

                    let s_content = Set.add (make_LinEx (Var(Q(s),s_idx0))) s_content

                    let size = 8
                    if [|0..size-1|] |> Array.forall (fun ix ->
                                                        let (found,value) = memory.TryGetValue(int(k)+ix)
                                                        if found then
                                                          match value with
                                                            | Si(ix',size',ix_content) -> ix' = ix &&
                                                                                          size' = size &&
                                                                                          sets_intersect ix_content s_content
                                                            | _ -> false
                                                        else false
                                                     ) then
                        for i = 0 to size-1 do memory.[int(k)+i] <- Si(i, size, s_content)
                                                         
//                        eprintfn "elimNEW %A" i
                        Instr.Nop
                    else
//                        eprintfn "NO STORE2 %A" i
                        i
                    
                else
                    if !dodump then printfn "MEM2> stack outside %A (%A)" memory m'
                    i
                
            | _ ->
                if m_vars |> Set.exists (R_occursIn_LinEx R64.rsp) then
                    if !dodump then printfn "MEM4 > clear"
                    memCLEAR()
                else
                    if !dodump then printfn "MEM3> memunknown %A" m

                i

    | i -> i


let ai_map_instr_omit_frame_pointer (memory:stackT) (iregfile:iregs) (i:Instr) =
    let m_vars = ea_variations iregfile (M.B(R64.rsp))
    
//    match m_vars |> set_tryFind (function (k,map) -> map.Count = 1 && (match Map.tryFind (ParamQ(0)) map with | Some(x) -> x = 1L | None -> false)) with
    match m_vars |> set_tryFind (function (k,map) -> map.Count = 1 && (match Map.tryFind (Var(R64.rsp,1)) map with | Some(x) -> x = 1L | None -> false)) with
        | Some((k1,_) as m') ->
           let i = ref i
           for m in M_of_Instr(!i) do  (* for M8/M16/M32/M64/M128/M256 *)
             let m_vars = ea_variations iregfile m
//             match m_vars |> set_tryFind (function (k,map) -> map.Count = 1 && (match Map.tryFind (ParamQ(0)) map with | Some(x) -> x = 1L | None -> false)) with
             match m_vars |> set_tryFind (function (k,map) -> map.Count = 1 && (match Map.tryFind (Var(R64.rsp,1)) map with | Some(x) -> x = 1L | None -> false)) with
               | Some((k,_) as m') ->
                   i := (!i |> Map_M_in_Instr (fun x -> if x = m then M.B_D(R64.rsp,k-k1) else x))
               | None -> ()
           !i
        | _ -> i



let ai_map_instr2 (memory:stackT) (iregfile:iregs) (i:Instr) =
    let all_m = Ms_of_Instr(i) @ Mds_of_Instr(i) @ Md_of_Instr(i)
    match all_m with // fold over all_m?
        | [] -> i
        | _::_::_ -> failwith "ai_map_instr2"
        | [(m_size,m)] when List.exists((=)R64.rsp) (R64_of_M(m)) -> i // eprintfn "found stackloc RSP"; i // should occur AFTER omit_frame_pointer
        | [(m_size,m)] ->
            let variations = ea_variations iregfile m
            let isCurrent (Q(r)) (ix:int) = ix = (iregfile.[r] |> fst)

//            eprintfn "ai_map_instr2: %A" m
            match variations
                  |> set_tryPick
                                (fun linex -> let (k,map) = linex
                                              let xs = Array.toList(Map.toArray(map))
//                                              eprintfn "  cand %A" linex
                                              match xs with
                                                   | [(ParamQ(q),ni); (Var(rJ,xj),nj)]
                                                   | [(Var(rJ,xj),nj); (ParamQ(q),ni)]

//                                                   | [(Var(rI,xi),ni); (Var(rJ,xj),nj)]
                                                     when isCurrent rJ xj && (ni = 1L || nj = 1L) && (not (ni = 1L && nj = 1L)) ->
//                                                     when ((xi = 1 && (not(isCurrent rI xi)) && isCurrent rJ xj) || (xj = 1 && isCurrent rI xi)) && (ni = 1L || nj = 1L) && (not (ni = 1L && nj = 1L)) ->
                                                       if ni = 1L && nj = 1L then eprintfn "ai_map_instr2: BOTH =1"
                                 
                                                       let offset_R64 (Q(r)) ofs = Q(r+ofs)
                                 
                                                       let s1 = Var(Q(100+q),1)
                                                       let s2 = Var(rJ,xj)
//                                                       let s1 = if xi = 1 && not(isCurrent rI xi) then Var(offset_R64 rI 100,xi)
//                                                                else Var(rI,xi)
//                                                       let s2 = if xj = 1 && not(isCurrent rJ xj) then Var(offset_R64 rJ 100,xj)
//                                                                else Var(rJ,xj)
                                 
                                                       let linEX = (k,Map [(s1,ni);(s2,nj)])
                                                       let opt = eaOption_of_linex(linEX)
//                                                       eprintfn "linex = %A: opt = %A" linEX opt
                                                       opt
                                                   | _ -> None)
             with | Some(m') -> Map_M_in_Instr (fun x -> if x = m then m' else x) i
                  | None -> i
   
(*
let ai_map_instr2 (memory:stackT) (iregfile:iregs) (i:Instr) =
    let all_m = Ms_of_Instr(i) @ Mds_of_Instr(i) @ Md_of_Instr(i)
    match all_m with // fold over all_m?
        | [] -> i
        | _::_::_ -> failwith "ai_map_instr2"
        | [(m_size,m)] when List.exists((=)R64.rsp) (R64_of_M(m)) -> eprintfn "found stackloc RSP"; i // should occur AFTER omit_frame_pointer
        | [(m_size,m)] ->
            let variations = ea_variations iregfile m
            let isCurrent (Q(r)) (ix:int) = ix = (iregfile.[r] |> List.head |> fst)

            match variations
                  |> set_tryPick
                                (fun linex -> let (k,map) = linex
                                              let xs = Array.toList(Map.toArray(map))
                                              match xs with

                                                   | [(Var(rI,xi),ni); (Var(rJ,xj),nj)]
                                                     when ((xi = 1 && (not(isCurrent rI xi)) && isCurrent rJ xj) || (xj = 1 && isCurrent rI xi)) && (ni = 1L || nj = 1L) && (not (ni = 1L && nj = 1L)) ->
                                                       if ni = 1L && nj = 1L then eprintfn "ai_map_instr2: BOTH =1"
                                 
                                                       let offset_R64 (Q(r)) ofs = Q(r+ofs)
                                 
                                                       let s1 = if xi = 1 && not(isCurrent rI xi) then Var(offset_R64 rI 100,xi)
                                                                else Var(rI,xi)
                                                       let s2 = if xj = 1 && not(isCurrent rJ xj) then Var(offset_R64 rJ 100,xj)
                                                                else Var(rJ,xj)
                                 
                                                       let linEX = (k,Map [(s1,ni);(s2,nj)])
                                                       let opt = eaOption_of_linex(linEX)
                                                       opt
                                                   | _ -> None)
             with | Some(m') -> Map_M_in_Instr (fun x -> if x = m then m' else x) i
                  | None -> i
*)

// does not include stack pointer and/or hidden parameters... (R64.r10,-2); (R64.rax,-1); (R64.rsp,0); 
let paramRegs = [| (R64.rdi,1); (R64.rsi,2); (R64.rdx,3); (R64.rcx,4); (R64.r8,5); (R64.r9,6) |]


let do_AI (tDict:Dictionary<String,IBlock>) (cc:String []) =
    let iregsIn  = new Dictionary<string,iregs>()
    let iregsOut = new Dictionary<string,iregs>()
    let stackIn  = new Dictionary<string,Set<int * stackItem>>()
    let stackOut = new Dictionary<string,Set<int * stackItem>>()


    let bfs (tDict:Dictionary<String,IBlock>) =
        let ordered = new ResizeArray<String>()
        let visited = new HashSet<String>()
        let queued = new Queue<String>()
        match try_get_single_entrypoint(cc) with Some(entryPoint) -> queued.Enqueue(entryPoint)
                                               | None -> failwith("bfs: not single entrypoint")
        while queued.Count > 0 do
            let name = queued.Dequeue()
            if not(visited.Contains(name)) then
                visited.Add(name) |> ignore
                ordered.Add(name)
                let b = tDict.[name]
                b.DynSuccs |> Array.iter (fun(n) -> queued.Enqueue(n))
        ordered.ToArray()

    let cc = bfs(tDict)

//    eprintfn "AI order: %A" cc

    let cc = cc |> Array.mapi (fun i l -> let b = tDict.[l]
                                          let ix = if isFunction(b.Name) then 1 else ((i+1) * 1000000)
                                          (b,l,i,ix))

    let prep changed (b,l,i,ix) =
        if b.DynPreds = [||] then
            let iregfile = Array.init 117 (* HACK *) (fun i -> (ix, Set.singleton (make_LinEx(Var(Q(i),ix)))))

            let stacks = let onstackParams = new ResizeArray<Int32 * stackItem>()
                         for q = 7 to 20 do let set = Set.singleton(make_LinEx(ParamQ(q)))
                                            for j = 0 to 7 do onstackParams.Add(8+(q-7)*8+j, Si(j,8,set))
                         onstackParams.ToArray() |> Set.ofArray

//            let paramRegs = [| (R64.r10,-2); (R64.rax,-1); (R64.rsp,0); (R64.rdi,1); (R64.rsi,2); (R64.rdx,3); (R64.rcx,4); (R64.r8,5); (R64.r9,6) |]

            paramRegs |> Array.iter (fun (r,q) -> iregfile.[r.Int] <- (ix, Set.singleton(make_LinEx(ParamQ(q)))))
//            paramRegs |> Array.iter (fun (r,q) -> iregfile.[r.Int] <- (ix, Set.ofList [make_LinEx(Var(Q(r.Int),ix)); make_LinEx(ParamQ(q))]))

            (iregfile,stacks)
        else 
            let mutable stacks = None

            if !dodump then eprintfn "\n\n========================== AI run START %s (%A)" l b.DynPreds

            let t00 = DateTime.Now
            let sets = Array.create 117 (* HACK *) None

            for pred in b.DynPreds do
                if not (iregsOut.ContainsKey(pred)) then
                                                        changed := true
                                                        if !dodump then eprintfn "    ! not contained %A" pred
                                                    else
                                                          let regs       = iregsOut.[pred]
                                                          let p_stackout = stackOut.[pred]

                                                          for i = 0 to (* HAAACK *) 115 do (* iteri *)
                                                             let (r,rcontent) = regs.[i] //|> List.head
                                                             sets.[i] <- Some(match sets.[i] with None -> rcontent | Some(x) -> Set.intersect x rcontent)

                                                          stacks <- Some(match stacks with None -> p_stackout | Some(x) -> Set.intersect x p_stackout)
                                                       
            let iregfile = sets |> Array.mapi (fun i opt -> let selfref = make_LinEx(Var(Q(i),ix)) // make_LinEx(Map[Var(Q(i),r),1L]) rcontent
                                                            let set = match opt with None -> Set.empty | Some(s) -> s
                                                            (ix, Set.add selfref set))

            let stack = match stacks with None -> Set.empty | Some(s) -> s
            (iregfile,stack)
            

    let rec loop niter =
        let changed = ref false
//        eprintfn "(start %d)" niter

        for (b,l,i,ix) in cc do
            let (iregfile,stack) = prep changed (b,l,i,ix)

            let changedhere = not(iregsIn.ContainsKey(l)) || iregsIn.[l] <> iregfile || (not(stackIn.ContainsKey(l))) || stackIn.[l] <> stack

            if !dodump then eprintfn "\n\n========================== AI run END %s (changed %A)" l changedhere
            let t01 = DateTime.Now

            if changedhere then
                dump_iregfile iregfile
                changed := true
                iregsIn.[l] <- Array.copy iregfile
                stackIn.[l] <- stack
                let stack' = new stackT()
                for (key,item) in stack do stack'.[key] <- item

                didmemclear := false
                b.Instrs |> Array.iter (interp_ stack' iregfile)
                if !didmemclear then eprintfn "memCLEAR"

                let stack'' = ArrayMisc.array_of_collection(stack') |> Array.map (fun(KeyValue(k,v)) -> (k,v)) |> Set.ofArray
                iregsOut.[l] <- iregfile
                stackOut.[l] <- stack''

                dump_iregfile iregfile

                if !dodump then eprintfn "AI state %s -------------------" l
                if !dodump then eprintfn "stack %A\n" stack''

        if !changed then
//            eprintf "(AI loop %d)" niter
            loop (niter+1) 
        else niter

    let niter = loop 1
    ((iregsIn,iregsOut),(stackIn,stackOut))


let do_AI1 (tDict:Dictionary<String,IBlock>) (cc:String []) ((iregsIn,iregsOut),(stackIn,stackOut)) (mapInstr:stackT -> iregs -> Instr -> Instr) =
    for lbl in cc do
        let b = tDict.[lbl]
        
        let iregfile = Array.copy((iregsIn:Dictionary<String,iregs>).[lbl])
        let stack    = (stackIn:Dictionary<String,Set<int*stackItem>>).[lbl]
        let stack'   = new stackT()
        for (key,item) in stack do stack'.[key] <- item

(*        eprintfn "AI1 block %s (%d instrs = %A)" lbl b.Instrs.Length b.Instrs *)

        b.Instrs <- b.Instrs |> Array.map (fun i -> let i1 = mapInstr stack' iregfile i
                                                    interp_ stack' iregfile i // XXX i1
                                                    i1)


// hoist out scaling of indices to the function entry point

let do_AI3 (tDict:Dictionary<string, IBlock>) cc =
    let extra_srcs = new Dictionary<R64,Int64>()
    let extra_dsts = new HashSet<R64>()

    let (yes,no) = Array.partition isFunction cc

    for lbl in cc do
        let b = tDict.[lbl]
        b.Instrs |> Array.iter (fun(i) -> let visited = new HashSet<R64>()
                                          // HACK
                                          let doReg (r:R64) (scale:Int64) = if r.Int >= 42 then let (found,value) = extra_srcs.TryGetValue(r)
                                                                                                // eprintfn "extra_srcs: %A s=%d" r scale
                                                                                                extra_srcs.[r] <- gcd64 scale (if found then value else 0L) // simply collect?
                                                                                                visited.Add(r) |> ignore


                                          M_of_Instr(i) |> List.iter (function EA64(ofs,baseOpt,scaledIndexOpt) -> match baseOpt        with None -> () | Some({Base=b})          -> doReg b 1L
                                                                                                                   match scaledIndexOpt with None -> () | Some({Index=i;Scale=s}) -> doReg i s
                                                                             | EA32(_) -> failwith "do_AI3: xxxEA32xxx"
                                                                             | _ -> ())
                                                                                                                                                                       
                                          Rs_of_Instr(i)  |> List.iter (fun r -> if r.Int >= 42 then if not(visited.Contains(Q(r.Int))) then extra_srcs.[Q(r.Int)] <- 1L)
                                          Rds_of_Instr(i) |> List.iter (fun r -> if r.Int >= 42 then extra_dsts.Add(Q(r.Int)) |> ignore)
                                          Rd_of_Instr(i)  |> List.iter (fun r -> if r.Int >= 42 then extra_dsts.Add(Q(r.Int)) |> ignore))

    if extra_dsts.Count = 0 && extra_srcs.Count > 0 && yes.Length = 1 then
        let func   = yes.[0]
        let b_func = tDict.[func]
        
        

        let newinstrs = ArrayMisc.array_of_collection(extra_srcs)
                        |> Array.map (fun (KeyValue(d,factor)) -> match paramRegs |> Array.tryPick (fun(reg,q) -> if d.Int = q+100 then Some(reg) else None) with
                                                                   | None ->
                                                                        let memParam = d.Int - 100 (* HACK *)
                                                                        if factor = 1L then
                                                                           M64sR64d(I_M64sR64d.Movq, {S=M64.S(EA64(Iexpr.Int32(8 + (memParam-7)*8), Some{Base=R64.rsp}, None)); D=d})
                                      
                                                                        else
                                                                           M64sR64d(I_M64sR64d.Imulq(Lit.Int64(factor)), {S=M64.S(EA64(Iexpr.Int32(8 + (memParam-7)*8), Some{Base=R64.rsp}, None)); D=d})

//                                                                        failwithf "do_AI3: not found: d=%A, factor=%d, (%s)" d factor !Global.InFileName
                                                                   | Some(s) ->
                                                                        // let s = Q(d.Int - 100) // HACK
                                                                        if s = R64.rsp then failwithf "[xxxRSPxxx] %s: %A, factor = %d" !Global.InFileName s factor // DEL
                                                                        if                     factor = 1L                 then R64sR64d(I_R64sR64d.Movq,                     {S=s;D=d})
                                                                        elif                   factor = 2L                 then R64d(I_R64d.Leaq(M.B_SI(s,s,1L)),                 {D=d})
                                                                        elif TargetIsAMD() && (factor = 4L || factor = 8L) then R64d(I_R64d.Leaq(M.SI(s,factor)),                 {D=d})
                                                                                                                           else R64sR64d(I_R64sR64d.Imulq(Lit.Int64(factor)), {S=s;D=d}))

        let instrs = Array.toList(Array.append newinstrs b_func.Instrs)
        let instrs = instrs |> moveDown_Instrs_like b_func (fun(i) -> List.exists (fun(r:R) -> r.Int >= 42 (* HACK *)) (Rd_of_Instr(i) @ Rds_of_Instr(i) @ Rs_of_Instr(i)))
        b_func.Instrs <- Array.ofList(instrs)
                                         
        let map_SI {Index=i;Scale=s} = let (found,value) = extra_srcs.TryGetValue(i) in {Index=i;Scale=s/(if found then value else 1L)}
        let map1 = function K64rip(x) -> K64rip(x)
                          | K32eip(x) -> K32eip(x)
                          | K(x) -> K(x)
                          | EA32(ofs,baseOpt,scaledIndexOpt)    -> failwith "xxxEA32(ofs,baseOpt,scaledIndexOpt)"
                          | EA64(ofs,baseOpt,None)              -> EA64(ofs,baseOpt,None)
                          | EA64(ofs,baseOpt,Some(scaledIndex)) -> EA64(ofs,baseOpt,Some(map_SI scaledIndex))
                       
        for lbl in cc do
                let b = tDict.[lbl]
                b.Instrs <- b.Instrs |> Array.map (Map_M_in_Instr map1)
                           
(***)
