
    
  let calc_fw_deps (idx,orig_order,idmap,deps,lastreads,lastwrite) aInstr =
    let id = Id.makeNew () in
    let addDepsR r_res =
        match operandMap_find r_res lastwrite with   (* RAW *)
        | None -> return
        | Some(w_id,w_lat) -> idmap_addE w_id ((RAW w_lat), id, w_id, w_lat)
    let addDepsW (w_res,_) deps =                    (* WAW and WAR *)
      fold_right
        (fun r_id -> idmap_addE r_id (WAR, id, r_id, 1))
        (operandMap_findE w_res lastreads) 
        (match operandMap_find w_res lastwrite with
           | Some (w_id,_) -> idmap_addE w_id (WAW, id, w_id, 1) deps
           | None -> deps)

                 
    let rs = Instr.inputs_of_aInstr aInstr
    let lat = Instr.latency_of_aInstr aInstr
    let ws = List.map (fun d -> (d,lat)) (Instr.outputs_of_aInstr aInstr)
    
    let lastreads' = List.fold_right (operandMap_addE' id) rs lastreads
    (idx+1,
       IdMap.add id idx orig_order,
       IdMap.add id aInstr idmap,
     
       List.fold_right 
         addDepsW ws (fold_right addDepsR rs (IdMap.add id [] deps)),
     
       List.fold_right (fun (res,_) -> OperandMap.add res []) ws lastreads',
       List.fold_right 
         (fun (res,lat) -> OperandMap.add res (id,lat)) ws lastwrite)

    
  let (idm0,gtm0) = (IdMap.empty, OperandMap.empty)
  let s0 = (0, idm0, idm0, idm0, gtm0, gtm0)

  let (_,orig_order,idmap,fw_deps,_,_) = List.fold_left calc_fw_deps s0 aInstrs
  let bw_deps_cnt =
    IdMap.fold 
      (fun _ -> List.fold_right (fun dep -> idmap_inc (depToSucc dep)))
      fw_deps (IdMap.map toZero1 fw_deps) 
  (orig_order, 
      (fun id -> IdMap.find id idmap), 
      (fun id -> IdMap.find id fw_deps), 
      bw_deps_cnt)

(* map the forward and backward-dependencies to critical-path lengths *)
let calc_critical_path_lengths fw_deps bw_deps_cnt = 
  let fold_start_instrs id cnt = if cnt = 0 then cons id else return
  let startinstrs = IdMap.fold fold_start_instrs bw_deps_cnt []
  let rec calc_cplen' id cplen = 
    if idmap_exists id cplen then cplen else calc_cplen'' id cplen
  and calc_cplen'' id cplen =
      match fw_deps id with
          | [] -> IdMap.add id 0 cplen
          | deps ->
              let cplen' = fold_right calc_cplen' (map depToSucc deps) cplen
              let depToCPLen' d = depToLatency(d) + IdMap.find (depToSucc d) cplen'
              IdMap.add id (max_of (map depToCPLen' deps)) cplen'
  let cplens = fold_right calc_cplen' startinstrs IdMap.empty 
  fun id -> IdMap.find id cplens

(***)









type allocationStatus =			(* UNIT ALLOCATION STATUS ***********)
  | Free of exUnit			(*   <1> is available for use       *)
  | Busy_until of exUnit * int		(*   <1> is blocked until t >= <2>  *)

let initialStatus_of_resources xs = List.map (fun x -> Free x) xs

let maybeReleaseResource t = function
  | Busy_until(r,until) when t > until -> Free r
  | x -> x

let isSuitableFreeResource u = function Free r -> r = u | _ -> false

let rec issueInstrM'' = function
  | And xs -> iterM issueInstrM'' xs
  | Or xs -> memberM xs >>= issueInstrM''
  | Use(u) -> issueInstrM'' (Use_N(u,1))
  | Use_N(u,n) ->
      fetchStateM () >>= fun (t,res) ->
        match selectFirst (isSuitableFreeResource u) res with
	  | None -> failM
	  | Some(Free r,res0) -> storeStateM (t, (Busy_until(r,t+n-1))::res0)
	  | _ -> failwith "issueInstrM"

let issueInstrM' (_,_,instr) =
  issueInstrM'' (resUse_of_aInstr instr)

let issueInstrM idmap instrs ready =
  iterM issueInstrM' instrs >>= fun _ ->
    selectM ready >>= fun ((CP_Id(cplen,id),rs) as p) ->
      issueInstrM'' (resUse_of_aInstr (idmap id)) >>= fun _ ->
        unitM p

(***)

let flatInsSched idmap fw_deps cplens =
  (* move instructions from 'preready' to 'ready' *)
  let rec addNewReady t ready preready =
    if ESTIdSet.is_empty preready then (ready,preready)
    else 
      let EST_Id(est,id) as x = ESTIdSet.min_elt preready in
      if t < est then (ready,preready)
      else 
	let instr = idmap id in
	let bias = match instr with
	  | (i,_) when instr_is_inlinableVecCopy i -> -25
(* XXX push last users! *)
	  | _ -> 0
        in
	let ready' = insert_list cmp_cpid (CP_Id(bias + cplens id,id)) ready
	in addNewReady t ready' (ESTIdSet.remove x preready)

  (* adapt earliest starting time. decrement the number of bw deps of an
   * instruction. if it is zero, move the instr from waiting to preready. *)
  and adaptEST t dep (preready,waiting) =
    let id = depToSucc dep in
    let (r_nBWdeps,r_est) = IdMap.find id waiting in
    let r_est' = max (t + (depToLatency dep)) r_est in
      if r_nBWdeps > 1 then (preready, IdMap.add id (r_nBWdeps-1, r_est') waiting)
      else (ESTIdSet.add (EST_Id(r_est',id)) preready, IdMap.remove id waiting)

  and loop resStat t nFreeSlots instrsT instrs (ready,preready) waiting =
    let complete () = rev (instrsT @ instrs)
    and startNewCycle () = 
      let (_,(_,resStat')) = 
	value_of_option (runM (iterM issueInstrM') instrsT (t,resStat)) in
      let t' = t + 1 in
      loop (map (maybeReleaseResource t') resStat') t' maxInstrThruput 
	   [] (instrsT @ instrs) (addNewReady t' ready preready) waiting in
    match (ready,preready,waiting) with
      | ([],p,w) when ESTIdSet.is_empty p && w = IdMap.empty -> complete ()
      | ([],p,_) when not (ESTIdSet.is_empty p) -> startNewCycle ()
      | ([],_,_) -> failwith "flatInsSched"
      | (_::_,_,_) when nFreeSlots = 0 -> startNewCycle ()
      | (_::_ as ready,_,_) ->
	  (* select an instr with the longest critical path that may be issued
	   * simultaneously with the ones already selected earlier. *)
	  match runM (issueInstrM idmap instrsT) ready (t,resStat) with
            | None -> startNewCycle ()
            | Some((CP_Id(cplen,id),ready'),_) ->
		let instr = idmap id in
		let ready'' = ready' in

		let cplen = cplens id in
		let () = assert (cplen >= 0) in
	        let (preready',waiting') = 
  		  fold_right (adaptEST t) (fw_deps id) (preready,waiting) in
		loop resStat t (nFreeSlots-1) ((t,cplen,instr)::instrsT)
		     instrs (addNewReady t ready'' preready') waiting' in
  loop (initialStatus_of_resources exUnits0) 0 maxInstrThruput [] []

(***)

let llschedule_aInstrs options aInstrs =
  let (orig_order,idmap,fw_deps,bw_deps_cnt) = calc_dependencies aInstrs in
  let toCP = calc_critical_path_lengths fw_deps bw_deps_cnt
  let foldReady id (n,_) = if n=0 then cons (CP_Id(toCP id,id)) else return in
  let all = IdMap.map (fun cnt -> (cnt, 0)) bw_deps_cnt in
  let ready = sort cmp_cpid (IdMap.fold foldReady all []) in
  let waiting = fold_right (fun (CP_Id(_,i)) -> IdMap.remove i) ready all 
  in List.map 
       (fun (t,cplen,(instr,annotations)) -> 
	  (instr,[T t; CPl cplen] @ annotations)) 
       (flatInsSched idmap fw_deps toCP (ready, ESTIdSet.empty) waiting)

let llschedule_procedure options p = 
  Instr.map_aInstrs_in_procedure (llschedule_aInstrs options) p
(*
= function
  | Procedure(name,args,aDecls,cDecls,aInstrs,loop) ->
      let aInstrs = llschedule_aInstrs options aInstrs
      in Procedure(name,args,aDecls,cDecls,aInstrs,loop)
*)

