module ListScheduling_AMD_K10

open X64instr

open ListScheduling


let (@) ncycles resource = TAKE(resource,ncycles)

let xuse_of_Instr(x:Instr) =
    match x with
        | R64sR64d(I_R64sR64d.Movq, _) -> 1@Dcod
        | R64sM64d(I_R64sM64d.Movq, _) -> ALL [1@Dcod; 1@Fstore]

        | M128sXd((I_M128sXd.Movaps | I_M128sXd.Movapd), _) -> ALL [1@Dcod; 1@Load]

        | M64sXd((I_M64sXd.Movq | I_M64sXd.Movsd | I_M64sXd.Movddup), _) -> ALL [1@Dcod; 1@Load]

        | M64sR64d(I_M64sR64d.Movq, _) -> ALL [1@Dcod; 1@Load]


        | R64d(I_R64d.Leaq(_), _) -> 1@Dcod

        | XsXd((I_XsXd.Movaps | I_XsXd.Movapd), _) -> ALL [1@Dcod; FIRST [1@Fmul; 1@Fadd; 1@Fstore]]
        
        | XsXds((I_XsXds.Subpd | I_XsXds.Subps | I_XsXds.Addps | I_XsXds.Addpd), _) -> ALL [1@Dcod; 1@Fadd]
        | XsXds((I_XsXds.Mulps | I_XsXds.Mulpd), _) -> ALL [1@Dcod; 1@Fmul]


        | XsXds((I_XsXds.Xorps | I_XsXds.Xorpd | I_XsXds.Unpcklps | I_XsXds.Unpcklpd | I_XsXds.Unpckhps | I_XsXds.Unpckhpd | I_XsXds.Shufpd _ | I_XsXds.Shufps _), _) -> ALL [1@Dcod; FIRST [1@Fmul; 1@Fadd]]


        | M128sXds((I_M128sXds.Subpd | I_M128sXds.Subps | I_M128sXds.Addps | I_M128sXds.Addpd), _) -> ALL [1@Dcod; 1@Load; 1@Fadd]
        | M128sXds((I_M128sXds.Mulps | I_M128sXds.Mulpd), _) -> ALL [1@Dcod; 1@Load; 1@Fmul]

        | M128sXds((I_M128sXds.Xorps | I_M128sXds.Xorpd | I_M128sXds.Unpcklps | I_M128sXds.Unpcklpd | I_M128sXds.Unpckhps | I_M128sXds.Unpckhpd | I_M128sXds.Shufpd _ | I_M128sXds.Shufps _), _) -> ALL [1@Dcod; 1@Load; FIRST [1@Fmul; 1@Fadd]]

        | M64sR64ds((I_M64sR64ds.Addq | I_M64sR64ds.Subq), _) -> ALL [1@Dcod; 1@Load]

        | R64ds(I_R64ds.Decq, _) -> 1@Dcod
        | R64sR64s(I_R64sR64s.Testq, _) -> 1@Dcod
        | C(_) -> 1@Dcod
        

        | XsM128d((I_XsM128d.Movapd | I_XsM128d.Movaps), _) -> ALL [1@Fstore; 1@Dcod; 1@Dcod]

        | R64sR64d(I_R64sR64d.Imulq(_),_) -> ALL [1@Dcod; 1@Alu0]
        | R32sR32d(I_R32sR32d.Imull(_),_) -> ALL [1@Dcod; 1@Alu0]

        | M64sR64d(I_M64sR64d.Imulq(_),_) -> ALL [1@Dcod; 1@Dcod; 1@Dcod; 1@Alu0] (* vector path *)
        | M32sR32d(I_M32sR32d.Imull(_),_) -> ALL [1@Dcod; 1@Dcod; 1@Dcod; 1@Alu0] (* vector path *)


        | I0(I_I0.Retq) -> 1@Dcod
        | I0(I_I0.Ret _) -> ALL [1@Dcod; 1@Dcod] (* directpath double *)


        | M64ds(I_M64ds.Shlq(_), _) -> ALL [1@Dcod; 1@Load; 1@Fstore]
                 
        | _ -> 
//            eprintfn "AMDk10: unknown xuse: %A" x
            1@Dcod


(* all int register only instructions have latency 1 *)

let latency_of_Instr(x:Instr) =
    match x with

        | M64sR64d(I_M64sR64d.Movq, _) -> 3
        | R64sM64d(I_R64sM64d.Movq, _) -> 3

        | R32d(I_R32d.Leal(_), _) -> 2
        | R64d(I_R64d.Leaq(_), _) -> 2
                
        | R32sR32d(I_R32sR32d.Imull(_),_) -> 3
        | R64sR64d(I_R64sR64d.Imulq(_),_) -> 4
                

        (* all other integer register-to-register instructions have latency 1 *)

                                  
        | R8d _ | R16d _ | R32d _ | R64d _ -> 1

        
        | R8ds _ | R16ds _ | R32ds _ | R64ds _ -> 1
        | R8s _ | R16s _ | R32s _ | R64s _ -> 1
        | R8sR8s _ | R16sR16s _ | R32sR32s _ | R64sR64s _ -> 1

        
        | R8sR8d _ | R8sR16d _ | R8sR32d _ | R8sR64d _ | R16sR16d _ | R16sR32d _ | R16sR64d _ | R32sR32d _ | R32sR64d _ | R64sR64d _ -> 1
        | R8dsR8ds _ | R16dsR16ds _ | R32dsR32ds _ | R64dsR64ds _ -> 1
        | R8sR16sR16ds _ | R8sR32sR32ds _ | R8sR64sR64ds _ -> 1
        | R8sR8ds _ | R8sR16ds _ | R8sR32ds _ | R8sR64ds _ | R16sR16ds _ | R32sR32ds _ | R64sR64ds _ -> 1

                                  (* wrong numbers for shldw/... *)


        (* all other memory-to-register/register-to-memory instructions have latency 4 *)

        | M8d _ | M16d _ | M32d _ | M64d _ -> 4
        | M8ds _ | M16ds _ | M32ds _ | M64ds _ | M128ds _ -> 4
        | M8s _ | M16s _ | M32s _ | M64s _ -> 4

        | M8sR8d _ | M8sR16d _ | M8sR32d _ | M8sR64d _ | M16sR16d _ | M16sR32d _ | M16sR64d _ | M32sR32d _ | M32sR64d _ | M64sR64d _ -> 4
        
        | R8sM8d _ | R16sM16d _ | R32sM32d _ | R64sM64d _ -> 4

        | M8sR8ds _ | M16sR16ds _ | M32sR32ds _ | M64sR64ds _ -> 4
        | R8sM8ds _ | R16sM16ds _ | R32sM32ds _ | R64sM64ds _ -> 4
        | R8sR16sM16ds _ | R8sR32sM32ds _ | R8sR64sM64ds _ -> 4
        
        
        | M8dsR8ds _ | M16dsR16ds _ | M32dsR32ds _ | M64dsR64ds _ -> 4
        | R8dsM8ds _ | R16dsM16ds _ | R32dsM32ds _ | R64dsM64ds _ -> 4
        | R8sM8s _ | R16sM16s _ | R32sM32s _ | R64sM64s _ -> 4
        | M8sR8s _ | M16sR16s _ | M32sR32s _ | M64sR64s _ -> 4

(* exception:
imul

*)


(*

        | R64ds(I_R64ds.Decq,_) -> 1
        | R64sR64s(I_R64sR64s.Testq,_) -> 1
*)

        | M128sXd((I_M128sXd.Movaps | I_M128sXd.Movapd),_) -> 4


        | XsXd((I_XsXd.Movapd | I_XsXd.Movaps), _) -> 2

        | XsXds((I_XsXds.Xorpd | I_XsXds.Xorps), _) -> 2
        | XsXds((I_XsXds.Subpd | I_XsXds.Subps | I_XsXds.Addps | I_XsXds.Addpd | I_XsXds.Mulps | I_XsXds.Mulpd), _) -> 4
        | XsXds((I_XsXds.Shufpd _ | I_XsXds.Shufps _), _) -> 2

        | M128sXds((I_M128sXds.Xorpd | I_M128sXds.Xorps), _) -> 4
        | M128sXds((I_M128sXds.Subpd | I_M128sXds.Subps | I_M128sXds.Addps | I_M128sXds.Addpd | I_M128sXds.Mulps | I_M128sXds.Mulpd), _) -> 6
        | M128sXds((I_M128sXds.Shufpd _ | I_M128sXds.Shufps _), _) -> 4

        | XsM128d((I_XsM128d.Movapd | I_XsM128d.Movaps), _) -> 2



        | I0(I_I0.Retq) | I0(I_I0.Ret _) -> 4 (* ret _ -> directpath double *)
        (* leave directpath double *)

        | C(_) -> 4
        | I0 _ -> 4
//        | Y0 _ -> 4

        (* indirect jumps/subroutine calls *)
        | NR64s _ | NM64s _ -> 4

        | _ -> //eprintfn "AMDk10: unknown latency %A" x;
               4



let schedParams = { XResources           = [ Fadd;Fmul;Fstore; Load;Load; Alu0;Alu1;Alu2; Dcod;Dcod;Dcod ]
                    MaxInstrsPerCycle    = 3
                    MaxIbytesPerCycle    = 32
                    MaxReadPorts         = 0
                    MaxOoOcycles         = 4
                    ReadPortNotNeededFor = 100000
                    Mode                 = CritPathCC
                    GetXUse              = xuse_of_Instr
                    GetLatency           = latency_of_Instr
                    NextXRes             = id }




