module SSEtoAVX

open X64instr

(*
   to_2opForm : Instr -> option<Instr []>
   of_2opForm : Instr -> option<Instr []>


   sse_to_avx

   avx_to_sse
*)

(*
SSEtoAVX

AVXtoSSE
*)

(***)

let convert_Xds i op ds = match op with I_Xds.Pslld(l)  -> XsXd(I_XsXd.Vpslld(l),  {S=ds;D=ds})
                                      | I_Xds.Pslldq(l) -> XsXd(I_XsXd.Vpslldq(l), {S=ds;D=ds})
                                      | I_Xds.Psllq(l)  -> XsXd(I_XsXd.Vpsllq(l),  {S=ds;D=ds})
                                      | I_Xds.Psllw(l)  -> XsXd(I_XsXd.Vpsllw(l),  {S=ds;D=ds})
                                      | I_Xds.Psrad(l)  -> XsXd(I_XsXd.Vpsrad(l),  {S=ds;D=ds})
                                      | I_Xds.Psraw(l)  -> XsXd(I_XsXd.Vpsraw(l),  {S=ds;D=ds})
                                      | I_Xds.Psrld(l)  -> XsXd(I_XsXd.Vpsrld(l),  {S=ds;D=ds})
                                      | I_Xds.Psrldq(l) -> XsXd(I_XsXd.Vpsrldq(l), {S=ds;D=ds})
                                      | I_Xds.Psrlq(l)  -> XsXd(I_XsXd.Vpsrlq(l),  {S=ds;D=ds})
                                      | I_Xds.Psrlw(l)  -> XsXd(I_XsXd.Vpsrlw(l),  {S=ds;D=ds})
                                      | I_Xds.Extrq(_)  -> i
(***)

let convert_XsXd i op s d = match op with I_XsXd.Aesimc             -> XsXd(I_XsXd.Vaesimc,             {S=s;D=d})
                                        | I_XsXd.Aeskeygenassist(l) -> XsXd(I_XsXd.Vaeskeygenassist(l), {S=s;D=d})
                                        | I_XsXd.Cvtdq2pd           -> XsXd(I_XsXd.Vcvtdq2pd,           {S=s;D=d})
                                        | I_XsXd.Cvtdq2ps           -> XsXd(I_XsXd.Vcvtdq2ps,           {S=s;D=d})
                                        | I_XsXd.Cvtpd2dq           -> XsXd(I_XsXd.Vcvtpd2dq,           {S=s;D=d})
                                        | I_XsXd.Cvtpd2ps           -> XsXd(I_XsXd.Vcvtpd2ps,           {S=s;D=d})
                                        | I_XsXd.Cvtps2dq           -> XsXd(I_XsXd.Vcvtps2dq,           {S=s;D=d})
                                        | I_XsXd.Cvtps2pd           -> XsXd(I_XsXd.Vcvtps2pd,           {S=s;D=d})
                                        | I_XsXd.Cvttpd2dq          -> XsXd(I_XsXd.Vcvttpd2dq,          {S=s;D=d})
                                        | I_XsXd.Cvttps2dq          -> XsXd(I_XsXd.Vcvttps2dq,          {S=s;D=d})
                                        | I_XsXd.Movapd             -> XsXd(I_XsXd.Vmovapd,             {S=s;D=d})
                                        | I_XsXd.Movaps             -> XsXd(I_XsXd.Vmovaps,             {S=s;D=d})
                                        | I_XsXd.Movddup            -> XsXd(I_XsXd.Vmovddup,            {S=s;D=d})
                                        | I_XsXd.Movdqa             -> XsXd(I_XsXd.Vmovdqa,             {S=s;D=d})
                                        | I_XsXd.Movdqu             -> XsXd(I_XsXd.Vmovdqu,             {S=s;D=d})
                                        | I_XsXd.Movq               -> XsXd(I_XsXd.Vmovq,               {S=s;D=d})
                                        | I_XsXd.Movshdup           -> XsXd(I_XsXd.Vmovshdup,           {S=s;D=d})
                                        | I_XsXd.Movsldup           -> XsXd(I_XsXd.Vmovsldup,           {S=s;D=d})
                                        | I_XsXd.Movupd             -> XsXd(I_XsXd.Vmovupd,             {S=s;D=d})
                                        | I_XsXd.Movups             -> XsXd(I_XsXd.Vmovups,             {S=s;D=d})
                                        | I_XsXd.Pabsb              -> XsXd(I_XsXd.Vpabsb,              {S=s;D=d})
                                        | I_XsXd.Pabsd              -> XsXd(I_XsXd.Vpabsd,              {S=s;D=d})
                                        | I_XsXd.Pabsw              -> XsXd(I_XsXd.Vpabsw,              {S=s;D=d})
                                        | I_XsXd.Phminposuw         -> XsXd(I_XsXd.Vphminposuw,         {S=s;D=d})
                                        | I_XsXd.Pmovsxbd           -> XsXd(I_XsXd.Vpmovsxbd,           {S=s;D=d})
                                        | I_XsXd.Pmovsxbq           -> XsXd(I_XsXd.Vpmovsxbq,           {S=s;D=d})
                                        | I_XsXd.Pmovsxbw           -> XsXd(I_XsXd.Vpmovsxbw,           {S=s;D=d})
                                        | I_XsXd.Pmovsxdq           -> XsXd(I_XsXd.Vpmovsxdq,           {S=s;D=d})
                                        | I_XsXd.Pmovsxwd           -> XsXd(I_XsXd.Vpmovsxwd,           {S=s;D=d})
                                        | I_XsXd.Pmovsxwq           -> XsXd(I_XsXd.Vpmovsxwq,           {S=s;D=d})
                                        | I_XsXd.Pmovzxbd           -> XsXd(I_XsXd.Vpmovzxbd,           {S=s;D=d})
                                        | I_XsXd.Pmovzxbq           -> XsXd(I_XsXd.Vpmovzxbq,           {S=s;D=d})
                                        | I_XsXd.Pmovzxbw           -> XsXd(I_XsXd.Vpmovzxbw,           {S=s;D=d})
                                        | I_XsXd.Pmovzxdq           -> XsXd(I_XsXd.Vpmovzxdq,           {S=s;D=d})
                                        | I_XsXd.Pmovzxwd           -> XsXd(I_XsXd.Vpmovzxwd,           {S=s;D=d})
                                        | I_XsXd.Pmovzxwq           -> XsXd(I_XsXd.Vpmovzxwq,           {S=s;D=d})
                                        | I_XsXd.Pshufd(l)          -> XsXd(I_XsXd.Vpshufd(l),          {S=s;D=d})
                                        | I_XsXd.Pshufhw(l)         -> XsXd(I_XsXd.Vpshufhw(l),         {S=s;D=d})
                                        | I_XsXd.Pshuflw(l)         -> XsXd(I_XsXd.Vpshuflw(l),         {S=s;D=d})
                                        | I_XsXd.Rcpps              -> XsXd(I_XsXd.Vrcpps,              {S=s;D=d})
                                        | I_XsXd.Roundpd(l)         -> XsXd(I_XsXd.Vroundpd(l),         {S=s;D=d})
                                        | I_XsXd.Roundps(l)         -> XsXd(I_XsXd.Vroundps(l),         {S=s;D=d})
                                        | I_XsXd.Rsqrtps            -> XsXd(I_XsXd.Vrsqrtps,            {S=s;D=d})
                                        | I_XsXd.Sqrtpd             -> XsXd(I_XsXd.Vsqrtpd,             {S=s;D=d})
                                        | I_XsXd.Sqrtps             -> XsXd(I_XsXd.Vsqrtps,             {S=s;D=d})
                                        | I_XsXd.Vphminposuw         | I_XsXd.Vcvtdq2pd   | I_XsXd.Vcvtdq2ps | I_XsXd.Vcvtpd2dq | I_XsXd.Vcvtpd2ps  -> i
                                        | I_XsXd.Vcvtps2pd           | I_XsXd.Vpermilps _ | I_XsXd.Vcvtps2dq | I_XsXd.Vpabsw    | I_XsXd.Vmovups    -> i
                                        | I_XsXd.Vpermilpd _         | I_XsXd.Vmovshdup   | I_XsXd.Vmovsldup | I_XsXd.Vpabsb    | I_XsXd.Vpabsd     -> i 
                                        | I_XsXd.Vaeskeygenassist _  | I_XsXd.Vaesimc     | I_XsXd.Vphsubbw  | I_XsXd.Vphsubdq  | I_XsXd.Vphsubwd   -> i 
                                        | I_XsXd.Vcvttpd2dq          | I_XsXd.Vcvttps2dq  | I_XsXd.Vfrczpd   | I_XsXd.Vfrczps   | I_XsXd.Vfrczsd    | I_XsXd.Vfrczss    -> i
                                        | I_XsXd.Vmovapd             | I_XsXd.Vmovaps     | I_XsXd.Vmovddup  | I_XsXd.Vmovdqa   | I_XsXd.Vmovdqu    | I_XsXd.Vmovq      -> i
                                        | I_XsXd.Vphaddbd            | I_XsXd.Vphaddbq    | I_XsXd.Vphaddbw  | I_XsXd.Vphadddq  | I_XsXd.Vphaddwd   | I_XsXd.Vphaddwq   -> i
                                        | I_XsXd.Vphaddubd           | I_XsXd.Vphaddubq   | I_XsXd.Vphaddubw | I_XsXd.Vphaddudq | I_XsXd.Vphadduwd  | I_XsXd.Vphadduwq  -> i
                                        | I_XsXd.Vpmovsxbd           | I_XsXd.Vpmovsxbq   | I_XsXd.Vpmovsxbw | I_XsXd.Vpmovsxdq | I_XsXd.Vpmovsxwd  | I_XsXd.Vpmovsxwq  -> i
                                        | I_XsXd.Vpmovzxbd           | I_XsXd.Vpmovzxbq   | I_XsXd.Vpmovzxbw | I_XsXd.Vpmovzxdq | I_XsXd.Vpmovzxwd  | I_XsXd.Vpmovzxwq  -> i
                                        | I_XsXd.Vprotb _            | I_XsXd.Vprotd _    | I_XsXd.Vprotq _  | I_XsXd.Vprotw _  | I_XsXd.Vpshufd _  | I_XsXd.Vpshufhw _ -> i
                                        | I_XsXd.Vpslld _            | I_XsXd.Vpslldq _   | I_XsXd.Vpsllq _  | I_XsXd.Vpsllw _  | I_XsXd.Vpshuflw _ | I_XsXd.Vmovupd    -> i
                                        | I_XsXd.Vpsrad _            | I_XsXd.Vpsraw _    | I_XsXd.Vpsrld _  | I_XsXd.Vpsrldq _ | I_XsXd.Vpsrlq _   | I_XsXd.Vpsrlw _   -> i
                                        | I_XsXd.Vroundpd _          | I_XsXd.Vroundps _  | I_XsXd.Vrcpps    | I_XsXd.Vrsqrtps  | I_XsXd.Vsqrtpd    | I_XsXd.Vsqrtps    -> i
                                        
(***)

let convert_XsXds i op s ds = match op with I_XsXds.Addpd        -> XsXsXd(I_XsXsXd.Vaddpd,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Addps        -> XsXsXd(I_XsXsXd.Vaddps,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Addsd        -> XsXsXd(I_XsXsXd.Vaddsd,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Addss        -> XsXsXd(I_XsXsXd.Vaddss,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Addsubpd     -> XsXsXd(I_XsXsXd.Vaddsubpd,     {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Addsubps     -> XsXsXd(I_XsXsXd.Vaddsubps,     {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Aesdec       -> XsXsXd(I_XsXsXd.Vaesdec,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Aesdeclast   -> XsXsXd(I_XsXsXd.Vaesdeclast,   {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Aesenc       -> XsXsXd(I_XsXsXd.Vaesenc,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Aesenclast   -> XsXsXd(I_XsXsXd.Vaesenclast,   {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Andnpd       -> XsXsXd(I_XsXsXd.Vandnpd,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Andnps       -> XsXsXd(I_XsXsXd.Vandnps,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Andpd        -> XsXsXd(I_XsXsXd.Vandpd,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Andps        -> XsXsXd(I_XsXsXd.Vandps,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Blendpd(l)   -> XsXsXd(I_XsXsXd.Vblendpd(l),   {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Blendps(l)   -> XsXsXd(I_XsXsXd.Vblendps(l),   {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Blendvpd     -> failwith "blendvpd"
                                          | I_XsXds.Blendvps     -> failwith "blendvps"
                                          | I_XsXds.Cmppd(l)     -> XsXsXd(I_XsXsXd.Vcmppd(l),     {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Cmpps(l)     -> XsXsXd(I_XsXsXd.Vcmpps(l),     {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Cmpsd(l)     -> XsXsXd(I_XsXsXd.Vcmpsd(l),     {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Cmpss(l)     -> XsXsXd(I_XsXsXd.Vcmpss(l),     {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Cvtsd2ss     -> XsXsXd(I_XsXsXd.Vcvtsd2ss,     {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Cvtss2sd     -> XsXsXd(I_XsXsXd.Vcvtss2sd,     {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Divpd        -> XsXsXd(I_XsXsXd.Vdivpd,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Divps        -> XsXsXd(I_XsXsXd.Vdivps,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Divsd        -> XsXsXd(I_XsXsXd.Vdivsd,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Divss        -> XsXsXd(I_XsXsXd.Vdivss,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Dppd(l)      -> XsXsXd(I_XsXsXd.Vdppd(l),      {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Dpps(l)      -> XsXsXd(I_XsXsXd.Vdpps(l),      {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Haddpd       -> XsXsXd(I_XsXsXd.Vhaddpd,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Haddps       -> XsXsXd(I_XsXsXd.Vhaddps,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Hsubpd       -> XsXsXd(I_XsXsXd.Vhsubpd,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Hsubps       -> XsXsXd(I_XsXsXd.Vhsubps,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Insertps(l)  -> XsXsXd(I_XsXsXd.Vinsertps(l),  {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Maxpd        -> XsXsXd(I_XsXsXd.Vmaxpd,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Maxps        -> XsXsXd(I_XsXsXd.Vmaxps,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Maxsd        -> XsXsXd(I_XsXsXd.Vmaxsd,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Maxss        -> XsXsXd(I_XsXsXd.Vmaxss,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Minpd        -> XsXsXd(I_XsXsXd.Vminpd,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Minps        -> XsXsXd(I_XsXsXd.Vminps,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Minsd        -> XsXsXd(I_XsXsXd.Vminsd,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Minss        -> XsXsXd(I_XsXsXd.Vminss,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Movhlps      -> XsXsXd(I_XsXsXd.Vmovhlps,      {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Movlhps      -> XsXsXd(I_XsXsXd.Vmovlhps,      {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Movsd        -> XsXsXd(I_XsXsXd.Vmovsd,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Movss        -> XsXsXd(I_XsXsXd.Vmovss,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Mpsadbw(l)   -> XsXsXd(I_XsXsXd.Vmpsadbw(l),   {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Mulpd        -> XsXsXd(I_XsXsXd.Vmulpd,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Mulps        -> XsXsXd(I_XsXsXd.Vmulps,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Mulsd        -> XsXsXd(I_XsXsXd.Vmulsd,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Mulss        -> XsXsXd(I_XsXsXd.Vmulss,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Orpd         -> XsXsXd(I_XsXsXd.Vorpd,         {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Orps         -> XsXsXd(I_XsXsXd.Vorps,         {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Packssdw     -> XsXsXd(I_XsXsXd.Vpackssdw,     {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Packsswb     -> XsXsXd(I_XsXsXd.Vpacksswb,     {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Packusdw     -> XsXsXd(I_XsXsXd.Vpackusdw,     {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Packuswb     -> XsXsXd(I_XsXsXd.Vpackuswb,     {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Paddb        -> XsXsXd(I_XsXsXd.Vpaddb,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Paddd        -> XsXsXd(I_XsXsXd.Vpaddd,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Paddq        -> XsXsXd(I_XsXsXd.Vpaddq,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Paddsb       -> XsXsXd(I_XsXsXd.Vpaddsb,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Paddsw       -> XsXsXd(I_XsXsXd.Vpaddsw,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Paddusb      -> XsXsXd(I_XsXsXd.Vpaddusb,      {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Paddusw      -> XsXsXd(I_XsXsXd.Vpaddusw,      {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Paddw        -> XsXsXd(I_XsXsXd.Vpaddw,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Palignr(l)   -> XsXsXd(I_XsXsXd.Vpalignr(l),   {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pand         -> XsXsXd(I_XsXsXd.Vpand,         {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pandn        -> XsXsXd(I_XsXsXd.Vpandn,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pavgb        -> XsXsXd(I_XsXsXd.Vpavgb,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pavgw        -> XsXsXd(I_XsXsXd.Vpavgw,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pblendvb     -> failwith "pblendvb"
                                          | I_XsXds.Pblendw(l)   -> failwith "pblendw"
                                          | I_XsXds.Pclmulqdq(l) -> XsXsXd(I_XsXsXd.Vpclmulqdq(l), {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pcmpeqb      -> XsXsXd(I_XsXsXd.Vpcmpeqb,      {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pcmpeqd      -> XsXsXd(I_XsXsXd.Vpcmpeqd,      {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pcmpeqq      -> XsXsXd(I_XsXsXd.Vpcmpeqq,      {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pcmpeqw      -> XsXsXd(I_XsXsXd.Vpcmpeqw,      {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pcmpgtb      -> XsXsXd(I_XsXsXd.Vpcmpgtb,      {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pcmpgtd      -> XsXsXd(I_XsXsXd.Vpcmpgtd,      {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pcmpgtq      -> XsXsXd(I_XsXsXd.Vpcmpgtq,      {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pcmpgtw      -> XsXsXd(I_XsXsXd.Vpcmpgtw,      {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Phaddd       -> XsXsXd(I_XsXsXd.Vphaddd,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Phaddsw      -> XsXsXd(I_XsXsXd.Vphaddsw,      {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Phaddw       -> XsXsXd(I_XsXsXd.Vphaddw,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Phsubd       -> XsXsXd(I_XsXsXd.Vphsubd,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Phsubsw      -> XsXsXd(I_XsXsXd.Vphsubsw,      {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Phsubw       -> XsXsXd(I_XsXsXd.Vphsubw,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pmaddubsw    -> XsXsXd(I_XsXsXd.Vpmaddubsw,    {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pmaddwd      -> XsXsXd(I_XsXsXd.Vpmaddwd,      {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pmaxsb       -> XsXsXd(I_XsXsXd.Vpmaxsb,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pmaxsd       -> XsXsXd(I_XsXsXd.Vpmaxsd,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pmaxsw       -> XsXsXd(I_XsXsXd.Vpmaxsw,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pmaxub       -> XsXsXd(I_XsXsXd.Vpmaxub,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pmaxud       -> XsXsXd(I_XsXsXd.Vpmaxud,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pmaxuw       -> XsXsXd(I_XsXsXd.Vpmaxuw,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pminsb       -> XsXsXd(I_XsXsXd.Vpminsb,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pminsd       -> XsXsXd(I_XsXsXd.Vpminsd,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pminsw       -> XsXsXd(I_XsXsXd.Vpminsw,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pminub       -> XsXsXd(I_XsXsXd.Vpminub,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pminud       -> XsXsXd(I_XsXsXd.Vpminud,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pminuw       -> XsXsXd(I_XsXsXd.Vpminuw,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pmuldq       -> XsXsXd(I_XsXsXd.Vpmuldq,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pmulhrsw     -> XsXsXd(I_XsXsXd.Vpmulhrsw,     {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pmulhuw      -> XsXsXd(I_XsXsXd.Vpmulhuw,      {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pmulhw       -> XsXsXd(I_XsXsXd.Vpmulhw,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pmulld       -> XsXsXd(I_XsXsXd.Vpmulld,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pmullw       -> XsXsXd(I_XsXsXd.Vpmullw,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pmuludq      -> XsXsXd(I_XsXsXd.Vpmuludq,      {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Por          -> XsXsXd(I_XsXsXd.Vpor,          {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Psadbw       -> XsXsXd(I_XsXsXd.Vpsadbw,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pshufb       -> XsXsXd(I_XsXsXd.Vpshufb,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Psignb       -> XsXsXd(I_XsXsXd.Vpsignb,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Psignd       -> XsXsXd(I_XsXsXd.Vpsignd,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Psignw       -> XsXsXd(I_XsXsXd.Vpsignw,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pslld        -> XsXsXd(I_XsXsXd.Vpslld,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Psllq        -> XsXsXd(I_XsXsXd.Vpsllq,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Psllw        -> XsXsXd(I_XsXsXd.Vpsllw,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Psrad        -> XsXsXd(I_XsXsXd.Vpsrad,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Psraw        -> XsXsXd(I_XsXsXd.Vpsraw,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Psrld        -> XsXsXd(I_XsXsXd.Vpsrld,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Psrlq        -> XsXsXd(I_XsXsXd.Vpsrlq,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Psrlw        -> XsXsXd(I_XsXsXd.Vpsrlw,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Psubb        -> XsXsXd(I_XsXsXd.Vpsubb,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Psubd        -> XsXsXd(I_XsXsXd.Vpsubd,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Psubq        -> XsXsXd(I_XsXsXd.Vpsubq,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Psubsb       -> XsXsXd(I_XsXsXd.Vpsubsb,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Psubsw       -> XsXsXd(I_XsXsXd.Vpsubsw,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Psubusb      -> XsXsXd(I_XsXsXd.Vpsubusb,      {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Psubusw      -> XsXsXd(I_XsXsXd.Vpsubusw,      {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Psubw        -> XsXsXd(I_XsXsXd.Vpsubw,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Punpckhbw    -> XsXsXd(I_XsXsXd.Vpunpckhbw,    {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Punpckhdq    -> XsXsXd(I_XsXsXd.Vpunpckhdq,    {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Punpckhqdq   -> XsXsXd(I_XsXsXd.Vpunpckhqdq,   {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Punpckhwd    -> XsXsXd(I_XsXsXd.Vpunpckhwd,    {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Punpcklbw    -> XsXsXd(I_XsXsXd.Vpunpcklbw,    {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Punpckldq    -> XsXsXd(I_XsXsXd.Vpunpckldq,    {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Punpcklqdq   -> XsXsXd(I_XsXsXd.Vpunpcklqdq,   {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Punpcklwd    -> XsXsXd(I_XsXsXd.Vpunpcklwd,    {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Pxor         -> XsXsXd(I_XsXsXd.Vpxor,         {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Rcpss        -> XsXsXd(I_XsXsXd.Vrcpss,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Roundsd(l)   -> XsXsXd(I_XsXsXd.Vroundsd(l),   {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Roundss(l)   -> XsXsXd(I_XsXsXd.Vroundss(l),   {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Rsqrtss      -> XsXsXd(I_XsXsXd.Vrsqrtss,      {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Shufpd(l)    -> XsXsXd(I_XsXsXd.Vshufpd(l),    {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Shufps(l)    -> XsXsXd(I_XsXsXd.Vshufps(l),    {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Sqrtsd       -> XsXsXd(I_XsXsXd.Vsqrtsd,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Sqrtss       -> XsXsXd(I_XsXsXd.Vsqrtss,       {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Subpd        -> XsXsXd(I_XsXsXd.Vsubpd,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Subps        -> XsXsXd(I_XsXsXd.Vsubps,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Subsd        -> XsXsXd(I_XsXsXd.Vsubsd,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Subss        -> XsXsXd(I_XsXsXd.Vsubss,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Unpckhpd     -> XsXsXd(I_XsXsXd.Vunpckhpd,     {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Unpckhps     -> XsXsXd(I_XsXsXd.Vunpckhps,     {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Unpcklpd     -> XsXsXd(I_XsXsXd.Vunpcklpd,     {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Unpcklps     -> XsXsXd(I_XsXsXd.Vunpcklps,     {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Xorpd        -> XsXsXd(I_XsXsXd.Vxorpd,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Xorps        -> XsXsXd(I_XsXsXd.Vxorps,        {S1=s;S2=ds;D=ds})
                                          | I_XsXds.Extrq         | I_XsXds.Insertq | I_XsXds.Insertq_(_) -> i

(***)

let convert_XsXs i op s1 s2 = match op with I_XsXs.Comisd      -> XsXs(I_XsXs.Vcomisd,     {S1=s1;S2=s2})
                                          | I_XsXs.Comiss      -> XsXs(I_XsXs.Vcomiss,     {S1=s1;S2=s2})
                                          | I_XsXs.Ucomisd     -> XsXs(I_XsXs.Vucomisd,    {S1=s1;S2=s2})
                                          | I_XsXs.Ucomiss     -> XsXs(I_XsXs.Vucomiss,    {S1=s1;S2=s2})
                                          | I_XsXs.Maskmovdqu  -> XsXs(I_XsXs.Vmaskmovdqu, {S1=s1;S2=s2})
                                          | I_XsXs.Ptest       -> XsXs(I_XsXs.Vptest,      {S1=s1;S2=s2})
                                          | I_XsXs.Pcmpistri   -> XsXs(I_XsXs.Vpcmpistri,  {S1=s1;S2=s2})
                                          | I_XsXs.Pcmpistrm   -> XsXs(I_XsXs.Vpcmpistrm,  {S1=s1;S2=s2})
                                          | I_XsXs.Pcmpestri   -> XsXs(I_XsXs.Vpcmpestri,  {S1=s1;S2=s2})
                                          | I_XsXs.Pcmpestrm   -> XsXs(I_XsXs.Vpcmpestrm,  {S1=s1;S2=s2})
                                          | I_XsXs.Vptest       | I_XsXs.Vtestpd    | I_XsXs.Vtestps    -> i
                                          | I_XsXs.Vmaskmovdqu  | I_XsXs.Vcomisd    | I_XsXs.Vcomiss    -> i
                                          | I_XsXs.Vucomisd     | I_XsXs.Vucomiss   | I_XsXs.Vpcmpistri -> i
                                          | I_XsXs.Vpcmpistrm   | I_XsXs.Vpcmpestri | I_XsXs.Vpcmpestrm -> i

(***)

let convert_M32sXs  i op s1 s2 = match op with I_M32sXs.Comiss      -> M32sXs(I_M32sXs.Vcomiss,      {S1=s1;S2=s2})
                                             | I_M32sXs.Ucomiss     -> M32sXs(I_M32sXs.Vucomiss,     {S1=s1;S2=s2})
                                             | I_M32sXs.Vcomiss      | I_M32sXs.Vucomiss -> i
let convert_M64sXs  i op s1 s2 = match op with I_M64sXs.Comisd      -> M64sXs(I_M64sXs.Vcomisd,      {S1=s1;S2=s2})
                                             | I_M64sXs.Ucomisd     -> M64sXs(I_M64sXs.Vucomisd,     {S1=s1;S2=s2})
                                             | I_M64sXs.Vcomisd      | I_M64sXs.Vucomisd -> i
let convert_M128sXs i op s1 s2 = match op with I_M128sXs.Pcmpestri  -> M128sXs(I_M128sXs.Vpcmpestri, {S1=s1;S2=s2})
                                             | I_M128sXs.Pcmpestrm  -> M128sXs(I_M128sXs.Vpcmpestrm, {S1=s1;S2=s2})
                                             | I_M128sXs.Pcmpistri  -> M128sXs(I_M128sXs.Vpcmpistri, {S1=s1;S2=s2})
                                             | I_M128sXs.Pcmpistrm  -> M128sXs(I_M128sXs.Vpcmpistrm, {S1=s1;S2=s2})
                                             | I_M128sXs.Ptest      -> M128sXs(I_M128sXs.Vptest,     {S1=s1;S2=s2})
                                             | I_M128sXs.Vpcmpestri  | I_M128sXs.Vpcmpestrm | I_M128sXs.Vpcmpistri | I_M128sXs.Vpcmpistrm -> i
                                             | I_M128sXs.Vptest      | I_M128sXs.Vtestpd    | I_M128sXs.Vtestps                           -> i
(***)

let convert_M16sXd  i op s d = match op with I_M16sXd.Pmovsxbq            -> M16sXd(I_M16sXd.Vpmovsxbq,             {S=s;D=d})
                                           | I_M16sXd.Pmovzxbq            -> M16sXd(I_M16sXd.Vpmovzxbq,             {S=s;D=d})
                                           | I_M16sXd.Vpmovsxbq            | I_M16sXd.Vpmovzxbq -> i
let convert_M32sXd  i op s d = match op with I_M32sXd.Movd                -> M32sXd(I_M32sXd.Vmovd,                 {S=s;D=d})
                                           | I_M32sXd.Movss               -> M32sXd(I_M32sXd.Vmovss,                {S=s;D=d})
                                           | I_M32sXd.Pmovsxbd            -> M32sXd(I_M32sXd.Vpmovsxbd,             {S=s;D=d})
                                           | I_M32sXd.Pmovsxwq            -> M32sXd(I_M32sXd.Vpmovsxwq,             {S=s;D=d})
                                           | I_M32sXd.Pmovzxbd            -> M32sXd(I_M32sXd.Vpmovzxbd,             {S=s;D=d})
                                           | I_M32sXd.Pmovzxwq            -> M32sXd(I_M32sXd.Vpmovzxwq,             {S=s;D=d})
                                           | I_M32sXd.Vmovss               | I_M32sXd.Vmovd     | I_M32sXd.Vbroadcastss -> i
                                           | I_M32sXd.Vpmovsxbd            | I_M32sXd.Vpmovsxwq | I_M32sXd.Vpmovzxbd     | I_M32sXd.Vpmovzxwq -> i
let convert_M64sXd  i op s d = match op with I_M64sXd.Movddup             -> M64sXd(I_M64sXd.Vmovddup,              {S=s;D=d})
                                           | I_M64sXd.Movq                -> M64sXd(I_M64sXd.Vmovq,                 {S=s;D=d})
                                           | I_M64sXd.Movsd               -> M64sXd(I_M64sXd.Vmovsd,                {S=s;D=d})
                                           | I_M64sXd.Pmovsxbw            -> M64sXd(I_M64sXd.Vpmovsxbw,             {S=s;D=d})
                                           | I_M64sXd.Pmovsxdq            -> M64sXd(I_M64sXd.Vpmovsxdq,             {S=s;D=d})
                                           | I_M64sXd.Pmovsxwd            -> M64sXd(I_M64sXd.Vpmovsxwd,             {S=s;D=d})
                                           | I_M64sXd.Pmovzxbw            -> M64sXd(I_M64sXd.Vpmovzxbw,             {S=s;D=d})
                                           | I_M64sXd.Pmovzxdq            -> M64sXd(I_M64sXd.Vpmovzxdq,             {S=s;D=d})
                                           | I_M64sXd.Pmovzxwd            -> M64sXd(I_M64sXd.Vpmovzxwd,             {S=s;D=d})
                                           | I_M64sXd.Vmovddup             | I_M64sXd.Vmovq  | I_M64sXd.Vpmovsxdq | I_M64sXd.Vpmovzxbw | I_M64sXd.Cvtpi2pd  -> i
                                           | I_M64sXd.Vpmovsxbw            | I_M64sXd.Vmovsd | I_M64sXd.Vpmovsxwd | I_M64sXd.Vpmovzxdq | I_M64sXd.Vpmovzxwd -> i
                                           
let convert_M128sXd i op s d = match op with I_M128sXd.Aesimc             -> M128sXd(I_M128sXd.Vaesimc,             {S=s;D=d})
                                           | I_M128sXd.Aeskeygenassist(l) -> M128sXd(I_M128sXd.Vaeskeygenassist(l), {S=s;D=d})
                                           | I_M128sXd.Cvtdq2pd           -> M128sXd(I_M128sXd.Vcvtdq2pd,           {S=s;D=d})
                                           | I_M128sXd.Cvtdq2ps           -> M128sXd(I_M128sXd.Vcvtdq2ps,           {S=s;D=d})
                                           | I_M128sXd.Cvtpd2dq           -> M128sXd(I_M128sXd.Vcvtpd2dq,           {S=s;D=d})
                                           | I_M128sXd.Cvtpd2ps           -> M128sXd(I_M128sXd.Vcvtpd2ps,           {S=s;D=d})
                                           | I_M128sXd.Cvtps2dq           -> M128sXd(I_M128sXd.Vcvtps2dq,           {S=s;D=d})
                                           | I_M128sXd.Cvtps2pd           -> M128sXd(I_M128sXd.Vcvtps2pd,           {S=s;D=d})
                                           | I_M128sXd.Cvttpd2dq          -> M128sXd(I_M128sXd.Vcvttpd2dq,          {S=s;D=d})
                                           | I_M128sXd.Cvttps2dq          -> M128sXd(I_M128sXd.Vcvttps2dq,          {S=s;D=d})
                                           | I_M128sXd.Lddqu              -> M128sXd(I_M128sXd.Vlddqu,              {S=s;D=d})
                                           | I_M128sXd.Movapd             -> M128sXd(I_M128sXd.Vmovapd,             {S=s;D=d})
                                           | I_M128sXd.Movaps             -> M128sXd(I_M128sXd.Vmovaps,             {S=s;D=d})
                                           | I_M128sXd.Movdqa             -> M128sXd(I_M128sXd.Vmovdqa,             {S=s;D=d})
                                           | I_M128sXd.Movdqu             -> M128sXd(I_M128sXd.Vmovdqu,             {S=s;D=d})
                                           | I_M128sXd.Movupd             -> M128sXd(I_M128sXd.Vmovupd,             {S=s;D=d})
                                           | I_M128sXd.Movups             -> M128sXd(I_M128sXd.Vmovups,             {S=s;D=d})
                                           | I_M128sXd.Pabsb              -> M128sXd(I_M128sXd.Vpabsb,              {S=s;D=d})
                                           | I_M128sXd.Pabsd              -> M128sXd(I_M128sXd.Vpabsd,              {S=s;D=d})
                                           | I_M128sXd.Pabsw              -> M128sXd(I_M128sXd.Vpabsw,              {S=s;D=d})
                                           | I_M128sXd.Phminposuw         -> M128sXd(I_M128sXd.Vphminposuw,         {S=s;D=d})
                                           | I_M128sXd.Pshufd(l)          -> M128sXd(I_M128sXd.Vpshufd(l),          {S=s;D=d})
                                           | I_M128sXd.Pshufhw(l)         -> M128sXd(I_M128sXd.Vpshufhw(l),         {S=s;D=d})
                                           | I_M128sXd.Pshuflw(l)         -> M128sXd(I_M128sXd.Vpshuflw(l),         {S=s;D=d})
                                           | I_M128sXd.Rcpps              -> M128sXd(I_M128sXd.Vrcpps,              {S=s;D=d})
                                           | I_M128sXd.Roundpd(l)         -> M128sXd(I_M128sXd.Vroundps(l),         {S=s;D=d})
                                           | I_M128sXd.Roundps(l)         -> M128sXd(I_M128sXd.Vroundpd(l),         {S=s;D=d})
                                           | I_M128sXd.Rsqrtps            -> M128sXd(I_M128sXd.Vrsqrtps,            {S=s;D=d})
                                           | I_M128sXd.Sqrtpd             -> M128sXd(I_M128sXd.Vsqrtpd,             {S=s;D=d})
                                           | I_M128sXd.Sqrtps             -> M128sXd(I_M128sXd.Vsqrtps,             {S=s;D=d})
                                           | I_M128sXd.Vaeskeygenassist _  | I_M128sXd.Vaesimc   | I_M128sXd.Vcvttpd2dq  -> i
                                           | I_M128sXd.Vcvttps2dq          | I_M128sXd.Vfrczpd   | I_M128sXd.Vfrczps     -> i
                                           | I_M128sXd.Vfrczsd             | I_M128sXd.Vfrczss   | I_M128sXd.Vcvtdq2pd   -> i
                                           | I_M128sXd.Vcvtdq2ps           | I_M128sXd.Vcvtpd2dq | I_M128sXd.Vcvtpd2ps   -> i
                                           | I_M128sXd.Vcvtps2dq           | I_M128sXd.Vcvtps2pd | I_M128sXd.Vmovapd     -> i
                                           | I_M128sXd.Vmovaps             | I_M128sXd.Vmovdqa   | I_M128sXd.Vmovdqu     -> i
                                           | I_M128sXd.Vmovupd             | I_M128sXd.Vmovups   | I_M128sXd.Vpabsb      -> i
                                           | I_M128sXd.Vpabsd              | I_M128sXd.Vpabsw    | I_M128sXd.Vphaddbd    -> i
                                           | I_M128sXd.Vphaddbq            | I_M128sXd.Vphaddbw  | I_M128sXd.Vphadddq    -> i
                                           | I_M128sXd.Vphaddwd            | I_M128sXd.Vphaddwq  | I_M128sXd.Vphaddubd   -> i
                                           | I_M128sXd.Vphaddubq           | I_M128sXd.Vphaddubw | I_M128sXd.Vphaddudq   -> i
                                           | I_M128sXd.Vphadduwd           | I_M128sXd.Vphadduwq | I_M128sXd.Vphsubbw    -> i
                                           | I_M128sXd.Vphsubdq            | I_M128sXd.Vphsubwd  | I_M128sXd.Vprotb _    -> i
                                           | I_M128sXd.Vprotd _            | I_M128sXd.Vprotq _  | I_M128sXd.Vphminposuw -> i
                                           | I_M128sXd.Vlddqu              | I_M128sXd.Vprotw _  | I_M128sXd.Vpshuflw _  -> i
                                           | I_M128sXd.Vpshufhw _          | I_M128sXd.Vpshufd _ | I_M128sXd.Vroundps _  -> i
                                           | I_M128sXd.Vroundpd _          | I_M128sXd.Vrcpps    | I_M128sXd.Vrsqrtps    -> i
                                           | I_M128sXd.Vsqrtpd             | I_M128sXd.Vsqrtps  -> i

(***)

let convert_M8sXds   i op s ds = match op with I_M8sXds.Pinsrb(l)      -> M8sXsXd(I_M8sXsXd.Vpinsrb(l),        {S1=s;S2=ds;D=ds})
let convert_M16sXds  i op s ds = match op with I_M16sXds.Pinsrw(l)     -> M16sXsXd(I_M16sXsXd.Vpinsrw(l),      {S1=s;S2=ds;D=ds})
let convert_M32sXds  i op s ds = match op with I_M32sXds.Addss         -> M32sXsXd(I_M32sXsXd.Vaddss,          {S1=s;S2=ds;D=ds})
                                             | I_M32sXds.Cmpss(l)      -> M32sXsXd(I_M32sXsXd.Vcmpss(l),       {S1=s;S2=ds;D=ds}) 
                                             | I_M32sXds.Divss         -> M32sXsXd(I_M32sXsXd.Vdivss,          {S1=s;S2=ds;D=ds})
                                             | I_M32sXds.Insertps(l)   -> M32sXsXd(I_M32sXsXd.Vinsertps(l),    {S1=s;S2=ds;D=ds})
                                             | I_M32sXds.Maxss         -> M32sXsXd(I_M32sXsXd.Vmaxss,          {S1=s;S2=ds;D=ds})
                                             | I_M32sXds.Minss         -> M32sXsXd(I_M32sXsXd.Vminss,          {S1=s;S2=ds;D=ds})
                                             | I_M32sXds.Mulss         -> M32sXsXd(I_M32sXsXd.Vmulss,          {S1=s;S2=ds;D=ds})
                                             | I_M32sXds.Roundss(l)    -> M32sXsXd(I_M32sXsXd.Vroundss(l),     {S1=s;S2=ds;D=ds})
                                             | I_M32sXds.Rsqrtss       -> M32sXsXd(I_M32sXsXd.Vrsqrtss,        {S1=s;S2=ds;D=ds})
                                             | I_M32sXds.Sqrtss        -> M32sXsXd(I_M32sXsXd.Vsqrtss,         {S1=s;S2=ds;D=ds})
                                             | I_M32sXds.Subss         -> M32sXsXd(I_M32sXsXd.Vsubss,          {S1=s;S2=ds;D=ds})
                                             | I_M32sXds.Pinsrd(l)     -> M32sXsXd(I_M32sXsXd.Vpinsrd(l),      {S1=s;S2=ds;D=ds})
                                             | I_M32sXds.Cvtsi2sdl     -> M32sXsXd(I_M32sXsXd.Vcvtsi2sdl,      {S1=s;S2=ds;D=ds})
                                             | I_M32sXds.Cvtss2sd      -> M32sXsXd(I_M32sXsXd.Vcvtss2sd,       {S1=s;S2=ds;D=ds})
                                             | I_M32sXds.Cvtsi2ssl     -> M32sXsXd(I_M32sXsXd.Vcvtsi2ssl,      {S1=s;S2=ds;D=ds}) 
let convert_M64sXds  i op s ds = match op with I_M64sXds.Addsd         -> M64sXsXd(I_M64sXsXd.Vaddsd,          {S1=s;S2=ds;D=ds})
                                             | I_M64sXds.Cmpsd(l)      -> M64sXsXd(I_M64sXsXd.Vcmpsd(l),       {S1=s;S2=ds;D=ds})
                                             | I_M64sXds.Divsd         -> M64sXsXd(I_M64sXsXd.Vdivsd,          {S1=s;S2=ds;D=ds})
                                             | I_M64sXds.Maxsd         -> M64sXsXd(I_M64sXsXd.Vmaxsd,          {S1=s;S2=ds;D=ds})
                                             | I_M64sXds.Minsd         -> M64sXsXd(I_M64sXsXd.Vminsd,          {S1=s;S2=ds;D=ds})
                                             | I_M64sXds.Movhpd        -> M64sXsXd(I_M64sXsXd.Vmovhpd,         {S1=s;S2=ds;D=ds})
                                             | I_M64sXds.Movhps        -> M64sXsXd(I_M64sXsXd.Vmovhps,         {S1=s;S2=ds;D=ds})
                                             | I_M64sXds.Movlpd        -> M64sXsXd(I_M64sXsXd.Vmovlpd,         {S1=s;S2=ds;D=ds})
                                             | I_M64sXds.Movlps        -> M64sXsXd(I_M64sXsXd.Vmovlps,         {S1=s;S2=ds;D=ds})
                                             | I_M64sXds.Mulsd         -> M64sXsXd(I_M64sXsXd.Vmulsd,          {S1=s;S2=ds;D=ds})
                                             | I_M64sXds.Roundsd(l)    -> M64sXsXd(I_M64sXsXd.Vroundsd(l),     {S1=s;S2=ds;D=ds})
                                             | I_M64sXds.Sqrtsd        -> M64sXsXd(I_M64sXsXd.Vsqrtsd,         {S1=s;S2=ds;D=ds})
                                             | I_M64sXds.Subsd         -> M64sXsXd(I_M64sXsXd.Vsubsd,          {S1=s;S2=ds;D=ds})
                                             | I_M64sXds.Pinsrq(l)     -> M64sXsXd(I_M64sXsXd.Vpinsrq(l),      {S1=s;S2=ds;D=ds})
                                             | I_M64sXds.Cvtsd2ss      -> M64sXsXd(I_M64sXsXd.Vcvtsd2ss,       {S1=s;S2=ds;D=ds})
                                             | I_M64sXds.Cvtsi2sdq     -> M64sXsXd(I_M64sXsXd.Vcvtsi2sdq,      {S1=s;S2=ds;D=ds})
                                             | I_M64sXds.Cvtsi2ssq     -> M64sXsXd(I_M64sXsXd.Vcvtsi2ssq,      {S1=s;S2=ds;D=ds})
                                             | I_M64sXds.Cvtpi2ps      -> i 
let convert_M128sXds i op s ds = match op with I_M128sXds.Addpd        -> M128sXsXd(I_M128sXsXd.Vaddpd,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Addps        -> M128sXsXd(I_M128sXsXd.Vaddps,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Addsubpd     -> M128sXsXd(I_M128sXsXd.Vaddsubpd,     {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Addsubps     -> M128sXsXd(I_M128sXsXd.Vaddsubps,     {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Aesdec       -> M128sXsXd(I_M128sXsXd.Vaesdec,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Aesdeclast   -> M128sXsXd(I_M128sXsXd.Vaesdeclast,   {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Aesenc       -> M128sXsXd(I_M128sXsXd.Vaesenc,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Aesenclast   -> M128sXsXd(I_M128sXsXd.Vaesenclast,   {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Andnpd       -> M128sXsXd(I_M128sXsXd.Vandnpd,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Andnps       -> M128sXsXd(I_M128sXsXd.Vandnps,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Andpd        -> M128sXsXd(I_M128sXsXd.Vandpd,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Andps        -> M128sXsXd(I_M128sXsXd.Vandps,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Blendpd(l)   -> M128sXsXd(I_M128sXsXd.Vblendpd(l),   {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Blendps(l)   -> M128sXsXd(I_M128sXsXd.Vblendps(l),   {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Blendvpd     -> failwith "blendvpd"
                                             | I_M128sXds.Blendvps     -> failwith "blendvps"
                                             | I_M128sXds.Cmppd(l)     -> M128sXsXd(I_M128sXsXd.Vcmppd(l),     {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Cmpps(l)     -> M128sXsXd(I_M128sXsXd.Vcmpps(l),     {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Divpd        -> M128sXsXd(I_M128sXsXd.Vdivpd,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Divps        -> M128sXsXd(I_M128sXsXd.Vdivps,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Dppd(l)      -> M128sXsXd(I_M128sXsXd.Vdppd(l),      {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Dpps(l)      -> M128sXsXd(I_M128sXsXd.Vdpps(l),      {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Haddpd       -> M128sXsXd(I_M128sXsXd.Vhaddpd,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Haddps       -> M128sXsXd(I_M128sXsXd.Vhaddps,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Hsubpd       -> M128sXsXd(I_M128sXsXd.Vhsubpd,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Hsubps       -> M128sXsXd(I_M128sXsXd.Vhsubps,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Maxpd        -> M128sXsXd(I_M128sXsXd.Vmaxpd,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Maxps        -> M128sXsXd(I_M128sXsXd.Vmaxps,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Minpd        -> M128sXsXd(I_M128sXsXd.Vminpd,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Minps        -> M128sXsXd(I_M128sXsXd.Vminps,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Mpsadbw(l)   -> M128sXsXd(I_M128sXsXd.Vmpsadbw(l),   {S1=s;S2=ds;D=ds}) 
                                             | I_M128sXds.Mulpd        -> M128sXsXd(I_M128sXsXd.Vmulpd,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Mulps        -> M128sXsXd(I_M128sXsXd.Vmulps,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Orpd         -> M128sXsXd(I_M128sXsXd.Vorpd,         {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Orps         -> M128sXsXd(I_M128sXsXd.Vorps,         {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Packssdw     -> M128sXsXd(I_M128sXsXd.Vpackssdw,     {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Packsswb     -> M128sXsXd(I_M128sXsXd.Vpacksswb,     {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Packusdw     -> M128sXsXd(I_M128sXsXd.Vpackusdw,     {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Packuswb     -> M128sXsXd(I_M128sXsXd.Vpackuswb,     {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Paddb        -> M128sXsXd(I_M128sXsXd.Vpaddb,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Paddd        -> M128sXsXd(I_M128sXsXd.Vpaddd,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Paddq        -> M128sXsXd(I_M128sXsXd.Vpaddq,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Paddsb       -> M128sXsXd(I_M128sXsXd.Vpaddsb,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Paddsw       -> M128sXsXd(I_M128sXsXd.Vpaddsw,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Paddusb      -> M128sXsXd(I_M128sXsXd.Vpaddusb,      {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Paddusw      -> M128sXsXd(I_M128sXsXd.Vpaddusw,      {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Paddw        -> M128sXsXd(I_M128sXsXd.Vpaddw,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pand         -> M128sXsXd(I_M128sXsXd.Vpand,         {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pandn        -> M128sXsXd(I_M128sXsXd.Vpandn,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pavgb        -> M128sXsXd(I_M128sXsXd.Vpavgb,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pavgw        -> M128sXsXd(I_M128sXsXd.Vpavgw,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pblendvb     -> failwith "pblendvb"
                                             | I_M128sXds.Pblendw(l)   -> failwith "pblendw"
                                             | I_M128sXds.Pclmulqdq(l) -> M128sXsXd(I_M128sXsXd.Vpclmulqdq(l), {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pcmpeqb      -> M128sXsXd(I_M128sXsXd.Vpcmpeqb,      {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pcmpeqd      -> M128sXsXd(I_M128sXsXd.Vpcmpeqd,      {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pcmpeqq      -> M128sXsXd(I_M128sXsXd.Vpcmpeqq,      {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pcmpeqw      -> M128sXsXd(I_M128sXsXd.Vpcmpeqw,      {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pcmpgtb      -> M128sXsXd(I_M128sXsXd.Vpcmpgtb,      {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pcmpgtd      -> M128sXsXd(I_M128sXsXd.Vpcmpgtd,      {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pcmpgtq      -> M128sXsXd(I_M128sXsXd.Vpcmpgtq,      {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pcmpgtw      -> M128sXsXd(I_M128sXsXd.Vpcmpgtw,      {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Phaddd       -> M128sXsXd(I_M128sXsXd.Vphaddd,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Phaddsw      -> M128sXsXd(I_M128sXsXd.Vphaddsw,      {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Phaddw       -> M128sXsXd(I_M128sXsXd.Vphaddw,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Phsubd       -> M128sXsXd(I_M128sXsXd.Vphsubd,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Phsubsw      -> M128sXsXd(I_M128sXsXd.Vphsubsw,      {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Phsubw       -> M128sXsXd(I_M128sXsXd.Vphsubw,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pmaddubsw    -> M128sXsXd(I_M128sXsXd.Vpmaddubsw,    {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pmaddwd      -> M128sXsXd(I_M128sXsXd.Vpmaddwd,      {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pmaxsb       -> M128sXsXd(I_M128sXsXd.Vpmaxsb,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pmaxsd       -> M128sXsXd(I_M128sXsXd.Vpmaxsd,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pmaxsw       -> M128sXsXd(I_M128sXsXd.Vpmaxsw,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pmaxub       -> M128sXsXd(I_M128sXsXd.Vpmaxub,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pmaxud       -> M128sXsXd(I_M128sXsXd.Vpmaxud,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pmaxuw       -> M128sXsXd(I_M128sXsXd.Vpmaxuw,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pminsb       -> M128sXsXd(I_M128sXsXd.Vpminsb,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pminsd       -> M128sXsXd(I_M128sXsXd.Vpminsd,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pminsw       -> M128sXsXd(I_M128sXsXd.Vpminsw,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pminub       -> M128sXsXd(I_M128sXsXd.Vpminub,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pminud       -> M128sXsXd(I_M128sXsXd.Vpminud,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pminuw       -> M128sXsXd(I_M128sXsXd.Vpminuw,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pmuldq       -> M128sXsXd(I_M128sXsXd.Vpmuldq,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pmulhrsw     -> M128sXsXd(I_M128sXsXd.Vpmulhrsw,     {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pmulhuw      -> M128sXsXd(I_M128sXsXd.Vpmulhuw,      {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pmulhw       -> M128sXsXd(I_M128sXsXd.Vpmulhw,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pmulld       -> M128sXsXd(I_M128sXsXd.Vpmulld,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pmullw       -> M128sXsXd(I_M128sXsXd.Vpmullw,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pmuludq      -> M128sXsXd(I_M128sXsXd.Vpmuludq,      {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Por          -> M128sXsXd(I_M128sXsXd.Vpor,          {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Psadbw       -> M128sXsXd(I_M128sXsXd.Vpsadbw,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pshufb       -> M128sXsXd(I_M128sXsXd.Vpshufb,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Psignb       -> M128sXsXd(I_M128sXsXd.Vpsignb,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Psignd       -> M128sXsXd(I_M128sXsXd.Vpsignd,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Psignw       -> M128sXsXd(I_M128sXsXd.Vpsignw,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pslld        -> M128sXsXd(I_M128sXsXd.Vpslld,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Psllq        -> M128sXsXd(I_M128sXsXd.Vpsllq,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Psllw        -> M128sXsXd(I_M128sXsXd.Vpsllw,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Psrad        -> M128sXsXd(I_M128sXsXd.Vpsrad,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Psraw        -> M128sXsXd(I_M128sXsXd.Vpsraw,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Psrld        -> M128sXsXd(I_M128sXsXd.Vpsrld,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Psrlq        -> M128sXsXd(I_M128sXsXd.Vpsrlq,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Psrlw        -> M128sXsXd(I_M128sXsXd.Vpsrlw,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Psubb        -> M128sXsXd(I_M128sXsXd.Vpsubb,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Psubd        -> M128sXsXd(I_M128sXsXd.Vpsubd,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Psubq        -> M128sXsXd(I_M128sXsXd.Vpsubq,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Psubsb       -> M128sXsXd(I_M128sXsXd.Vpsubsb,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Psubsw       -> M128sXsXd(I_M128sXsXd.Vpsubsw,       {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Psubusb      -> M128sXsXd(I_M128sXsXd.Vpsubusb,      {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Psubusw      -> M128sXsXd(I_M128sXsXd.Vpsubusw,      {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Psubw        -> M128sXsXd(I_M128sXsXd.Vpsubw,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Punpckhbw    -> M128sXsXd(I_M128sXsXd.Vpunpckhbw,    {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Punpckhdq    -> M128sXsXd(I_M128sXsXd.Vpunpckhdq,    {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Punpckhqdq   -> M128sXsXd(I_M128sXsXd.Vpunpckhqdq,   {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Punpckhwd    -> M128sXsXd(I_M128sXsXd.Vpunpckhwd,    {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Punpcklbw    -> M128sXsXd(I_M128sXsXd.Vpunpcklbw,    {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Punpckldq    -> M128sXsXd(I_M128sXsXd.Vpunpckldq,    {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Punpcklqdq   -> M128sXsXd(I_M128sXsXd.Vpunpcklqdq,   {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Punpcklwd    -> M128sXsXd(I_M128sXsXd.Vpunpcklwd,    {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Pxor         -> M128sXsXd(I_M128sXsXd.Vpxor,         {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Shufpd(l)    -> M128sXsXd(I_M128sXsXd.Vshufpd(l),    {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Shufps(l)    -> M128sXsXd(I_M128sXsXd.Vshufps(l),    {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Subpd        -> M128sXsXd(I_M128sXsXd.Vsubpd,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Subps        -> M128sXsXd(I_M128sXsXd.Vsubps,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Unpckhpd     -> M128sXsXd(I_M128sXsXd.Vunpckhpd,     {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Unpckhps     -> M128sXsXd(I_M128sXsXd.Vunpckhps,     {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Unpcklpd     -> M128sXsXd(I_M128sXsXd.Vunpcklpd,     {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Unpcklps     -> M128sXsXd(I_M128sXsXd.Vunpcklps,     {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Xorpd        -> M128sXsXd(I_M128sXsXd.Vxorpd,        {S1=s;S2=ds;D=ds})
                                             | I_M128sXds.Xorps        -> M128sXsXd(I_M128sXsXd.Vxorps,        {S1=s;S2=ds;D=ds})

(***)

let convert_R32sXd i op s d = match op with I_R32sXd.Movd  -> R32sXd(I_R32sXd.Vmovd, {S=s;D=d})
                                          | I_R32sXd.Vmovd -> i
let convert_R64sXd i op s d = match op with I_R64sXd.Movd  -> R64sXd(I_R64sXd.Vmovd, {S=s;D=d})
                                          | I_R64sXd.Movq  -> R64sXd(I_R64sXd.Vmovq, {S=s;D=d})
                                          | I_R64sXd.Vmovd  | I_R64sXd.Vmovq -> i

(***)

let convert_XsR32d i op s d = match op with I_XsR32d.Movd          -> XsR32d(I_XsR32d.Vmovd,          {S=s;D=d})
                                          | I_XsR32d.Extractpsl(l) -> XsR32d(I_XsR32d.Vextractpsl(l), {S=s;D=d})
                                          | I_XsR32d.Pmovmskbl     -> XsR32d(I_XsR32d.Vpmovmskbl,     {S=s;D=d})
                                          | I_XsR32d.Pextrb(l)     -> XsR32d(I_XsR32d.Vpextrb(l),     {S=s;D=d})
                                          | I_XsR32d.Pextrd(l)     -> XsR32d(I_XsR32d.Vpextrd(l),     {S=s;D=d})
                                          | I_XsR32d.Pextrw(l)     -> XsR32d(I_XsR32d.Vpextrw(l),     {S=s;D=d})
                                          | I_XsR32d.Movmskpdl     -> XsR32d(I_XsR32d.Vmovmskpdl,     {S=s;D=d})
                                          | I_XsR32d.Movmskpsl     -> XsR32d(I_XsR32d.Vmovmskpsl,     {S=s;D=d})
                                          | I_XsR32d.Cvtsd2sil     -> XsR32d(I_XsR32d.Vcvtsd2sil,     {S=s;D=d}) 
                                          | I_XsR32d.Cvttsd2sil    -> XsR32d(I_XsR32d.Vcvttsd2sil,    {S=s;D=d}) 
                                          | I_XsR32d.Cvtss2sil     -> XsR32d(I_XsR32d.Vcvtss2sil,     {S=s;D=d}) 
                                          | I_XsR32d.Cvttss2sil    -> XsR32d(I_XsR32d.Vcvttss2sil,    {S=s;D=d}) 
                                          | I_XsR32d.Vcvtsd2sil     | I_XsR32d.Vcvttsd2sil | I_XsR32d.Vcvtss2sil | I_XsR32d.Vcvttss2sil -> i
                                          | I_XsR32d.Vmovmskpsl     | I_XsR32d.Vmovmskpdl  | I_XsR32d.Vpmovmskbl | I_XsR32d.Vmovd       -> i
                                          | I_XsR32d.Vextractpsl _  | I_XsR32d.Vpextrb _   | I_XsR32d.Vpextrd _  | I_XsR32d.Vpextrw _   -> i
let convert_XsR64d i op s d = match op with I_XsR64d.Movd          -> XsR64d(I_XsR64d.Vmovd,          {S=s;D=d})
                                          | I_XsR64d.Movq          -> XsR64d(I_XsR64d.Vmovq,          {S=s;D=d})
                                          | I_XsR64d.Pmovmskbq     -> XsR64d(I_XsR64d.Vpmovmskbq,     {S=s;D=d})
                                          | I_XsR64d.Pextrb(l)     -> XsR64d(I_XsR64d.Vpextrb(l),     {S=s;D=d})
                                          | I_XsR64d.Pextrq(l)     -> XsR64d(I_XsR64d.Vpextrq(l),     {S=s;D=d})
                                          | I_XsR64d.Pextrw(l)     -> XsR64d(I_XsR64d.Vpextrw(l),     {S=s;D=d})
                                          | I_XsR64d.Extractpsq(l) -> XsR64d(I_XsR64d.Vextractpsq(l), {S=s;D=d})
                                          | I_XsR64d.Movmskpdq     -> XsR64d(I_XsR64d.Vmovmskpdq,     {S=s;D=d})
                                          | I_XsR64d.Movmskpsq     -> XsR64d(I_XsR64d.Vmovmskpsq,     {S=s;D=d})
                                          | I_XsR64d.Cvtsd2siq     -> XsR64d(I_XsR64d.Vcvtsd2siq,     {S=s;D=d}) 
                                          | I_XsR64d.Cvttsd2siq    -> XsR64d(I_XsR64d.Vcvttsd2siq,    {S=s;D=d}) 
                                          | I_XsR64d.Cvtss2siq     -> XsR64d(I_XsR64d.Vcvtss2siq,     {S=s;D=d}) 
                                          | I_XsR64d.Cvttss2siq    -> XsR64d(I_XsR64d.Vcvttss2siq,    {S=s;D=d}) 
                                          | I_XsR64d.Vpmovmskbq     | I_XsR64d.Vmovmskpdq  | I_XsR64d.Vmovmskpsq | I_XsR64d.Vmovd        | I_XsR64d.Vmovq -> i
                                          | I_XsR64d.Vextractpsq _  | I_XsR64d.Vpextrb _   | I_XsR64d.Vpextrq _  | I_XsR64d.Vpextrw _   -> i
                                          | I_XsR64d.Vcvtsd2siq     | I_XsR64d.Vcvttsd2siq | I_XsR64d.Vcvtss2siq | I_XsR64d.Vcvttss2siq -> i
(***)

let convert_R8sXds  i op s ds = match op with I_R8sXds.Pinsrb(l)  -> R8sXsXd(I_R8sXsXd.Vpinsrb(l),   {S1=s;S2=ds;D=ds})
let convert_R16sXds i op s ds = match op with I_R16sXds.Pinsrw(l) -> R16sXsXd(I_R16sXsXd.Vpinsrw(l), {S1=s;S2=ds;D=ds})
let convert_R32sXds i op s ds = match op with I_R32sXds.Cvtsi2sdl -> R32sXsXd(I_R32sXsXd.Vcvtsi2sdl, {S1=s;S2=ds;D=ds})
                                            | I_R32sXds.Cvtsi2ssl -> R32sXsXd(I_R32sXsXd.Vcvtsi2ssl, {S1=s;S2=ds;D=ds})
                                            | I_R32sXds.Pinsrd(l) -> R32sXsXd(I_R32sXsXd.Vpinsrd(l), {S1=s;S2=ds;D=ds})
let convert_R64sXds i op s ds = match op with I_R64sXds.Cvtsi2sdq -> R64sXsXd(I_R64sXsXd.Vcvtsi2sdq, {S1=s;S2=ds;D=ds})
                                            | I_R64sXds.Cvtsi2ssq -> R64sXsXd(I_R64sXsXd.Vcvtsi2ssq, {S1=s;S2=ds;D=ds})
                                            | I_R64sXds.Pinsrq(l) -> R64sXsXd(I_R64sXsXd.Vpinsrq(l), {S1=s;S2=ds;D=ds})

(***)

let convert_XsM8d   i op s d = match op with I_XsM8d.Pextrb(l)     -> XsM8d(I_XsM8d.Vpextrb(l),      {S=s;D=d})
                                           | I_XsM8d.Vpextrb _     -> i
let convert_XsM16d  i op s d = match op with I_XsM16d.Pextrw(l)    -> XsM16d(I_XsM16d.Vpextrw(l),    {S=s;D=d})
                                           | I_XsM16d.Vpextrw _    -> i
let convert_XsM32d  i op s d = match op with I_XsM32d.Extractps(l) -> XsM32d(I_XsM32d.Vextractps(l), {S=s;D=d})
                                           | I_XsM32d.Movd         -> XsM32d(I_XsM32d.Vmovd,         {S=s;D=d})
                                           | I_XsM32d.Movss        -> XsM32d(I_XsM32d.Vmovss,        {S=s;D=d})
                                           | I_XsM32d.Pextrd(l)    -> XsM32d(I_XsM32d.Vpextrd(l),    {S=s;D=d})
                                           | I_XsM32d.Vmovd         | I_XsM32d.Vmovss | I_XsM32d.Movntss | I_XsM32d.Vpextrd _ | I_XsM32d.Vextractps _ -> i
let convert_XsM64d  i op s d = match op with I_XsM64d.Movhpd       -> XsM64d(I_XsM64d.Vmovhpd,       {S=s;D=d})
                                           | I_XsM64d.Movhps       -> XsM64d(I_XsM64d.Vmovhps,       {S=s;D=d}) 
                                           | I_XsM64d.Movlpd       -> XsM64d(I_XsM64d.Vmovlpd,       {S=s;D=d})
                                           | I_XsM64d.Movlps       -> XsM64d(I_XsM64d.Vmovlps,       {S=s;D=d})
                                           | I_XsM64d.Movq         -> XsM64d(I_XsM64d.Vmovq,         {S=s;D=d})
                                           | I_XsM64d.Movsd        -> XsM64d(I_XsM64d.Vmovsd,        {S=s;D=d})
                                           | I_XsM64d.Pextrq(l)    -> XsM64d(I_XsM64d.Vpextrq(l),    {S=s;D=d})
                                           | I_XsM64d.Vmovhpd       | I_XsM64d.Vmovhps | I_XsM64d.Vmovlpd   | I_XsM64d.Vmovlps -> i
                                           | I_XsM64d.Vmovq         | I_XsM64d.Vmovsd  | I_XsM64d.Vpextrq _ | I_XsM64d.Movntsd -> i
let convert_XsM128d i op s d = match op with I_XsM128d.Movapd      -> XsM128d(I_XsM128d.Vmovapd,     {S=s;D=d})
                                           | I_XsM128d.Movaps      -> XsM128d(I_XsM128d.Vmovaps,     {S=s;D=d})
                                           | I_XsM128d.Movdqa      -> XsM128d(I_XsM128d.Vmovdqa,     {S=s;D=d})
                                           | I_XsM128d.Movdqu      -> XsM128d(I_XsM128d.Vmovdqu,     {S=s;D=d})
                                           | I_XsM128d.Movntdq     -> XsM128d(I_XsM128d.Vmovntdq,    {S=s;D=d})
                                           | I_XsM128d.Movntdqa    -> XsM128d(I_XsM128d.Vmovntdqa,   {S=s;D=d})
                                           | I_XsM128d.Movntpd     -> XsM128d(I_XsM128d.Vmovntpd,    {S=s;D=d})
                                           | I_XsM128d.Movntps     -> XsM128d(I_XsM128d.Vmovntps,    {S=s;D=d})
                                           | I_XsM128d.Movupd      -> XsM128d(I_XsM128d.Vmovupd,     {S=s;D=d})
                                           | I_XsM128d.Movups      -> XsM128d(I_XsM128d.Vmovups,     {S=s;D=d})
                                           | I_XsM128d.Vmovapd      | I_XsM128d.Vmovaps   | I_XsM128d.Vmovdqa  | I_XsM128d.Vmovdqu  | I_XsM128d.Vmovupd -> i
                                           | I_XsM128d.Vmovntdq     | I_XsM128d.Vmovntdqa | I_XsM128d.Vmovntpd | I_XsM128d.Vmovntps | I_XsM128d.Vmovups -> i

(***)

let convert(i:Instr) = match i with Xds(op,{DS=ds})           -> convert_Xds      i op    ds
                                  | XsXd(op,{S=s;D=d})        -> convert_XsXd     i op s  d
                                  | XsXds(op,{S=s;DS=ds})     -> convert_XsXds    i op s  ds
                                  | XsXs(op,{S1=s1;S2=s2})    -> convert_XsXs     i op s1 s2
                                  | M32sXs(op,{S1=s1;S2=s2})  -> convert_M32sXs   i op s1 s2
                                  | M64sXs(op,{S1=s1;S2=s2})  -> convert_M64sXs   i op s1 s2
                                  | M128sXs(op,{S1=s1;S2=s2}) -> convert_M128sXs  i op s1 s2
                                  | M16sXd(op,{S=s;D=d})      -> convert_M16sXd   i op s  d
                                  | M32sXd(op,{S=s;D=d})      -> convert_M32sXd   i op s  d
                                  | M64sXd(op,{S=s;D=d})      -> convert_M64sXd   i op s  d
                                  | M128sXd(op,{S=s;D=d})     -> convert_M128sXd  i op s  d
                                  | M32sXds(op,{S=s;DS=ds})   -> convert_M32sXds  i op s  ds
                                  | M64sXds(op,{S=s;DS=ds})   -> convert_M64sXds  i op s  ds
                                  | M128sXds(op,{S=s;DS=ds})  -> convert_M128sXds i op s  ds
                                  | R32sXd(op,{S=s;D=d})      -> convert_R32sXd   i op s  d
                                  | R64sXd(op,{S=s;D=d})      -> convert_R64sXd   i op s  d
                                  | XsR32d(op,{S=s;D=d})      -> convert_XsR32d   i op s  d
                                  | XsR64d(op,{S=s;D=d})      -> convert_XsR64d   i op s  d
                                  | R8sXds(op,{S=s;DS=ds})    -> convert_R8sXds   i op s  ds
                                  | R16sXds(op,{S=s;DS=ds})   -> convert_R16sXds  i op s  ds
                                  | R32sXds(op,{S=s;DS=ds})   -> convert_R32sXds  i op s  ds
                                  | R64sXds(op,{S=s;DS=ds})   -> convert_R64sXds  i op s  ds
                                  | XsM8d(op,{S=s;D=d})       -> convert_XsM8d    i op s  d
                                  | XsM16d(op,{S=s;D=d})      -> convert_XsM16d   i op s  d
                                  | XsM32d(op,{S=s;D=d})      -> convert_XsM32d   i op s  d
                                  | XsM64d(op,{S=s;D=d})      -> convert_XsM64d   i op s  d
                                  | XsM128d(op,{S=s;D=d})     -> convert_XsM128d  i op s  d
                                  | XsXsXd _ -> i
                                  | XsXsXsXd _ -> i
                                  | XsXsM128ds _ -> i
                                  | XsM128sXd _ -> i
                                  | XsM32sXsXd _ | XsM64sXsXd _ | XsM128sXsXd _ -> i
                                  | R8sXsXd _ | R16sXsXd _ | R32sXsXd _ | R64sXsXd _ -> i
                                  | M256sYsYd _ -> i
                                  | XsYsYd _ -> i
                                  | YsXd _ -> i
                                  | YsYd _ -> i
                                  | YsM128d _ | YsM256d _ -> i
                                  | YsR32d _ | YsR64d _ -> i
                                  | YsYsYd _ -> i
                                  | M32sYd _ | M64sYd _ | M128sYd _ | M256sYd _ -> i
                                  | M8sXsXd _ | M16sXsXd _ | M32sXsXd _ | M64sXsXd _ | M128sXsXd _ -> i
                                  | M32sXsXsXd _ | M64sXsXsXd _ | M128sXsXsXd _ -> i
                                  | I0 _ -> i
                                  | V0 _ -> i
                                  | C _ -> i
                                  | M8d _ | M16d _ | M32d _ | M64d _ -> i
                                  | R8d _ | R16d _ | R32d _ | R64d _ -> i
                                  | M8ds _ | M16ds _ | M32ds _ | M64ds _ | M128ds _ -> i
                                  | R8ds _ | R16ds _ | R32ds _ | R64ds _ -> i
                                  | M8s _ | M16s _ | M32s _ | M64s _ -> i
                                  | R8s _ | R16s _ | R32s _ | R64s _ -> i
                                  | NR64s _ | NM64s _ -> i
                                  | R8sR8s _ | R16sR16s _ | R32sR32s _ | R64sR64s _ -> i
                                  | R8sM8d _ | R16sM16d _ | R32sM32d _ | R64sM64d _ -> i
                                  | R8sR8d _ | R8sR16d _ | R8sR32d _ | R8sR64d _ | R16sR16d _ | R16sR32d _ | R16sR64d _ | R32sR32d _ | R32sR64d _ | R64sR64d _ -> i
                                  | M8sR8d _ | M8sR16d _ | M8sR32d _ | M8sR64d _ | M16sR16d _ | M16sR32d _ | M16sR64d _ | M32sR32d _ | M32sR64d _ | M64sR64d _ -> i
                                  | R8dsR8ds _ | R16dsR16ds _ | R32dsR32ds _ | R64dsR64ds _ -> i
                                  | M8sR8ds _ | M16sR16ds _ | M32sR32ds _ | M64sR64ds _ -> i
                                  | R8sM8ds _ | R16sM16ds _ | R32sM32ds _ | R64sM64ds _ -> i
                                  | R8sR16sM16ds _ | R8sR32sM32ds _ | R8sR64sM64ds _ -> i
                                  | R8sR16sR16ds _ | R8sR32sR32ds _ | R8sR64sR64ds _ -> i
                                  | M8dsR8ds _ | M16dsR16ds _ | M32dsR32ds _ | M64dsR64ds _ -> i
                                  | R8dsM8ds _ | R16dsM16ds _ | R32dsM32ds _ | R64dsM64ds _ -> i
                                  | R8sR8ds _ | R8sR16ds _ | R8sR32ds _ | R8sR64ds _ | R16sR16ds _ | R32sR32ds _ | R64sR64ds _ -> i
                                  | R8sM8s _ | R16sM16s _ | R32sM32s _ | R64sM64s _ -> i
                                  | M8sR8s _ | M16sR16s _ | M32sR32s _ | M64sR64s _ -> i

