#light (* -*- tuareg -*- *)

(* advantages of doing things THIS way, not with lex/yacc: ... *)
open Instr
open AsmAstIO

open InstrIO;;

(***)

type slice = 
  | Verbatim of ASTline list
  | Candidate of string * Instr list

(* TBD only take exactly one loop without output-vars! *)
let rec ends_soon = function
  | [] -> false
  | item :: items ->
      match item with
        | Instr(Instr.MRd(MRd.Popq,_)) -> ends_soon items
        | Instr(Instr.RMds _) -> ends_soon items
        | Instr(Instr.C0(C0.Ret,_)) -> true
        | Label _ -> ends_soon items
        | _ -> false

let rec slice2 label bag = function
  | [] -> None
  | item :: items ->
      match item with 
        | Instr i ->
            match i with
              | Instr.C0(C0.Ret,_) -> Some(i :: bag, items)
              | Instr.C((Call | Jmp),_) -> None
              | Instr.C((Jo | Jno | Jc | Jnc | Jz | Jnz | Ja | Jna | Js | Jns | Jp | Jnp | Jl | Jnl | Jg | Jng), l) -> 
                  if l <> label (* || not (ends_soon items) *) then None
                  else Some(i :: bag, items)
              | _ -> slice2 label (i :: bag) items
        | Tagged _ | Directive _ | Label _ -> None

let rec slice bag bags = function
  | [] -> List.rev (Verbatim (List.rev bag) :: bags)
  | item :: items ->
      match item with
        | Tagged _ | Instr _ | Directive _ -> slice (item :: bag) bags items
        | Label l ->
            match slice2 {Sk.S=l} [] items with
              | Some(instrs,post) (* when List.exists (fun (i:Instr) -> i.IsVecInstr) instrs *) -> 
                  slice [] (Candidate(l,List.rev instrs) :: Verbatim (List.rev bag) :: bags) post
              | _ -> slice (item::bag) bags items

(***)

let eliminate_useless_labels (lines:ASTline list) =
  let collect_id (all:Set<string>) = function
    | Tagged(_,args) | Directive(_,args) -> List.fold_left (fun set (i:ASTitem) -> set + i.Ids) all args
    | Instr _ -> failwith "eliminate_useless_labels: collect_id (Instr _)" (* (C(_,id)) -> Set.add id all *)
    | _ -> all
  let all_id_uses = List.fold_left collect_id Set.empty lines
  let useful_label = function
    | Label x -> all_id_uses.Contains(x) || (Set.exists (fun (s:string) -> s.Contains(x)) all_id_uses)
    | _ -> true
  let (good,bad) = List.partition useful_label lines
(*
  Printf.printfn "eliminate_useless_labels: %d -> (good %d, bad %d)" lines.Length good.Length bad.Length
  Printf.printfn "|all_id_uses| = %d" all_id_uses.Count
*)
  good;;

(***)

let nextReader_vreg (reg:X) (aInstrs:Instr list) =
  let rec loop pre = function
    | [] -> None
    | (instr : Instr)::xs ->
        let (i_d,i_ds,i_s) = (InstrIO.Get_X_d_in_Instr instr, InstrIO.Get_X_ds_in_Instr instr, InstrIO.Get_X_s_in_Instr instr) (* instr.GetXs () *)
        if Set.mem reg i_ds || Set.mem reg i_s then Some(pre,instr,xs)
        else if Set.mem reg i_d then None
        else loop (instr::pre) xs
  in loop [] aInstrs


let nextReader_memop (m:M) (aInstrs:Instr list) =
  let rec loop pre = function
    | [] -> None
    | (instr : Instr)::xs ->
        let (i_d,i_ds,i_s) = (InstrIO.Get_M_d_in_Instr instr, InstrIO.Get_M_ds_in_Instr instr, InstrIO.Get_M_s_in_Instr instr) (* instr.GetDstMs (), instr.GetSrcMs ()) *)
        if      Set.mem m i_s || Set.mem m i_ds  then Some(pre,instr,xs)
        else if Set.mem m i_d  then None
        else loop (instr::pre) xs
  in loop [] aInstrs

let writer_of_vreg (reg:X) (instrs:Instr list) =
  let rec loop bag = function
    | [] -> None
    | (i:Instr)::xs ->
        let (i_d,i_ds,i_s) = (InstrIO.Get_X_d_in_Instr i, InstrIO.Get_X_ds_in_Instr i, InstrIO.Get_X_s_in_Instr i) (* i.GetXs () *)
        if Set.mem reg i_d then Some(bag,i,xs) else loop (i::bag) xs
  loop [] instrs

let aInstrs_doNotRead_vreg (aInstrs:Instr list) (v:X) = nextReader_vreg v aInstrs = None

let instrs_doNotUse_vreg (xs:Instr list) (v:X) =
  not (List.exists (fun (i:Instr) -> Set.mem v (InstrIO.Get_X_in_Instr i)) xs)
(* let (i_d,i_ds,i_s) = i.GetXs () in List.mem v i_d || List.mem v i_ds || List.mem v i_s) xs) *)

let instrs_doNotUse_memop (xs:Instr list) (m:M) =
  not (List.exists (fun (i:Instr) -> Set.mem m (InstrIO.Get_M_in_Instr i)) xs)
(* let (i_d,i_s) = (i.GetSrcMs (),i.GetDstMs ()) in List.mem m i_d || List.mem m i_s) xs) *)


(* not conservative, but aggressive! *)
let eliminate_useless_veccopies (instrs:Instr list) =
  let rec loop (bag:Instr list) (gone:Instr list) = function
    | [] -> (List.rev bag,List.rev gone)
    | (XdMs(op,{D=d;S=s}) as x)::xs ->
        match nextReader_vreg d xs with
          | None -> failwith "eliminate_useless_veccopies"
          | Some(pre,(MdXs(op,{D=d';S=s'}) as x'),post) 
            when d' = s && s' = d && (not (List.exists (fun (i:Instr) -> Set.mem s (InstrIO.Get_M_in_Instr i)) pre)) && aInstrs_doNotRead_vreg post d ->
(*            when d' = s && s' = d && (not (List.exists (fun (i:Instr) -> List.mem s (i.GetSrcMs ()) || List.mem s (i.GetDstMs ())) pre)) && aInstrs_doNotRead_vreg post d -> *)

              loop bag (x'::x::gone) ((List.rev pre) @ post)
          | _ -> loop (x::bag) gone xs
    | x::xs -> loop (x::bag) gone xs
  let (good,bad) = loop [] [] instrs
(*  Printf.printf "eliminate_useless_veccopies: %d -> (%d + %d useless)\n" instrs.Length good.Length bad.Length *)
  good;;

let eliminate_useless_vecloads (instrs:Instr list) =
  let rec loop (bag:Instr list) (gone:Instr list) = function
    | [] -> (List.rev bag,List.rev gone)
    | (XdMs(_,{D=d;S=_}) as x)::xs when aInstrs_doNotRead_vreg xs d -> loop bag (x::gone) xs
    | x::xs -> loop (x::bag) gone xs
  let (bag,gone) = loop [] [] instrs
(*  Printf.printfn "eliminate_useless_vecloads: %d -> (%d OK + %d useless)" instrs.Length bag.Length gone.Length *)
  bag;;

(***)
(* from regrealloc1.ml *)

let veccopy_to_rename (instrs:Instr list) = 
  let rec loop (bag:Instr list) (gone:Instr list) = function
    | [] -> (List.rev bag, List.rev gone)
    (* d <- s
       content of s unused            => swap d,s
       s <- t
    *)
    | (XdXMs(XdMXs.Movaps,{D=d;S=s}) as x) :: xs when aInstrs_doNotRead_vreg xs s ->
        let map_v (v:X) = if v = d then s else if v = s then d else v
        loop bag (x::gone) (List.map (fun (i:Instr) -> InstrIO.Map_X_in_Instr map_v i) xs)
(*        loop bag (x::gone) (List.map (fun (i:Instr) -> i.MapXs map_v map_v map_v) xs) *)
    | x::xs -> loop (x::bag) gone xs
  let (good,bad) = loop [] [] instrs
(*  Printf.printfn "veccopy_to_rename: %d -> (%d OK, %d eliminated)" instrs.Length good.Length bad.Length *)
  good;;

let vec_opt_memops (instrs:Instr list) =
  let rec loop (bag:Instr list) (gone:Instr list) = function
    | [] -> (List.rev bag, List.rev gone)
    | (MdXs(MdXs.Movaps,{D=m;S=s}) as x) :: xs 
      when m.IsStack &&
           (match nextReader_memop m xs with
              | Some(pre, XdsMXs(XdsMXs.Addps,{Ds=d;S=_}), post) ->
                  d <> s &&
                  instrs_doNotUse_vreg pre d &&
                  (nextReader_memop m post = None) (* may re-use *)
              | _ -> false) 
      -> match nextReader_memop m xs with
           | Some(pre,(XdsMXs(op,{Ds=d;S=_}) as y),post) ->
               let bag = (XdsXMs(op,{Ds=d;S=s}))::bag 
               loop bag (y::x::gone) ((List.rev pre) @ post)
           | _ -> failwith "vec_opt_memops"
      (* next user of `m' is last user of `m'
         user = I_Xds_Ms(op,ds,m) suchthat 
           1 ds is not used in between
      *)
    | x::xs -> loop (x::bag) gone xs
  let (good,bad) = loop [] [] instrs
(*  Printf.printfn "vec_opt_memops: %d -> (%d OK, %A eliminated)" instrs.Length good.Length bad *)
  good;;

let move_some_spills_up (instrs:Instr list) =
  let n = ref 0
  let rec loop (bag:Instr list) = function
    | [] -> List.rev bag
    | (MdXs(MdXs.Movaps,{D=m;S=s}) as x) :: xs
      when m.IsStack (* ??? *) &&
           (match writer_of_vreg s bag with
              | Some(pre,writer,post) -> pre.Length > 20 && instrs_doNotUse_memop pre m
              | None -> false)
      -> match writer_of_vreg s bag with
           | Some(pre,writer,post) ->
               incr n
               let bag = (List.rev pre) @ [x; writer] @ post
               loop bag xs
           | _ -> failwith "move_some_spills_up"
    | x::xs -> loop (x::bag) xs
  let good = loop [] instrs
(*  Printf.printfn "move_some_spills_up: %d -> (%d OK, moved %d times)" instrs.Length good.Length !n *)
  good;;

let move_some_reloads_down (instrs:Instr list) =
  let n = ref 0
  let rec loop (bag:Instr list) = function
    | [] -> List.rev bag
    | (XdMXs(XdMXs.Movaps,{D=d;S=m}) as x) :: xs
      when m.IsStack (* ??? *) &&
           (match nextReader_vreg d xs with
              | Some(pre,_,post) -> pre.Length > 20 && instrs_doNotUse_memop pre m
              | None -> failwith (Printf.sprintf "move_some_reloads_down (n=%d) %A %A" !n x xs))
      -> match nextReader_vreg d xs with
           | Some(pre,reader,post) ->
               incr n
               loop bag ((List.rev pre) @ (x::reader::post))
           | _ -> failwith "move_some_reloads_down"
    | x::xs -> loop (x::bag) xs
  let good = loop [] instrs
(*  Printf.printfn "move_some_reloads_down: %d -> (%d OK, moved %d times)" instrs.Length good.Length !n *)
  good;;

let swap2 r1 r2 r = if r=r1 then r2 else if r=r2 then r1 else r

let rot3 r1 r2 r3 r = if r=r1 then r2 else if r=r2 then r3 else if r=r3 then r1 else r

let veccopy_to_rename2 (instrs:Instr list) = 
  let rules_used = new System.Collections.Generic.Dictionary<string,int> ()
  let note (s:string) = 
    if not (rules_used.ContainsKey s) then rules_used.Add(s,1)
    else 
      ignore (rules_used.Remove s)
      rules_used.Add(s, rules_used.[s]+1)
  let rec loop (bag:Instr list) = function
    | [] -> List.rev bag
    | (XdMXs(XdMXs.Movaps,{D=d;S=s_ea}) as x) :: xs
      when (match nextReader_vreg d xs with 
              | Some(pre',XdsXMs(op,{Ds=ds;S=s}),post) -> d=ds && ds<>s && op.IsCommutative && (aInstrs_doNotRead_vreg post s)
              | _ -> false) ->
        note "rule1"
        match nextReader_vreg d xs with
          | Some(pre',XdsXMs(op,{Ds=ds;S=s}),post) ->
              let aInstr = XdsMXs(op,{Ds=s;S=s_ea})
              loop bag ((List.rev pre') @ (aInstr::(List.map (fun (i:Instr) -> InstrIO.Map_X_in_Instr (swap2 ds s) i) post)))
(*              loop bag ((List.rev pre') @ (aInstr::(List.map (fun (i:Instr) -> i.MapXs (swap2 ds s) (swap2 ds s) (swap2 ds s)) post))) *)
          | _ -> failwith "renameRegisters1 2B"

    | XdXMs(XdMXs.Movaps,{D=d;S=s}) :: aInstrs
      when (* instr_is_inlinableVecCopy instr && *)
           (match nextReader_vreg s aInstrs with
              | Some(pre',XdsXMs(op',{Ds=sd;S=k}),post) ->
                  sd<>k && sd=s && sd<>d && k<>s &&
                  op'.IsCommutative &&
                  (nextReader_vreg k ((List.rev pre') @ post) = None) &&
                  (writer_of_vreg k pre' = None)
              | _ -> false) ->
        note "rule 2c"
        let aInstrs = match nextReader_vreg s aInstrs with
                        | Some(pre',XdsXs(op',{Ds=sd;S=k}),post) ->
                            let aInstr = XdsXs(op',{Ds=k;S=s}) in 
                            aInstr::(List.map (fun (i:Instr) -> InstrIO.Map_X_in_Instr (rot3 d s k) i) ((List.rev pre') @ post))
(*                            aInstr::(List.map (fun (i:Instr) -> i.MapXs (rot3 d s k) (rot3 d s k) (rot3 d s k)) ((List.rev pre') @ post)) *)
                        | _ -> failwith "renameRegisters1 2C"
        loop bag aInstrs

    | x::xs -> loop (x::bag) xs
  let good = loop [] instrs
(*  Printf.printfn "veccopy_to_rename2: %d -> (%d OK, %A)" instrs.Length good.Length rules_used *)
  good;;

let do_const_access (const_base:R) (instrs:Instr list) =
  let occur = new System.Collections.Generic.Dictionary<string,int> ()
  let map_ea (ea:M) = 
    match ea with
      | K(_,k) ->
          if not (occur.ContainsKey(k)) then occur.Add(k,1)
          else let n = occur.[k] in begin ignore (occur.Remove(k)); occur.Add(k,n+1) end
      | _ -> ()
    ea
  let map_instr (i:Instr) = InstrIO.Map_M_in_Instr map_ea i
(*  let map_instr (i:Instr) = i.MapMs map_ea *)
  ignore (List.map map_instr instrs) (* use map, not iter, because this guarantees we are seeing each constant as often as we should *)
 
(*  Printf.printfn "const_xs:"  *)
  let vks = ref []
  for o in occur do vks := (o.Value,o.Key) :: !vks
  let vks_a = Array.of_list !vks
  Array.sort (fun (k1,_) (k2,_) -> compare k2 k1) vks_a
(*  Printf.printfn "\t %A" vks_a *)
  let o_map = new System.Collections.Generic.Dictionary<string,M> ()
  for i = 0 to min (vks_a.Length - 1) 15 do
    let i' = i*16
    let i' = if i>=128 then i-128 else i
    o_map.Add(snd (vks_a.[i]),RID(Ofs.Int (int64 i'),const_base))
  let map_ea (ea:M) = 
    match ea with
      | K(_,k) when o_map.ContainsKey(k) -> o_map.[k]
      | _ -> ea
  let map_instr (i:Instr) = InstrIO.Map_M_in_Instr map_ea i
(*  let map_instr (i:Instr) = i.MapMs map_ea *)

  let instrs = List.map map_instr instrs (* use map, not iter, because this guarantees we are seeing each constant as often as we should *)
  (o_map, instrs)
    
let do_paging (index:R) (index_size:int64) (instrs:Instr list) =
  let n = ref 0
  let map_ea (ea:M) = 
    match ea with
      | RID(Ofs.Int i, r_base) ->
          match List.tryfind (fun k -> let v = i - k*index_size in v >= -128L && v < 128L) [0L;1L;2L;4L;8L] with
            | None | Some 0L -> ea
            | Some k -> 
                incr n
                let ea' = RISID(Ofs.Int (i - k*index_size),r_base,index,k)
(*                Printf.printfn "M: %A -> %A" ea ea' *)
                ea'
      | _ -> ea
  let map_instr (i:Instr) = InstrIO.Map_M_in_Instr map_ea i
(*  let map_instr (i:Instr) = i.MapMs map_ea *)
  let instrs' = List.map map_instr instrs
(*  Printf.printfn "do_paging: did it %d times" !n *)
  instrs'
;;


(*
let veccopy_to_rename2 (instrs:Instr list) = 
  let rec loop (bag:Instr list) (gone:Instr list) = function
    | [] -> (List.rev bag, List.rev gone)

    | (I_Xd_Xs(XdXs.Movaps,d,s) as x) :: xs  (* 2B *)
      when (match nextReader_vreg d xs with 
              Some(pre',I_Xds_Xs(op',sd,k),post) -> sd<>k && sd=d && sd<>s && k<>s && op'.IsCommutative && ((nextReader_vreg k ((List.rev pre') @ post)) = None) && (not (List.exists (fun (i:Instr) -> let (i_d,_,_) = i.Xs in List.mem k i_d) pre'))
              | _ -> false) ->
        match nextReader_vreg d xs with
          | Some(pre',I_Xds_Xs(op',sd,k),post) ->
              let aInstr = I_Xds_Xs(op',k,s)
              loop bag (x::gone) (List.map (fun (i:Instr) -> i.MapXs (swap2 k d) (swap2 k d) (swap2 k d)) post)
          | _ -> failwith "renameRegisters1 2B"
    | x::xs -> loop (x::bag) gone xs
  let (good,bad) = loop [] [] instrs
  Printf.printfn "veccopy_to_rename2: %d -> (%d OK, %d eliminated)" instrs.Length good.Length bad.Length
  good;;

*)

(***)

(* OLD, vector only *)

(*
let vec_liveness (instrs0:Instr list) =
  let hlp (i:Instr) (augs,v_dead) = (* (aug,deadRs:Set<R>,deadXs:Set<X>) = *)
    let (v_d,v_ds,v_s) = i.XSets 
    let v_in   = v_ds + v_s
    let v_dead = v_d + v_dead 
    let v_kill = Set.intersect v_in v_dead
    let v_dead = v_dead - v_in
    let aug = (v_in,v_kill,v_d,i)
    (aug::augs, v_dead)
  let augmented = List.fold_right hlp instrs0 ([],Set.of_list X.All)
  augmented
;;
*)

(* written inside of the loop, dead outside *)
let liveness (i_written:Set<R>) (instrs0:Instr list) =
  let hlp (i:Instr) (augs,i_dead,v_dead) = (* (aug,deadRs:Set<R>,deadXs:Set<X>) = *)
    let i_d    = InstrIO.Get_R_d_in_Instr i
    let i_ds   = InstrIO.Get_R_ds_in_Instr i
    let i_s    = InstrIO.Get_R_s_in_Instr i
(*    let (i_d,i_ds,i_s) = i.GetRSets () *)
    let i_dead = i_dead + i_d
    let i_kill = Set.intersect i_s i_dead
    let i_dead = i_dead - i_s
    let i_in   = i_s + i_ds

    let v_d    = InstrIO.Get_X_d_in_Instr i
    let v_ds   = InstrIO.Get_X_ds_in_Instr i
    let v_s    = InstrIO.Get_X_s_in_Instr i

(*    let (v_d,v_ds,v_s) = i.GetXSets () *)
    let v_dead = v_dead + v_d
    let v_kill = Set.intersect v_s v_dead
    let v_dead = v_dead - v_s
    let v_in   = v_s + v_ds

(* out = in + gen - kill  *)

    let aug = (i_in,i_kill,i_d,v_in,v_kill,v_d,i)
    (aug::augs, i_dead, v_dead)
  let augmented = List.fold_right hlp instrs0 ([],i_written,Set.of_list X.All)
  augmented
;;


let prep_of_liveness ((i_in,i_kill,i_gen,v_in,v_kill,v_gen,instr):(Set<R> * Set<R> * Set<R> * Set<X> * Set<X> * Set<X> * Instr)) (totalsize,genes) =
  let n_i_ds = i_in.Count
  let n_i = i_in.Count + i_gen.Count

  let (n_v_ds,n_v_d) = (v_in.Count, v_gen.Count)
  let n_v = n_v_ds + n_v_d

  let n = n_i + n_v

  let i_in   = Set.to_list i_in
  let i_kill = List.map (fun (i:R) -> i.Int) (Set.to_list i_kill)
  let i_gen  = Set.to_list i_gen

  let v_in   = Set.to_list v_in
  let v_kill = List.map (fun (v:X) -> v.Int) (Set.to_list v_kill)
  let v_gen  = Set.to_list v_gen

  if n = 0 then failwith "0" (* integer or weird vector instruction -> hoist out! *)

  let sizes = Array.create (1<<<n) (-1)

  let lookupV ix ofs where what = if ix &&& (1 <<< (ofs + List.find_index ((=)what) where)) = 0 then X.Lo else X.Hi
  let lookupI ix ofs where what = if ix &&& (1 <<< (ofs + List.find_index ((=)what) where)) = 0 then R.Lo else R.Hi


  for ix = 0 to (1 <<< n) - 1 do
(*    Printf.printfn "%A in %A gen %A" instr i_in i_gen *)

(* SPURIOUS! *)
    let instr = InstrIO.Map_R_d_in_Instr (lookupI ix n_i_ds i_gen) instr
    let instr = InstrIO.Map_R_ds_in_Instr (lookupI ix 0 i_in) instr
    let instr = InstrIO.Map_R_s_in_Instr  (lookupI ix 0 i_in) instr

    let instr = InstrIO.Map_X_d_in_Instr (lookupV ix (n_i+n_v_ds) v_gen) instr
    let instr = InstrIO.Map_X_ds_in_Instr (lookupV ix n_i v_in) instr
    let instr = InstrIO.Map_X_s_in_Instr  (lookupV ix n_i v_in) instr
(* XXX ??? TBD CHK *)
(*    let instr1 = instr.MapRs (lookupI ix      n_i_ds  i_gen) (lookupI ix   0 i_in) (lookupI ix   0 i_in)
    let instr1 = instr.MapXs (lookupV ix (n_i+n_v_ds) v_gen) (lookupV ix n_i v_in) (lookupV ix n_i v_in)
*)

(*    Printf.printf "%A -> %A (ix=%d) size = %d\n" instr instr1 ix instr1.CodeSize *)
    sizes.[ix] <- CodeSize_of_Instr instr

  let min_size = Array.min sizes
  and max_size = Array.max sizes
  if max_size > min_size + 1 then failwith "size!"
  let costs = 
    if min_size = max_size then 0 
    else
      let mutable costs = 0
      for ix = 0 to (1<<<n)-1 do
        if sizes.[ix] = max_size then costs <- costs ||| (1 <<< ix)
      costs

  let v_in  = List.map (fun (v:X) -> v.Int) v_in
  let v_gen = List.map (fun (v:X) -> v.Int) v_gen

  let i_in  = List.map (fun (i:R) -> i.Int) i_in
  let i_gen = List.map (fun (i:R) -> i.Int) i_gen

  (totalsize+min_size, ((i_in,i_kill,i_gen,v_in,v_kill,v_gen,costs)::genes))
;;

open Genes01;;


let all_instrs_hack : Instr.Instr array ref = ref [| |]

type prep = (int list * int list * int list * int list * int list * int list * int);;

(* evaluates genes according to prepared data `prep' *)


let eval_genes (int_inputs:Set<R>) (n_i_freeLo:int) (n_i_freeHi:int) (vec_uninitialized:Set<X>) (g:Genes01) (p:prep array) =
  let mutable pos = 0 
 
  let mutable mappingI = 0
  for int_input in int_inputs do
    let i = int_input.Int
    if i >= 8 then mappingI <- mappingI ||| (1 <<< i)

  let mutable nIfreeLO = n_i_freeLo
  let mutable nIfreeHI = n_i_freeHi

  let mutable mapping = 0xff00
  let mutable nFreeLo = 8 - (List.filter (fun (v:X) -> v.Int < 8) (Set.to_list vec_uninitialized)).Length
  let mutable nFreeHi = 8 - (List.filter (fun (v:X) -> v.Int >= 8) (Set.to_list vec_uninitialized)).Length
  let mutable totalCosts = 0
  let mutable nRepair = 0
  for i = 0 to p.Length - 1 do
    let (i_in,i_kill,i_gen,v_in,v_kill,v_gen,costs) = p.[i]
    let mutable ix = 0

(* INT *)
    for j = 0 to i_in.Length - 1   do if mappingI &&& (1 <<< i_in.[j]) <> 0 then ix <- ix ||| (1 <<< j)
    for j = 0 to i_kill.Length - 1 do if mappingI &&& (1 <<< i_kill.[j]) = 0 then nIfreeLO <- nIfreeLO + 1 else nIfreeHI <- nIfreeHI + 1
    for j = 0 to i_gen.Length - 1 do
      let d = i_gen.[j]
      if g.[pos] then           (* upper half *)
        if      nIfreeHI > 0 then begin nIfreeHI <- nIfreeHI - 1; mappingI <- mappingI ||| (1 <<< d); ix <- ix ||| (1 <<< (j + i_in.Length)) end
        else if nIfreeLO > 0 then begin nIfreeLO <- nIfreeLO - 1; nRepair <- nRepair + 1; g.MutateClear1 pos; mappingI <- mappingI &&& (~~~ (1 <<< d)) end
        else failwith "eval_genes: Int1"
      else
        if      nIfreeLO > 0 then begin nIfreeLO <- nIfreeLO - 1; mappingI <- mappingI &&& (~~~ (1 <<< d)) end
        else if nIfreeHI > 0 then begin nIfreeHI <- nIfreeHI - 1; nRepair <- nRepair + 1; g.MutateSet1 pos; mappingI <- mappingI ||| (1 <<< d); ix <- ix ||| (1 <<< (j + i_in.Length)) end
        else failwith (Printf.sprintf "eval_genes: Int2 (instr #%d, nIfreelo %d, nIFreeHI %d)" i nIfreeLO nIfreeHI)
      pos <- pos + 1

(* VEC *)
    for j = 0 to v_in.Length - 1   do if mapping &&& (1 <<< v_in.[j]) <> 0 then ix <- ix ||| (1 <<< (j + i_in.Length + i_gen.Length))
    for j = 0 to v_kill.Length - 1 do if mapping &&& (1 <<< v_kill.[j]) = 0 then nFreeLo <- nFreeLo + 1 else nFreeHi <- nFreeHi + 1
    for j = 0 to v_gen.Length - 1 do
      let d = v_gen.[j]
      if g.[pos] then           (* upper half *)
        if nFreeHi > 0 then begin nFreeHi <- nFreeHi - 1; mapping <- mapping ||| (1 <<< d); ix <- ix ||| (1 <<< (j + i_in.Length + i_gen.Length + v_in.Length)) end
        else if nFreeLo > 0 then begin nFreeLo <- nFreeLo - 1; nRepair <- nRepair + 1; g.MutateClear1 pos; mapping <- mapping &&& (~~~ (1 <<< d)) end
        else failwith (Printf.sprintf "eval_genes: Vec1 (instr #%d, nIfreelo %d, nIFreeHI %d)" i nFreeLo nFreeHi)
      else
        if      nFreeLo > 0 then begin nFreeLo <- nFreeLo - 1; mapping <- mapping &&& (~~~ (1 <<< d)) end
        else if nFreeHi > 0 then begin nFreeHi <- nFreeHi - 1; nRepair <- nRepair + 1; g.MutateSet1 pos; mapping <- mapping ||| (1 <<< d); ix <- ix ||| (1 <<< (j + i_in.Length + i_gen.Length + v_in.Length)) end
        else failwith (Printf.sprintf "eval_genes: Vec2 (instr #%d %A, nIfreelo %d, nIFreeHI %d)" i (!all_instrs_hack).[i] nFreeLo nFreeHi)
      pos <- pos + 1

    totalCosts <- totalCosts + ((costs >>> ix) &&& 1)
(*  Printf.printf "%A\n" (totalCosts,nRepair,nFreeLo,nFreeHi,pos) *)
  (totalCosts,nRepair,nIfreeLO,nIfreeHI,nFreeLo,nFreeHi,pos);;

(***)

let express_genes (int_inputs:Set<R>) (int_written:Set<R>) (g:Genes01) (p:prep array) (arr:Instr array) =
  let mutable pos = 0 

  let mappingI = Array.init 16 (fun i -> i)
  for i in int_inputs do 
    let x = i.Int
    mappingI.[x] <- x

  let i_freeLo = ref (List.filter (fun i -> i < 8)  (List.map (fun  (i:R) -> i.Int) (Set.to_list int_written)))
  let i_freeHi = ref (List.filter (fun i -> i >= 8) (List.map (fun  (i:R) -> i.Int) (Set.to_list int_written)))

  let mapping = Array.init 16 (fun i -> i)
  let freeLo = ref [0;1;2;3;4;5;6;7]
  let freeHi = ref [8;9;10;11;12;13;14;15]
  let arrD = Array.copy arr
  for i = 0 to p.Length - 1 do
    let (i_in,i_kill,i_gen,v_in,v_kill,v_gen,_) = p.[i]


    for j = 0 to i_kill.Length - 1 do
      let r = mappingI.[i_kill.[j]]
      if r < 8 then i_freeLo := r :: !i_freeLo else i_freeHi := r :: !i_freeHi

    for j = 0 to i_gen.Length - 1 do
      let d = i_gen.[j]
      if g.[pos] then if !i_freeHi <> [] then begin mappingI.[d] <- List.hd !i_freeHi; i_freeHi := List.tl !i_freeHi end else failwith "express_genes (1i)"
      else            if !i_freeLo <> [] then begin mappingI.[d] <- List.hd !i_freeLo; i_freeLo := List.tl !i_freeLo end else failwith "express_genes (2i)"
      pos <- pos + 1


    let i_ds_map : (int * int) list = List.map (fun (i:int) -> (i, mappingI.[i])) i_in
    let i_d_map  : (int * int) list = List.map (fun (i:int) -> (i, mappingI.[i])) i_gen
    let map_i_d  (i:R) = let i = i.Int in Instr.R (int64 (snd (List.find (fun (x,_) -> x=i) i_d_map)))
    and map_i_ds (i:R) = let i = i.Int in Instr.R (int64 (snd (List.find (fun (x,_) -> x=i) i_ds_map)))

   
    for j = 0 to v_kill.Length - 1 do 
      let r = mapping.[v_kill.[j]] 
      if r < 8 then freeLo := r::!freeLo else freeHi := r::!freeHi
   


    for j = 0 to v_gen.Length - 1 do
      let d = v_gen.[j]
      if g.[pos] then if !freeHi <> [] then begin mapping.[d] <- List.hd !freeHi; freeHi := List.tl !freeHi end else failwith "express_genes (1v)"
      else            if !freeLo <> [] then begin mapping.[d] <- List.hd !freeLo; freeLo := List.tl !freeLo end else failwith "express_genes (2v)"
      pos <- pos + 1
 
    let v_ds_map : (int * int) list = List.map (fun (v:int) -> (v, mapping.[v])) v_in
    let v_d_map  : (int * int) list = List.map (fun (v:int) -> (v, mapping.[v])) v_gen
    let map_v_d  (v:X) = let v = v.Int in Instr.X (int64 (snd (List.find (fun (x,_) -> x=v) v_d_map)))
    and map_v_ds (v:X) = let v = v.Int in Instr.X (int64 (snd (List.find (fun (x,_) -> x=v) v_ds_map)))

    let instr = arr.[i]
    let instr = InstrIO.Map_R_d_in_Instr map_i_d instr
    let instr = InstrIO.Map_R_ds_in_Instr map_i_ds instr
    let instr = InstrIO.Map_R_s_in_Instr map_i_ds instr

    let instr = InstrIO.Map_X_d_in_Instr map_v_d instr
    let instr = InstrIO.Map_X_ds_in_Instr map_v_ds instr
    let instr = InstrIO.Map_X_s_in_Instr map_v_ds instr

    arrD.[i] <- instr 

(*    arrD.[i] <- (arr.[i].MapRs map_i_d map_i_ds map_i_ds).MapXs map_v_d map_v_ds map_v_ds *)

  arrD;;

(*
TBD assert that everything is THERE (nothing is lost by slicing etc)
    assert that order is correct
*)

let get_integer_inputs (instrs:Instr list) =
  let collect ((inputs,written):Set<R> * Set<R>) (i:Instr) =
    let i_gen = InstrIO.Get_R_d_in_Instr i
    let i_ds  = InstrIO.Get_R_ds_in_Instr i
    let i_s   = InstrIO.Get_R_s_in_Instr i 
(*    let (i_gen,i_ds,i_s) = i.GetRSets () *)
    let i_in = i_ds + i_s
    let inputs = inputs + (i_in - written)
    let written = written + i_gen 
    (inputs,written)
  let (inp,wr) = List.fold_left collect (Set.empty,Set.empty) instrs
  (inp,wr-inp)

open Genes01;;
open Ga;

let totalpopsize = 5000

let testGA (comm:MPI.TopologicalCommunicator) (text:string) =
  let popsize = totalpopsize (*  / comm.Size *)
  let rng = new System.Random (comm.Rank)


  let lexbuf = Lexing.from_string text
  let lines = try 
                AsmParser.ASTlines AsmLexer.token lexbuf
              with e -> let pos = lexbuf.EndPos in failwithf "Error near char %d\n" pos.Column

  let lines = eliminate_useless_labels lines

  let lines = List.map InstrIO.TryFindInstr_in_ASTline lines
(*  let lines = List.map (fun (l:ASTline) -> l.TryFindInstr) lines *)

  if comm.Rank = 0 then
    Printf.printfn "--- lines start" 
    for l in lines do
      Printf.printfn "%A" l
    Printf.printfn "--- lines end"

  let slices = slice [] [] lines

  if comm.Rank = 0 then Printf.printfn "# slices = %d" (List.length slices)

  let slices = List.filter (function Verbatim items -> items <> [] | _ -> true) slices
(*    let slices = List.map (fun (b,items:Line list) -> (b,items.Length,items)) slices
*)


  let outslices = ref []

  for slice in slices do
    match slice with
      | Verbatim xs -> 
          outslices := (Verbatim xs) :: !outslices

(*
          if comm.Rank = 0 then 
            for x in xs do
              Printf.printfn "\t%A" x
*)

(*
            Printf.printfn "Verbatim %A" x
*)
      | Candidate(lbl,instrs) ->
          let (int_inputs,int_written) = get_integer_inputs instrs
          if comm.Rank = 0 then
            Printf.printfn "(%d) int_inputs  = %A" int_inputs.Count int_inputs
            Printf.printfn "(%d) int_written = %A" int_written.Count int_written

          let n_i_freeLo = List.length (List.filter (fun (i:R) -> i.Int <  8) (Set.to_list int_written))
          let n_i_freeHi = List.length (List.filter (fun (i:R) -> i.Int >= 8) (Set.to_list int_written))

          let all_size = List.sum (List.map CodeSize_of_Instr instrs)

(***



          let (other,instrs) = List.partition (fun (i:Instr) -> Set.is_empty (InstrIO.Get_R_in_Instr i) && Set.is_empty (InstrIO.Get_X_in_Instr i)) instrs
(*          let (instrs,other) = List.partition (fun (i:Instr) -> match (i.GetRs (),i.GetXs ()) with | ([],[],[]),([],[],[]) -> false | _ -> true) instrs *)
          if comm.Rank = 0 then
            Printf.printf "others = %A\n" other

(*
          let instrs = List.zip [1..(List.length instrs)] instrs 
          let (vec_instrsx,rest) = List.partition (fun ((_,i):(int * Instr)) -> match i.Xs with | ([],[],[]) -> false | _ -> true) instrs
          let vec_instr_ixs = List.map fst vec_instrsx
          let vec_instrs = List.map snd vec_instrsx
*)
          let vec_instrs = instrs
          if comm.Rank = 0 then
            Printf.printfn "%s:" lbl
            Printf.printfn "|instrs| = %d\n" instrs.Length

(*
          let vec_instrs = eliminate_useless_vecloads vec_instrs
          let vec_instrs = eliminate_useless_veccopies vec_instrs
          let vec_instrs = veccopy_to_rename vec_instrs
          let vec_instrs = veccopy_to_rename2 vec_instrs
          let vec_instrs = move_some_spills_up    vec_instrs
          let vec_instrs = move_some_reloads_down vec_instrs

          let () = Lifespans.rename_stackcells vec_instrs

          if comm.Rank = 0 then
            Printf.printfn "%s:" lbl
            Printf.printfn "|instrs| = %d\n" instrs.Length

          let vec_instrs = vec_opt_memops vec_instrs
*)

          let xtra_index_reg = List.find (fun (r:R) -> not (Set.mem r (int_inputs + int_written))) R.All
          Printf.printfn "Picking XtraREG %A" xtra_index_reg
(*
          let vec_instrs = do_paging xtra_index_reg 256L vec_instrs
*)

          let xtra_index_reg2 = List.find (fun (r:R) -> r <> xtra_index_reg && not (Set.mem r (int_inputs + int_written))) R.All
          Printf.printfn "Picking XtraREG2 %A" xtra_index_reg2

          let (constants,vec_instrs) = do_const_access xtra_index_reg2 vec_instrs



          let instrs0_arr = Array.of_list vec_instrs
          all_instrs_hack := Array.of_list vec_instrs
          let (augs,intdead,vecdead) as x = liveness int_written vec_instrs
          let vec_uninitialized = (Set.of_list X.All) - vecdead

          if comm.Rank = 0 then 
            Printf.printfn "@intdead: %A\n" (Seq.to_list intdead)
            Printf.printfn "@vecdead: %A\n" (Seq.to_list vecdead)
(*          Printf.printfn "@start: %A\n" x
*)
(*
          for info in augs do
            let (vec_in,vec_kill,vec_gen,i) = info
            Printf.printfn "IN(%A), KILL(%A)\n\t\t\t\tINSTR: %A\n\t\tGEN(%A)" (Set.to_list vec_in) (Set.to_list vec_kill) i (Set.to_list vec_gen)
*)
          let asis = List.sum (List.map CodeSize_of_Instr vec_instrs)

          let (totalsize,prep) = List.fold_right prep_of_liveness augs (0,[])
          let ngenes = List.sum (List.map (fun (_,_,i_gen,_,_,v_gen,_) -> List.length i_gen + List.length v_gen) prep)
          let maxcosts = List.sum (List.map (fun (_,_,_,_,_,_,c) -> min c 1) prep)

          if comm.Rank = 0 then
            Printf.printf "totalsize = %d (hoisted)\n" totalsize
            Printf.printf "ngenes   = %d\n" ngenes
            Printf.printf "eval maxcosts = %d\n" maxcosts
            Printf.printf "eval asis = %d (%d + %d)\n" all_size totalsize (all_size - totalsize)

            (*            Printf.printf "prep = %A \n" prep *)

          let prep = Array.of_list prep

          let nbits = ngenes

          let all0 = Genes01.Replicate(0L,ngenes)
          let all1 = Genes01.Replicate(-1L,ngenes)
          let xs = ref [all0, "all0"; all1, "all1"]

(*
          let r = new System.Random()
          for i = 1 to 10000 do
            xs := (Genes01.Random(r,ngenes),"rnd")::!xs
*)
          if comm.Rank = 0 then
            for x in !xs do
              let (x,name) = x
              let c0 = x.Count
              let retval = eval_genes int_inputs n_i_freeLo n_i_freeHi vec_uninitialized x prep
              let (costs,_,_,_,_,_,_) = retval
              let c1 = x.Count
              Printf.printfn "eval %05d (%s) : %A => [%d]=>[%d]" costs name retval c0 c1

          let f (i:Genes01) =             
            let retval = eval_genes int_inputs n_i_freeLo n_i_freeHi vec_uninitialized i prep
            let (costs,_,_,_,_,_,_) = retval
            float 1.0 / (float costs)

          let init () = let i = Genes01.Random(rng,nbits) in (f i, i)
          let xover(xx:Genes01,yy:Genes01,i) = 
            let poss = [ rng.Next(nbits); rng.Next(nbits) ]
            let (xy,yx) = xx.CrossoverMulti(poss,yy)
            if rng.NextDouble () < 0.01 then xy.MutateFlip1(rng.Next(xy.Size))
            if rng.NextDouble () < 0.01 then yx.MutateFlip1(rng.Next(yx.Size))
            (f xy,xy),(f yx,yx)
          let mutate2(xx:Genes01,i) =
            let xx1 = xx.Clone ()
            let xx2 = xx.Clone ()
            xx1.MutateFlip1(rng.Next(xx1.Size))
            xx2.MutateFlip1(rng.Next(xx2.Size))
            (f xx1,xx1),(f xx2,xx2)

          let pop0 = 
            let pop0 = Array.init popsize (fun _ -> init ())
            Array.sort (fun (f1,_) (f2,_) -> compare f1 f2) pop0
            pop0

          let ga = new Ga.GenerationalGeneticAlgorithm<Genes01>(Verbosity=100, PopulationSize=popsize, EliminateDuplicates=false, MaxAsync=1, RNG=rng, MigrationRate=0.1, Population0=pop0, Selection = (* RouletteWheelSelection *) TournamentSelection(0.8,2), Crossover=xover, Mutate2=mutate2, Communicator=comm)

          let best = ga.Do 300

          let instrs1 = Array.to_list (express_genes int_inputs int_written best prep instrs0_arr)


***)
          if comm.Rank = 0 then
            Printf.printf "  label %s\n" lbl
            Printf.printf "    size = %d\n" (List.sum (List.map CodeSize_of_Instr instrs))
            Printf.printf "    instrs = %d\n" instrs.Length
          
          outslices := (Candidate(lbl, instrs)) :: !outslices
(*
  let mutable t0 = DateTime.Now

  let nrounds = 10
  if comm.Rank = 0 then System.Console.WriteLine("# rounds = {0}", nrounds)
*)

(*    if comm.Rank = 0 && round % 1000 = 0 then 
      let t = DateTime.Now
      Printf.printfn "ROUND = %d (%A)" round (t - t0)
      t0 <- t
*)

(*  ga.StopAll ()  *)
(*  Printf.printfn "(%d: %d)" comm.Rank !n_spin *)
(*
  System.Console.WriteLine(Printf.sprintf "(%d: %d)" comm.Rank !n_spin)
*)

  outslices := List.rev !outslices
(*
  if comm.Rank = 0 then
    for slice in slices do
      match slice with
        | Verbatim xs -> for (x:ASTline) in xs do x.Print ()
        | Candidate(lbl,instrs) -> (Label lbl).Print (); for i in instrs do (Instr i).Print ()
*)
;;

(***)

let main (args:string array) =
(*
  use env = new MPI.Environment(ref args, MPI.Threading.Single)
  let world = MPI.Communicator.world
  if world.Rank = 0 then Printf.printfn "env.Threading = %A" MPI.Environment.Threading
  let ndims = 3
  let cart_world_dims = 
    let hlp = ref (Array.create ndims 0)
    MPI.CartesianCommunicator.ComputeDimensions(world.Size,ndims,hlp)
    !hlp
  let cart_world = 
    let cart_world_periods = Array.create ndims true
    new MPI.CartesianCommunicator(world, ndims, cart_world_dims, cart_world_periods, true)

  let text = System.IO.File.ReadAllText(Sys.argv.[1])
  testGA cart_world text
*)
  ()
;;

main (System.Environment.GetCommandLineArgs ());;
