module NXopt

(*
type procedure =
    dictionary *
    all *
    entrypoint (=1) *
    exitpoints (1+)
*)

(* TBD group prefixes and instructions into one: don't eliminate prefixes AND following instruction if either is needed    :::    pinstr = (instr [] * instr) *)

open System
open System.Collections.Generic

open Ast
open X64instr

open InstrIO
open MoreInstrIO

open AstIO
open MoreAstIO

open NXyn
open NXyn.Aux
open NXyn.Aux.Global

open NXyn.Aux.IntMisc
open NXyn.Aux.OtherMisc
open NXyn.Aux.SetMisc

open Basics

open LocalScheduling

open OptSimple

open AbstractInterpretation


(************************************************************************************************************************************************************************************)



let get_all_used_intregs (cc:IBlock []) : Set<R64> =
                         (Set.singleton(R64.rsp),cc)
                         ||> Array.fold (fun used b -> (used,b.Instrs)
                                                       ||> Array.fold (fun used -> AllR_of_Instr
                                                                                   >> List.map (fun(r:R) -> Q(r.Int))
                                                                                   >> Set.ofList
                                                                                   >> Set.union used))

let get_free_global_intregs (cc:IBlock []) : Set<R64> =
                            Set.ofList(R64.All) - get_all_used_intregs(cc)
(*
                            (Set.empty,cc)
                            ||> Array.fold (fun used b -> (used,b.Instrs)
                                                          ||> Array.fold (fun used -> AllR_of_Instr
                                                                                      >> List.map (fun(r:R) -> Q(r.Int))
                                                                                      >> Set.ofList
                                                                                      >> Set.union used))
                            |> fun allused -> Set.ofList(R64.All) - allused
*)           

let alloc_free_global_intreg (preference:R64 -> Int32) (freeRegs:ref<Set<R64>>) : option<R64> =
                             if Set.isEmpty(!freeRegs) then None
                                                       else let r = Seq.maxBy preference !freeRegs
                                                            if preference(r) < 0 then None
                                                                                 else freeRegs := Set.remove r !freeRegs
                                                                                      Some(r)


(*
    let allused = 

    R64.All
    |> List.filter (fun r ->
                         r <> R64.rsp &&
                         cc |> Array.forall (fun b ->
                                                 b.Instrs |> Array.forall (fun i -> not(List.exists (fun(r':R) -> r'.Aliases(Rquad(r))) (AllR_of_Instr(i))))
                                             )
                     )

*)


(***)


let targets_of_Instr (lexSucc:string) (i:Instr) =
    let t1 = if Instr_is_Jfollow(i) then [] else [lexSucc] (* TBD assert that lexSucc <> "" *)
    let t2 = if not(Instr_is_Jthru(i) || Instr_is_Jfollow(i)) then []
                                                              else match i
                                                                    with | NR64s(_) | NM64s(_) -> failwith("NRs|NMs")
                                                                         | I0(I_I0.Retq | I_I0.Ret(_)) -> []
                                                                         | C(_,target) -> match target.S
                                                                                           with | Iexpr.Id({Id.Name = k}) -> [k] (* if k is global or extern=not_defined_by_label ==> unknown *)
                                                                                                | _ -> ["*unknown*"]
                                                                         | _ -> ["*unknown*"]
    t1 @ t2

(***)

(* eliminate_instructions_modifying_RSP
 **************************************
   purpose:       get rid of instructions that alter the stack pointer.
                  assert that stack accesses are of the form RSP+konst. (-> no on-stack random access arrays)

   prerequisites: stack may only be accessed by using register RSP. (abstract interpretation does that)
                  no aliasing of rsp may occur.
                  all push/pops need to be rewritten first, too.
   
   rationale:     expose all accesses to the stack, so deciding stack access aliasing becomes trivial.
   
   caveat:        the resulting code cannot run without adding proper stack modifying instructions.
*)

let eliminate_instructions_modifying_RSP (tDict:Dictionary<String,IBlock>) (cc:String []) =
    let fail(s:String) = failwithf "eliminate_instructions_modifying_RSP: %s: %s" !Global.InFileName s

    let offset_stack_access ofs (m:M) = match m
                                         with K(i) -> m
                                            | K32eip(i) -> m
                                            | K64rip(i) -> m
                                            | EA32(_,None,None) -> m
                                            | EA32(_,Some({Base=b}),None) when b <> R32.esp -> m
                                            | EA32(Iexpr.Int(o),Some({Base=b}),None) -> EA32(Iexpr.Int(o+ofs),Some({Base=R32.mESP}),None)
                                            | EA32(_,Some({Base=b}),None) -> fail("esp with unknown displacement")
                                            | EA32(_,None,Some({Index=i;Scale=s})) -> if i = R32.esp then fail(sprintf "esp as index: %A" m) else m
                                            | EA32(_,Some({Base=b}),Some({Index=i;Scale=s})) -> if b = R32.esp || i = R32.esp then fail("esp as base/index + other reg") else m
                                            | EA64(_,None,None) -> m
                                            | EA64(o,Some({Base=b}),None) when b <> R64.rsp -> m
                                            | EA64(Iexpr.Int(o),Some({Base=b}),None) -> EA64(Iexpr.Int(o+ofs),Some({Base=R64.mRSP}),None)
                                            | EA64(_,Some({Base=b}),None) -> fail("rsp with unknown displacement")
                                            | EA64(_,None,Some({Index=i;Scale=s})) -> if i = R64.rsp then fail(sprintf "rsp as index: %A" m) else m
                                            | EA64(_,Some({Base=b}),Some({Index=i;Scale=s})) -> if b = R64.rsp || i = R64.rsp then fail("rsp as base/index + other reg") else m
    let usesRSP(r:R) = r.Int = R.rsp.Int (* R aliases R64.rsp *)

    let sp = new Dictionary<String,bigint>()
    let rec dfs ofs l = let (got,value) = sp.TryGetValue(l)
                        if got then if value <> ofs then fail(sprintf "dfs(%s): ofs %d vs %d" l (int64 value) (int64 ofs))
                               else let b = tDict.[l]
                                    let i1 = new ResizeArray<Instr>()
                                    sp.[l] <- ofs

                                    (ofs,b.Instrs)
                                    ||> Array.fold (fun ofs -> function R64ds(I_R64ds.Addq(LitInt64(k)), {DS=ds}) when ds = R64.rsp -> ofs + (bigint(k))
                                                                      | R64ds(I_R64ds.Subq(LitInt64(k)), {DS=ds}) when ds = R64.rsp -> ofs - (bigint(k))
                                                                      | R64d(I_R64d.Leaq(EA64(Iexpr.Int(k),Some{Base=s},None)), {D=d}) when s=R64.rsp && s=d -> ofs + k
                                                                      | I0(Retq) as i -> if ofs <> 0I then fail("retq with bad %rsp") else i1.Add(i); 0I
                                                                      | I0(Ret(LitInt64(k))) as i -> if bigint k <> -ofs then fail("ret with bad %rsp") else i1.Add(i); 0I
                                                                      | i -> let i = Map_M_in_Instr (offset_stack_access ofs) i
                                                                             if List.exists usesRSP (AllRds_of_Instr(i)) then fail(sprintf "rsp in Rds: %A" i)
                                                                             if List.exists usesRSP (AllRd_of_Instr(i))  then fail(sprintf "rsp in Rd: %A"  i)
                                                                             if List.exists usesRSP (AllRs_of_Instr(i))  then fail(sprintf "rsp in Rs: %A"  i)
                                                                             let i = Map_R32_in_Instr (fun r -> if r = R32.mESP then R32.esp else r) i
                                                                             let i = Map_R64_in_Instr (fun r -> if r = R64.mRSP then R64.rsp else r) i
                                                                             i1.Add(i)
                                                                             ofs)
                                    |> fun ofs -> b.Instrs <- i1.ToArray()
                                                  for succ in b.DynSuccs do dfs ofs succ

    match try_get_single_entrypoint(cc) with Some(entryPoint) -> dfs 0I entryPoint
                                           | None -> fail("not single entrypoint")
(*
 * assign_unique_stack_locations
 *******************************
   purpose: reassign cells of private stack suchthat no aliasing is performed
            check that no partial reads/writes/modifies occur on the stack
*)


let offset_stack_accesses (by:int) (tDict:Dictionary<String,IBlock>) (cc:String []) =
    for l in cc do
        let b = tDict.[l]
        b.Instrs <- b.Instrs
                    |> Array.map (Map_M_in_Instr (function EA32(Iexpr.Int(ofs),Some{Base=b},None) when b = R32.esp -> EA32(Iexpr.Int(ofs + bigint(by)), Some{Base=b},None)
                                                         | EA64(Iexpr.Int(ofs),Some{Base=b},None) when b = R64.rsp -> EA64(Iexpr.Int(ofs + bigint(by)), Some{Base=b},None)
                                                         | x -> x))

let offset_private_stack_accesses (padding:Int32) (totalAligned:Int32) (tDict:Dictionary<String,IBlock>) (cc:String []) : unit =
    for l in cc do
        let b = tDict.[l]
        b.Instrs <- b.Instrs
                    |> Array.map (Map_M_in_Instr (function EA32(Iexpr.Int(ofs),Some{Base=b},None) when b = R32.esp && ofs <  0I -> EA32(Iexpr.Int(ofs - bigint(padding)),      Some{Base=b},None)
                                                         | EA64(Iexpr.Int(ofs),Some{Base=b},None) when b = R64.rsp && ofs <  0I -> EA64(Iexpr.Int(ofs - bigint(padding)),      Some{Base=b},None)
                                                         | EA32(Iexpr.Int(ofs),Some{Base=b},None) when b = R32.esp && ofs >= 0I -> EA32(Iexpr.Int(ofs + bigint(totalAligned)), Some{Base=b},None)
                                                         | EA64(Iexpr.Int(ofs),Some{Base=b},None) when b = R64.rsp && ofs >= 0I -> EA64(Iexpr.Int(ofs + bigint(totalAligned)), Some{Base=b},None)
                                                         | x -> x))
                                                                             


(* *)


              
(***)

let inline_mem_params1 (tDict:Dictionary<String,IBlock>) (cc:String []) =

    (* 0 heuristically reorder suchthat:... *)
    for lbl in cc do
            let b = tDict.[lbl]
            let instrs = b.Instrs |> Array.toList
            let instrs = instrs |> moveDown_Instrs_like b (fun(i) -> (Fs_of_Instr(i) <> [] || Fd_of_Instr(i) <> []) &&
                                                                     (match i with R64sR64d(_) | R64sR64ds(_) | M64sR64ds(_) | R64sR64s(_) | R64ds(_) | M64sR64s(_) | R64sM64s(_) -> true | _ -> false) &&
                                                                     List.forall ((<>)R.rsp) (Rds_of_Instr(i)) && // NEW
                                                                     List.forall ((<>)R.rsp) (Rd_of_Instr(i)))
            b.Instrs <- Array.ofList(instrs)



    
    (*** 1 *** ANALYZE ***)
    (* get all stack locations referenced by instructions in this connected component *)
    let (s_stacklocs,ds_stacklocs,d_stacklocs) = get_all_stacklocs tDict cc

    (* of all stack locations, only consider procedure parameters *)
    let isParam (size,ofs) = ofs >= 0
    let (s_params, ds_params, d_params) = (Array.filter isParam s_stacklocs, Array.filter isParam ds_stacklocs, Array.filter isParam d_stacklocs)

    (* only select non-aliasing read-only parameters *)
    let s_params  = s_params |> Array.filter (fun s_loc -> ds_params |> Array.forall (nonoverlapping_stackLoc s_loc) &&
                                                           d_params  |> Array.forall (nonoverlapping_stackLoc s_loc) &&
                                                           s_params  |> Array.forall (fun s_loc1 -> s_loc = s_loc1 || nonoverlapping_stackLoc s_loc s_loc1))

    (* recursively collect instructions reading some register <r> that is written by some ``load'' of an on-stack parameter *)
    let spurious = new HashSet<Int32 * Int32>()
    let allUses  = new Dictionary<Int32 * Int32, list<R64 * array<String * Int32>>>()

    let rec dfs (size,ofs) (r:R) (uses:ResizeArray<String * Int32>) (visited:String list) (b:IBlock) (ix:Int32) =
        if ix >= b.Instrs.Length then for succ in b.DynSuccs do if List.forall ((<>)succ) visited then dfs (size,ofs) r uses (succ::visited) tDict.[succ] 0
                                 else let instr = b.Instrs.[ix]
                                      if   List.exists ((=)r) (AllRds_of_Instr(instr))     then spurious.Add((size,ofs)) |> ignore (* too strict? *)
                                      elif not(List.exists ((=)r) (AllRd_of_Instr(instr))) then if List.exists ((=)r) (AllRs_of_Instr(instr)) then uses.Add((b.Name,ix))
                                                                                                dfs (size,ofs) r uses visited b (ix+1)

    (* starting at each ``load'' of an on-stack parameter to some register <X>, collect all uses of <X>  *)
    let rec add_instr_to_use (size,ofs) b ix = function M64sR64d(I_M64sR64d.Movq, {S=s; D=d}) ->
                                                            let uses = new ResizeArray<String * Int32>()
                                                            dfs (size,ofs) (Rquad(d)) uses [b.Name] b (ix+1)
                                                            let others = match allUses.TryGetValue((size,ofs)) with (true,value) -> value | (false,_) -> []
                                                            allUses.[(size,ofs)] <- (d, uses.ToArray()) :: others
                                                      | _ -> spurious.Add((size,ofs)) |> ignore

    (* get all readers of d, traverse connected component, building up use info *)
    for lbl in cc do let b = tDict.[lbl]
                     b.Instrs |> Array.iteri (fun ix instr -> Ms_of_Instr(instr) |> List.choose stackLocOption
                                                                                 |> List.filter (fun(memInfo) -> Array.exists ((=)memInfo) s_params)
                                                                                 |> List.iter (fun(size,ofs) -> add_instr_to_use (size,ofs) b ix instr))
                     

    (*** 2 *** ACT ***)
    (* scaling instructions will have to be inserted at the beginning of procedure entry point *)
    let scalingInstrs = new ResizeArray<Instr>()

    (* for every proper quad-sized on-stack parameter, collect scaling factors and new variants of instructions requiring fewer registers *)
    for KeyValue((size,ofs),uses) in allUses do
        if size = 8 && not(spurious.Contains((size,ofs))) then
            let factors   = new ResizeArray<int>()
            let newInstrs = new Dictionary<(String * Int32), Instr>()

            let appendInstrAndFactor l ix ofs s d = factors.Add(int(s))
                                                    newInstrs.[(l,ix)] <- M64sR64ds(I_M64sR64ds.Addq, {S=M64.S(EA64(Iexpr.Int32(ofs),Some{Base=R64.rsp},None)); DS=d})
            if uses |> List.forall (fun(r,u) -> u |> Array.forall (fun(l,ix) -> match tDict.[l].Instrs.[ix]
                                                                                 with | R64d(I_R64d.Leaq(EA64(_, Some{Base=b}, Some{Index=i;Scale=s})), {D=d})
                                                                                        when d = b && i = r && flags_overwritten_after outFlags__addq tDict.[l].Instrs (ix+1)
                                                                                             -> appendInstrAndFactor l ix ofs s d; true
                                                                                      | _ -> false)) then
                if factors.Count < 1 then failwith "inline_mem_params: factors = [||]"
                (* if all factors are equal and a power of two, then LEAQ can be replaced by ADDQ with in-memory operand + a scaling of the in-memory operand *)
                let f = factors.[0]
                if is_power_of_two(f) && factors.ToArray() |> Array.forall ((=)f) then
                    if f > 1 then scalingInstrs.Add(M64ds(I_M64ds.Shlq(Lit.Int32(lsb f)), {DS=M64.S(EA64(Iexpr.Int32(ofs),Some{Base=R64.rsp},None))}))
                    for KeyValue((lbl,ix),newInstr) in newInstrs do tDict.[lbl].Instrs.[ix] <- newInstr

    (* if any in-memory parameters require scaling by a power-of-two, insert scaling instructions at beginning of procedure *)
    let scalingInstrs = scalingInstrs.ToArray()
//    eprintfn "inline_mem_params: %A" scalingInstrs

    scalingInstrs

let inline_mem_params2 (scalingInstrs:Instr []) (tDict:Dictionary<String,IBlock>) (cc:String []) =
    if scalingInstrs.Length > 0 then match try_get_single_entrypoint(cc) with Some(l) -> let b = tDict.[l]
                                                                                         b.Instrs <- Array.append scalingInstrs b.Instrs
//                                                                                         eprintfn "b.Instrs = %A" b.Instrs
                                                                            | None -> failwith "inline_mem_params: no single entrypoint"

(***)

open OptSimple

type EAsplit = NoSplit
             | Split2 of int64
             | Split3 of int64
             | Split4 of int64
             | Split8 of int64

type Itmp = Eentry
          | Ientry of R64 * int64
          | Bentry of R64 * R64 * int64


type R_W = R | W

let doEAopt = true



(***)



(***)

let eaPhase2 (taken:list<R64>) (freeIRegs:ref<Set<R64>>) (tDict:Dictionary<String,IBlock>) (comp:String []) =
    let inoutDict = elimDC_returnLiveness false tDict comp

    let needs_regs = HashSet<R64>()

    for lbl in comp do
            let b = tDict.[lbl]
            for i in b.Instrs do
                  for r in (Rs_of_Instr(i) @ Rd_of_Instr(i) @ Rds_of_Instr(i)) do
                        if r.Int >= 42 then (* HACK *) needs_regs.Add(Q(r.Int)) |> ignore



    (* are `needs_regs' written to/read from/modified ? *)
    
(*
    for lbl in comp do
            let b = tDict.[lbl]
            eprintfn "block %A" lbl
            for i in b.Instrs do
                  let writes   = Rd_of_Instr(i)  |> List.exists (fun (r:R) -> needs_regs.Contains(Q(r.Int)))
                  let modifies = Rds_of_Instr(i) |> List.exists (fun (r:R) -> needs_regs.Contains(Q(r.Int)))
                  let reads    = Rs_of_Instr(i)  |> List.exists (fun (r:R) -> needs_regs.Contains(Q(r.Int)))


                  if writes || modifies || reads then
                     eprintfn "    "
                     if writes   then eprintf "[writes]"
                     if modifies then eprintf "[modifies]"
                     if reads    then eprintf "[reads]"
                     eprintfn " %A" i
                     
            
*)




    let moveAROUND(lbl) =
            let b = tDict.[lbl]
            let marker = R64sR64d(I_R64sR64d.Imulq(Lit.Int64(4711L)), {S=Q(-1);D=Q(-1)})
            let instrs = marker :: (Array.toList(b.Instrs))

            (* find instructions that use/alter flags. (we might need to move them down so we can insert IMULQ's) *)
            let instrs = instrs |> moveDown_Instrs_like b (fun(i) -> (Fs_of_Instr(i) <> [] || Fd_of_Instr(i) <> []) &&
                                                                     (match i with R64sR64d(_) | R64sR64ds(_) | M64sR64ds(_) | R64sR64s(_) | R64ds(_) | M64sR64s(_) | R64sM64s(_) -> true | _ -> false) &&
                                                                     List.forall ((<>)R.rsp) (Rds_of_Instr(i)) && // NEW
                                                                     List.forall ((<>)R.rsp) (Rd_of_Instr(i)))

            let (in0,out0) = inoutDict.[lbl]
            let instrs = OptForwardStoreToLoad.do_opt b in0 out0 (Array.ofList instrs) |> Array.toList


            let index = List.tryFindIndex ((=)marker) instrs
            if index = None then failwith "OJEOJE List.tryFindIndex ((=)marker) instrs = None"
            let instrs1 = instrs
            let instrs = List.filter ((<>)marker) instrs
            (index,instrs1,instrs) // b.Instrs <- Array.ofList(instrs)
//            index

    if needs_regs.Count > 0 then
         (* XXX in file OLD.EA1 *)

        let filtered = [| 1 .. needs_regs.Count |] |> Array.choose (fun _ -> alloc_free_global_intreg (fun _ -> 0) freeIRegs)


        // SOLLTE NIE PASSIEREN!!!
        if needs_regs.Count > filtered.Length then failwithf "ERROR: not enough registers: %s, need=%A, have=%A" !Global.InFileName (ArrayMisc.array_of_collection needs_regs) (ArrayMisc.array_of_collection filtered)
        else
          let taken = new HashSet<R64>(taken)
          let reg_substs = new Dictionary<R64,R64>()
          needs_regs |> ArrayMisc.array_of_collection |> Array.iteri (fun i r0 -> taken.Add(filtered.[i]) |> ignore
                                                                                  reg_substs.[r0] <- filtered.[i])

          for lbl in comp do
            let b = tDict.[lbl]
            b.Instrs <- b.Instrs |> Array.map (fun(i0) -> let i1 = i0 |> Map_R64_in_Instr (fun x -> let (found,value) = reg_substs.TryGetValue(x) in if found then value else x) in i1)



          elim_nops tDict comp

          
//          comp |> Array.iter (fun lbl -> let b = tDict.[lbl] in b.Instrs <- b.Instrs |> Array.filter (fun(instr) -> Instr_is_nop(instr) |> not))
                            
          let inoutDict = elimDC_returnLiveness false tDict comp

          for lbl in comp do
              let b = tDict.[lbl]

              let (in0,out0) = inoutDict.[lbl]

              let (yes,no) = b.Instrs |> Array.partition (fun (i) -> Fs_of_Instr(i) <> [] || Fd_of_Instr(i) <> [])


              let need_imul = b.Instrs |> Array.filter (fun(i) -> M_of_Instr(i) |> List.exists (function EA64(_,_,Some{Index=_;Scale=s}) -> List.forall ((<>)s) [0L;1L;2L;4L;8L] | _ -> false))
              // eprintfn "lbl = %s, yes.len = %d, no.len = %d, need-imul = %d" lbl yes.Length no.Length need_imul.Length

              let (index,instrs1,instrs) = moveAROUND(lbl)


              

//              eprintfn "lbl = %s" lbl
//              Lifespans.rename_stackcells(instrs1) // Array.toList(b.Instrs))

              if need_imul.Length = 0 then
                  // eprintfn "lbl = %s, yes.len = %d, no.len = %d, need-imul = %d , marker_IX = %A" lbl yes.Length no.Length need_imul.Length index
                  b.Instrs <- Array.ofList(instrs)
              else
                  // we need to do some reg alloc for doing EA calc
                  
                  // find out which (base,index) pairs are being used when i_mul is needed
                  let all_need_imul = new Dictionary<R64*R64, ResizeArray<Int64>>()
                 
                  instrs
                  |> List.iter (fun(i) -> M_of_Instr(i)
                                          |> List.iter (function EA64(_,Some{Base=b},Some{Index=i;Scale=s}) -> // when List.forall ((<>)s) [0L;1L;2L;4L;8L] ->
                                                                      let (found,value) = all_need_imul.TryGetValue((b,i))
                                                                      if found then value.Add(s)
                                                                      else let arr = new ResizeArray<Int64>()
                                                                           arr.Add(s)
                                                                           all_need_imul.[(b,i)] <- arr
                                                               | _ -> ()))

                  let plans = ref [[]]

                  let t0 = DateTime.Now
                  for KeyValue((b,i),factors) in all_need_imul do
                      
                      let sorted = factors.ToArray() |> Set.ofArray |> Set.toArray |> Array.sort
(*
                      eprintfn "(b,i) = %A    sorted = %A" (b,i) sorted 
                      eprintfn "  sorted.[%d] = %d" (sorted.Length/2) sorted.[sorted.Length/2]
                      eprintfn "  sorted.[%d] = %d" (sorted.Length/3) sorted.[sorted.Length/3]
                      eprintfn "  sorted.[%d] = %d" (sorted.Length/4) sorted.[sorted.Length/4]
                      eprintfn "  sorted.[%d] = %d" (sorted.Length/8) sorted.[sorted.Length/8]
  *)                    
                      let splits = ref [NoSplit]
                      let q2 = sorted.[sorted.Length/2] |> int |> find_near_power_of_two
                      if q2 >= 7 then
                         splits := (Split2(int64 q2)) :: !splits
                         let q3 = sorted.[sorted.Length/3] |> int
                         if q3 >= 7 && q3 <> q2 then
                            splits := (Split3(int64 q3)) :: !splits
                            let q4 = sorted.[sorted.Length/4] |> int |> find_near_power_of_two
                            if q4 >= 7 && q4 <> q3 && q4 <> q2 then
                               splits := (Split4(int64 q4)) :: !splits
                               let q8 = sorted.[sorted.Length/8] |> int |> find_near_power_of_two
                               if q8 >= 7 && q8 <> q4 && q8 <> q3 && q8 <> q2 then
                                  splits := (Split8(int64 q8)) :: !splits
                               
                      plans := !plans |> List.fold (fun plans1 plan -> !splits |> List.fold (fun plans1 split -> (((b,i),split)::plan) :: plans1) plans1) []

                  let t1 = DateTime.Now

////                  eprintfn "[EA] makePlans took  %A" (t1-t0)
                  let plans = !plans

                  let leaCosts  = 1.0
                  let imulCosts = 2.5
//                  let imulCosts = 1.5

                  // get free reg
                  let (in0,out0) = inoutDict.[lbl]
                  let tmpregs0 = R64.All
                                 |> List.filter (fun(x) -> x <> R64.rsp &&
                                                           not(taken.Contains(x)) &&
                                                           (Set.contains x !freeIRegs
                                                            || ((instrs |> List.forall (fun(i) -> not(List.exists (fun(y:R) -> y.Aliases(Rquad(x))) (AllR_of_Instr(i)))) &&
                                                                 // was: nextReader_of b (Operand_of_R(Rquad(x))) instrs = None &&
                                                                 not(Set.contains (Operand_of_R(Rquad(x))) out0)))))
                                 |> List.map (fun(x) -> (x,Eentry))
                  if tmpregs0 = [] then failwithf "tmpregs %s" !Global.InFileName
//                  eprintfn "ntmpregs = %d, tmpregs = %A" tmpregs0.Length tmpregs0


                  let measure_EAplan (randDouble:unit -> double) (plan) = 
                    let pDict = new Dictionary<R64 * R64, EAsplit>()
                    plan |> List.iter (fun((b,i),split) -> if split <> NoSplit then pDict.[(b,i)] <- split)

                    let tmpregs = ref tmpregs0

                    let make_leaq (b,(i,k)) d = R64d(I_R64d.Leaq(EA64(Iexpr.Zero,Some{Base=b},Some{Index=i;Scale=k})), {D=d})
                    let make_imulq (s,k) d =
                                   if                  k = 1L                              then [R64sR64d(I_R64sR64d.Movq,                {S=s;D=d})]
                                   elif                Array.exists ((=)k) [|2L;3L;5L;9L|] then [R64d(I_R64d.Leaq(M.B_SI(s,s, k-1L)),         {D=d})]
                                   elif TargetIsAMD() && Array.exists ((=)k) [|4L;8L|]     then [R64d(I_R64d.Leaq(M.SI(s,k)),                 {D=d})]
                                                                                           else let iEntries = (!tmpregs |> List.choose (function(r,Ientry(i',k')) when i' = s && k' <= k -> Some(r,k') | _ -> None)) @ [(s,1L)]
                                                                                                iEntries
                                                                                                |> List.tryPick (fun (i1,k1) ->
                                                                                                                     iEntries
                                                                                                                     |> List.tryPick (fun (i2,k2) ->
                                                                                                                                          [| 1L;2L;4L;8L |]
                                                                                                                                          |> Array.tryPick (fun scale ->
                                                                                                                                                                if k = k1 + k2 * scale then Some(i1,i2,scale)
                                                                                                                                                                                       else None)))
                                                                                                |> function Some(i1,i2,scale) when TargetIsAMD() -> [R64d(I_R64d.Leaq(M.B_SI(i1,i2,scale)), {D=d})]
                                                                                                          | _ -> if k &&& (k-1L) <> 0L then [R64sR64d(I_R64sR64d.Imulq(Lit.Int64(k)), {S=s;D=d})]
                                                                                                                                       else [R64sR64d(I_R64sR64d.Movq, {S=s;D=d});
                                                                                                                                             R64ds(I_R64ds.Shlq(Lit.Int32(lsb(int(k)))), {DS=d})] 



//                    let ti0 = DateTime.Now
                    let i1 = new ResizeArray<Instr>()

                    let costs = ref 0.0
                    let addInstr (i:Instr) =
                                 i1.Add(i)
                                 match i
                                  with | R64sR64d(I_R64sR64d.Imulq(_), _) -> costs := !costs + imulCosts
                                       | R64d(I_R64d.Leaq(_), _)          -> costs := !costs + leaCosts
                                       | _ -> () 

                    let doEAsplit b i (s:EAsplit) =
                                  match s
                                   with | NoSplit -> ()
                                        | Split2(q2) -> let k = List.find (fun k -> q2 % k = 0L) [8L;4L;2L;1L]
                                                        match !tmpregs
                                                         with | (t1,_)::rest when q2/k = 1L ->
                                                                        addInstr(make_leaq (b,(i,k)) t1)
                                                                        tmpregs := rest @ [(t1,Bentry(b,i,q2))]
                                                              | (t0,_)::(t1,_)::rest ->
                                                                        make_imulq (i,q2/k)   t0 |> List.iter addInstr
                                                                        make_leaq  (b,(t0,k)) t1 |> addInstr
                                                                        tmpregs := rest @ [(t0,Ientry(i,q2 / k)); (t1,Bentry(b,i,q2))]
                                                              | _ -> failwith "OJE: not enough regs for split2!"
                                        | Split3(q3) -> match !tmpregs
                                                         with | (t0,_)::(t1,_)::(t2,_)::rest ->
                                                                        let k = List.find (fun k -> q3 % k = 0L) [4L;2L;1L]
                                                                        make_imulq (i,q3/k)      t0 |> List.iter addInstr
                                                                        addInstr(make_leaq  (b,(t0,k))    t1)
                                                                        addInstr(make_leaq  (b,(t0,k*2L)) t2)
                                                                        tmpregs := rest @ [(t0,Ientry(i,q3 / k)); (t1,Bentry(b,i,q3)); (t2,Bentry(b,i,q3*2L))]
                                                              | _ -> failwith "OJE: not enough regs for split2!"

                                        | Split4(q4) -> match !tmpregs
                                                         with | (t0,_)::(t1,_)::(t2,_)::(t3,_)::rest ->
                                                                        let k = List.find (fun k -> q4 % k = 0L) [4L;2L;1L]
                                                                        make_imulq (i,q4/k)       t0 |> List.iter addInstr
                                                                        addInstr(make_leaq  (b, (t0,k))    t1)
                                                                        addInstr(make_leaq  (b, (t0,k*2L)) t2)
                                                                        addInstr(make_leaq  (t2,(t0,k))    t3)
                                                                        tmpregs := rest @ [(t0,Ientry(i,q4 / k)); (t1,Bentry(b,i,q4)); (t2,Bentry(b,i,q4*2L)); (t3,Bentry(b,i,q4*3L))]
                                                              | _ -> failwith "OJE: not enough regs for split2!"

                                        | Split8(q8) -> match !tmpregs
                                                         with | (t0,_)::(t1,_)::(t2,_)::(t3,_)::(t4,_)::(t5,_)::(t6,_)::(t7,_)::rest ->
                                                                        let k = List.find (fun k -> q8 % k = 0L) [2L;1L]
                                                                        make_imulq (i,q8/k)       t0 |> List.iter addInstr
                                                                        addInstr(make_leaq  (b, (t0,k))    t1)
                                                                        addInstr(make_leaq  (b, (t0,k*2L)) t2)
                                                                        addInstr(make_leaq  (b, (t0,k*4L)) t4)
                                                                        addInstr(make_leaq  (t2,(t0,k))    t3)
                                                                        addInstr(make_leaq  (t4,(t0,k))    t5)
                                                                        addInstr(make_leaq  (t4,(t0,k*2L)) t6)
                                                                        addInstr(make_leaq  (t6,(t0,k))    t7)
                                                                        tmpregs := rest @ [(t0,Ientry(i,q8 / k)); (t1,Bentry(b,i,q8)); (t2,Bentry(b,i,q8*2L)); (t3,Bentry(b,i,q8*3L)); (t4,Bentry(b,i,q8*4L)); (t5,Bentry(b,i,q8*5L)); (t6,Bentry(b,i,q8*6L)); (t7,Bentry(b,i,q8*7L))]
                                                              | _ -> failwith "OJE: not enough regs for split2!"
                     
                    (* use some least recently used heuristic for replacement *)
                    let rec loop ix = function // List.iteri? 
                      | [] -> ()
                      | instr::instrs ->
                          let i_need_imul = M_of_Instr(instr) |> List.exists (function EA64(_,_,Some{Index=_;Scale=s}) -> List.forall ((<>)s) [0L;1L;2L;4L;8L] | _ -> false)
                          if i_need_imul && ix >= (Option.get(index)) then
                              let mn = min ix (Option.get(index))
                              let mx = max ix (Option.get(index))
                              for i = mn to mx do eprintfn "error ix=... %d : %A" i instrs1.[i]
                              failwithf "%s:error ix=%d vs index=%d" !Global.InFileName ix (Option.get(index))

                          if not(i_need_imul) then addInstr(instr)
                          else match M_of_Instr(instr)
                                with | [EA64(ofs,Some{Base=b0},Some{Index=i0;Scale=s}) as m] ->
                                    (* TBD rename s -> "k" *)
                                         // if this combination of base+index is to be logarithmically split, do so at first use
                                         let (found,eaSplit) = pDict.TryGetValue((b0,i0))
                                         if found then doEAsplit b0 i0 eaSplit
                                         pDict.Remove((b0,i0)) |> ignore
                                                 

                                         // reorder so that register holding multiple of i0 is overwritten

                                         let t = match !tmpregs
                                                  with | [] -> failwith "EA: no temporary register available"
                                                       | (r,Eentry)::_ -> r
                                                       | entries0 -> let entries = match List.filter (function (_,Bentry(b',i',_)) -> b' <> b0 || i' <> i0
                                                                                                             | (_,Ientry(i',_))    -> i' <> i0
                                                                                                             | _ -> true) entries0
                                                                                    with | [] -> match List.filter (function (_,Ientry(i',_)) -> i' = i0 | _ -> false) entries0
                                                                                                  with | [] -> entries0
                                                                                                       | zs -> zs
                                                                                         | xs -> xs
                                                                          
                                                                     let n_entries = List.length(entries)
                                                                     entries.[min (n_entries - 1) (int(randDouble() * (double(n_entries))))] |> fst

//                                         let t = !tmpregs |> List.head |> fst
(*                                         
                                         match !tmpregs
                                          with | (_,Bentry(b',_,_))::_ when b' = b0 ->
                                                    match !tmpregs |> List.tryFind (function(r,Ientry(i',_)) -> i' = i0 | _ -> false)
                                                     with | None -> ()
                                                          | Some((t,content)) -> tmpregs := (t,content) :: (List.filter (fun(r,_) -> r <> t) !tmpregs)
                                               | _ -> ()


                                         let n_tmpregs = List.length(!tmpregs)
                                         let t = (!tmpregs).[min (n_tmpregs - 1) (int(randDouble() * (double(n_tmpregs))))] |> fst
*)                                         
//                                         let t = randDouble()
//                                         let t = !tmpregs |> List.head |> fst

                                         

                                         let bEntries = (!tmpregs |> List.choose (function(r,Bentry(b',i',k')) when b' = b0 && i' = i0 && k' <= s -> Some(r,k') | _ -> None)) @ [(b0,0L)]
                                         let iEntries = (!tmpregs |> List.choose (function(r,Ientry(i',k'))    when            i' = i0 && k' <= s -> Some(r,k') | _ -> None)) @ [(i0,1L)]

                                         // find all matching combinations
                                         let matching = new ResizeArray<Int32 * option<list<Instr> * R64 * Itmp> * M>()
                                         bEntries
                                         |> List.iter (fun (b,bK) ->
                                                           match s - bK
                                                            with | 0L -> matching.Add((-99999, None, (EA64(ofs,Some{Base=b},None))))
                                                                 | s -> [| 8L;4L;2L;1L |]
                                                                        |> Array.find (fun(k) -> s % k = 0L)
                                                                        |> fun(k) -> let s0 = s/k
                                                                                     matching.Add((int(s0), Some(make_imulq (i0,s0) t, t, Ientry(i0,s0)), EA64(ofs,Some{Base=b},Some{Index=t;Scale=k})))
                                                                
                                                                        iEntries
                                                                        |> List.iter (fun (i,iK) ->
                                                                                          [| 8L;4L;2L;1L |]
                                                                                          |> Array.tryFind (fun(k) -> s = iK * k)
                                                                                          |> function Some(k) -> matching.Add((int(iK) - (if b = b0 || i = i0 then 10000 else 1000), None, EA64(ofs,Some{Base=b},Some{Index=i;Scale=k})))
                                                                                                    | None -> ()))
                          
                                         // pick best matching combination
                                         if matching.Count = 0 then failwith "sorted = []"
                                         let (_,auxInstrsOpt,m') = matching.ToArray() |> Array.minBy get1of3

                                         // if necessary add auxiliary instruction(s)
                                         match auxInstrsOpt
                                          with | None -> ()
                                               | Some(auxInstrs,t,content) -> tmpregs := (List.filter (fun(r,_) -> r <> t) !tmpregs) @ [(t,content)]
//                                                                            if !isBEST then eprintfn "auxInstr = %A, tmpregs = %A" auxInstr !tmpregs
                                                                              auxInstrs |> List.iter addInstr
                                         // update tmpregs suchthat LRU order is maintained
                                         let r_in_m' = R64_of_M(m')
                                         let (yes,no) = List.partition (fun(r,_) -> List.exists ((=)r) r_in_m') !tmpregs
                                         tmpregs := no @ yes
                                                 
                                         // finally map original effective address and emit instruction
                                         instr |> Map_M_in_Instr (fun x -> if x = m then m' else x) |> addInstr

                                     | _ -> failwith "BAD!"
                                     
                          loop (ix+1) instrs
                    loop 0 instrs
                                          
//                    let ti1 = DateTime.Now
                    (!costs, i1.ToArray())


                  eprintfn "[EA %s] |plans| = %d" !Global.InFileName (List.length plans)
//                  System.Threading.Thread.Sleep(10000)
                  let t0 = DateTime.Now

                  let bestOption = ref None

//                  for p in plans do
//                      eprintfn "plan: %A" p

                  plans
                  |> List.iter (fun plan ->
                                    try let (costs,instrs) = measure_EAplan (fun _ -> 0.0) (plan)
                                        match !bestOption
                                         with | Some(costsBest,_,_) when costsBest <= costs -> ()
                                              | _ -> bestOption := Some(costs,instrs,plan)
                                    with _ -> ())
                  
//                  let measured_plans = plans |> List.choose (fun(plan) -> try Some(measure_EAplan(plan)) with _ -> None) // begin eprintfn "%s: plan = %A ==> EXN" !Global.InFileName plan; None end)

                  let t1 = DateTime.Now
                  eprintfn "[EA %s] measuring time = %A" !Global.InFileName (t1-t0)
//                  System.Threading.Thread.Sleep(10000)

                  match !bestOption
                   with | None -> failwith "nxas: no plans!"
                        | Some(bestcosts,bestinstrs,bestplan) ->
                            eprintfn "%s: bestplan = %A, bestcosts = %A" !Global.InFileName  bestplan bestcosts
                            let t0 = DateTime.Now

                            let bestcosts0 = bestcosts
                            // randomize a bit
                            let rng = new System.Random()
                            let best = ref (bestcosts,bestinstrs)
                            for i = 0 to 100 do
                                  let (vcosts,vinstrs) = measure_EAplan (fun () -> let r = rng.NextDouble() in r * r * r * r) bestplan
                                  let (bestcosts,bestinstrs) = !best
                                  if vcosts < bestcosts then
                                      eprintfn "%s: new best %A => %A" !Global.InFileName bestcosts0 vcosts
                                      LogInfo(sprintf "new best %A => %A" bestcosts0 vcosts)

                                      best := (vcosts,vinstrs)
                            let t1 = DateTime.Now
                                      
                            eprintfn "[EA %s] additional tuning time = %A" !Global.InFileName (t1-t0)

                            let (bestcosts,bestinstrs) = !best
                            b.Instrs <- bestinstrs
                            



(***)

(* must be used AFTER rewrite_push_pop *)
let remove_save_callee_saved_regs (tDict:Dictionary<String,IBlock>) (cc:String []) =
    let entryBlock = get_single_entrypoint(cc) |> fun l -> tDict.[l]
    let exitBlocks = get_nonempty_exitpoints tDict cc |> Array.map (fun l -> tDict.[l])

    (* instruction patterns to be recognized and removed *)
    let (|LoadFromStack|_|) set = function M64sR64d(I_M64sR64d.Movq,{S=M64.S(EA64(_,Some({Base=b}),None)); D=d}) when b = R64.rsp && Set.contains d set -> Some(d)
                                         | _ -> None

    let (|StoreToStack|_|) set = function R64sM64d(I_R64sM64d.Movq,{S=s;D=M64.S(EA64(_,Some({Base=b}),None))}) when b = R64.rsp && Set.contains s set -> Some(s)
                                        | _ -> None
 
    let get_AllR64(i) = AllR_of_Instr(i) |> List.map (fun(r:R) -> Q(r.Int)) |> Set.ofList

 
    (* step through entry block and exit blocks to find matching stores/loads of callee saved registers *)
    ((Set.empty,Set.ofList(R64.CalleeSaved)), entryBlock.Instrs)
    ||> Array.fold (fun (remove,saved) -> function StoreToStack saved s -> (Set.add s remove,Set.remove s saved)
                                                 | i -> (remove,saved - get_AllR64(i)))
    |> fun (remove,saved) -> (remove,exitBlocks)
    ||> Array.fold (fun remove exitBlock ->
                        (exitBlock.Instrs, (Set.empty,Set.ofList(R64.CalleeSaved)))
                        ||> Array.foldBack (fun i (remove1,saved) -> match i with LoadFromStack saved d -> (Set.add d remove1, Set.remove d saved)
                                                                                | i -> (remove1,saved - get_AllR64(i)))
                        |> fun (remove1,saved) -> Set.intersect remove1 remove)

    (* remove greatest common set of loads/stores of callee saved registers *)
    |> fun remove -> (([],remove), entryBlock.Instrs)
                     ||> Array.fold (fun (instrs,remove) -> function StoreToStack remove s -> (instrs,remove)
                                                                   | i -> (i :: instrs, remove - get_AllR64(i)))
                     |> fun (instrs,_) -> entryBlock.Instrs <- Array.ofList(List.rev(instrs))

                     exitBlocks
                     |> Array.iter (fun exitBlock -> (exitBlock.Instrs,([],remove))
                                                     ||> Array.foldBack (fun i (instrs,remove) -> match i with LoadFromStack remove d -> (instrs,remove)
                                                                                                             | i -> (i::instrs, remove - get_AllR64(i)))
                                                     |> fun (instrs,_) -> exitBlock.Instrs <- Array.ofList(instrs))

                     (* return set of callee saved registers for which load/stores have been deleted and pushes/pops need to be inserted later on *)
//                     eprintfn "remove = %A" (Seq.toList remove)
                     remove

(***)

let reallocStack (fastSPoption:option<Int32 * R64>) (how:ReassignStack) (tDict:Dictionary<String,IBlock>) (comp:String []) =

    let cc = comp |> Array.map (fun l -> tDict.[l])

    let inoutDict = elimDC_returnLiveness true tDict comp

    let alltobesaved = R64.CalleeSaved
                         |> Set.ofList
                         |> Set.remove R64.rsp
                         |> Set.intersect (Set.union (match fastSPoption with None -> Set.empty | Some(_,r) -> Set.singleton(r))
                                                     (get_all_used_intregs(cc)))
                         |> Set.toArray


    let n_alltobesaved = Array.length(alltobesaved)

    let mutable n_tobesaved_aligned = n_alltobesaved
    // keep aligned to 32 bytes
    while n_tobesaved_aligned % 4 <> 0 do n_tobesaved_aligned <- n_tobesaved_aligned + 1 // TBD closed form

    let aligned_size = (n_tobesaved_aligned * 8)
    let padding_size = ((n_tobesaved_aligned - n_alltobesaved) * 8)

    // duerfen nur vielfache von 32 schieben
    // rsp ist anfangs nicht aligned mod 32 = 0, sondern aligned mod 32 = ..
    offset_private_stack_accesses padding_size (n_alltobesaved * 8) tDict comp

//    eprintfn "padding = %d" padding_size
//    eprintfn "aligned_size = %d" aligned_size

    (* get_max_stack_size *)
    let stacksize = ref 0
    for l in comp do
          let b = tDict.[l]
          b.Instrs |> Array.iter (fun(i) -> (Ms_of_Instr(i) @ Mds_of_Instr(i) @ Md_of_Instr(i)) |> List.choose stackLocOption |> List.iter (fun(_,ofs) -> if ofs < 0 then stacksize := max !stacksize (-ofs)))

    while !stacksize % 32 <> 0 do incr(stacksize) // ? 
    let stacksize = !stacksize

//    eprintf "stacksize = %d" stacksize

    let useExtraPage = match how with ReassignStack.No -> false
                                    | ReassignStack.Yes | ReassignStack.Yes_Xsp -> true
                                    // TargetIsIntel()
      
    let stacksize = stacksize + (if not(useExtraPage) then 0 else (256 + (if fastSPoption = None then 0 else 256)))
                                                  
    let e_stack_size = padding_size + stacksize + 24 (* so that `0(%rsp)' is aligned to 32 *)
    if stacksize > 0 then offset_stack_accesses e_stack_size tDict comp
    
    // all regs are real again, so we can do AI again!
    (* this also allocates ONE (32-byte aligned) region on the stack *)
    (* optStack *)
    if useExtraPage then
       for lbl in comp do
               let b = tDict.[lbl]
               let (in0,out0) = inoutDict.[lbl]

               let inout0 = Set.union in0 out0
               
               // assert that accesses are non-overlapping
               let stackaccess = Dictionary<Int32,HashSet<Int32*Int32>>()
               b.Instrs |> Array.iter (fun instr ->
                                           Md_of_Instr(instr)
                                           |> List.choose stackLocOption
                                           |> List.iter (fun (size,ofs) ->
                                                             for i = 0 to size - 1 do
                                                                   let tbl = match stackaccess.TryGetValue(ofs+i)
                                                                              with | (true,tbl) -> tbl
                                                                                   | (false,_) -> let tbl = new HashSet<Int32*Int32>()
                                                                                                  stackaccess.[ofs+i] <- tbl
                                                                                                  tbl
                                                                   tbl.Add((size,ofs)) |> ignore))

               // get non-overlapping stack accesses
               let candidates = stackaccess
                                |> ArrayMisc.array_of_collection
                                |> Array.map (fun (KeyValue(pos,accesses)) -> (pos, ArrayMisc.array_of_collection(accesses)))
                                |> Array.choose (function (pos,[|access|]) -> Some(pos,access) | _ -> None)
                                |> Array.filter (fun(pos,(size,ofs)) -> pos = ofs && [| 0 .. size-1 |] |> Array.map ((+)ofs) |> Array.forall (fun(x) -> not(Set.contains (PstackB x) inout0)))
                                |> Array.map snd

//               eprintfn "[lbl = %s] candidates0 = %A (%d)" lbl candidates candidates.Length

               // add access information
               let candDict = new Dictionary<Int32*Int32, ResizeArray<R_W * Int32>>()
               candidates |> Array.iter (fun(size,ofs) -> candDict.[(size,ofs)] <- new ResizeArray<R_W * Int32>())

               let maybeAddAccess (what:R_W) (ix:Int32) (item:(Int32*Int32)) =
                                  match candDict.TryGetValue(item)
                                   with | (false,_) -> ()
                                        | (true,arr) -> arr.Add((what,ix))

               b.Instrs |> Array.iteri (fun ix instr -> (Ms_of_Instr(instr) @ Mds_of_Instr(instr)) |> List.choose stackLocOption |> List.iter (fun(size,ofs) -> maybeAddAccess R ix (size,ofs))
                                                        Md_of_Instr(instr)                         |> List.choose stackLocOption |> List.iter (fun(size,ofs) -> maybeAddAccess W ix (size,ofs)))

               let candidates = candDict
                                |> ArrayMisc.array_of_collection
                                |> Array.map (fun (KeyValue((size,ofs),accesses)) -> ((size,ofs),accesses.ToArray()))
                                |> Array.filter (fun ((size,ofs),accesses) ->
                                                     accesses <> [||] &&
                                                     W = fst(accesses.[0]) &&
                                                     R = fst(accesses.[accesses.Length-1]) &&
                                                     [1..accesses.Length-1] |> List.forall (fun(ix) -> not (fst(accesses.[ix-1]) = W && fst(accesses.[ix]) = W)))


               if candidates <> [||] then
                 let sizes = Dictionary<Int32, Int32>()

                 candidates |> Array.iter (fun ((size,_),accesses) ->
                                               sizes.[size] <- match sizes.TryGetValue(size)
                                                                with | (true,n) -> n
                                                                     | (false,_) -> 0)

                 let sizes = sizes |> ArrayMisc.array_of_collection |> Array.map (fun (KeyValue(size,n)) -> (size,n))
                 let bestsize = Array.maxBy snd sizes |> fst

                 let candidates = candidates |> Array.filter (fun((size,_),_) -> size = bestsize)

                 let splitcandidates = new ResizeArray<Int32 * Int32 * (Int32 * Int32)>()
                 candidates |> Array.iter (fun ((_,ofs),accesses) ->
                                               let rec loop (wtime:Int32) (n:Int32) xs =
                                                            match xs
                                                             with | [] | (W,_)::_ -> failwith "splitcandidates(1)"
                                                                  | [(R,rtime)] -> splitcandidates.Add((ofs,n+1,(wtime,rtime)))
                                                                  | (R,_)::(R,rtime)::others -> loop wtime (n+1) ((R,rtime)::others)
                                                                  | (R,rtime)::(W,wtime1)::others -> splitcandidates.Add((ofs,n+1,(wtime,rtime)))
                                                                                                     loop wtime1 1 others
                                               match accesses |> Array.toList
                                                with | (W,wtime)::others -> loop wtime 1 others
                                                     | _ -> failwith "splitcandidates(2)")
                                                                    
                 let ls = Array.toList(candidates) |> List.map (fun ((size,ofs),accesses) -> (ofs,accesses.Length,(accesses.[0] |> snd, accesses.[accesses.Length - 1] |> snd)))
//                 let ls = splitcandidates.ToArray() |> Array.toList
                 let ls' = Lifespans.join_lifespans ls


                 let howmany16 = 256 / bestsize
                 
                 let howmany = (256 * (1 + if fastSPoption = None then 0 else 1)) / bestsize
                 

//                 let (yes,no) = ls' |> List.partition (fun(i,_) -> i <= howmany)
//                 let yes_S = yes |> List.map snd |> List.map (List.map (fun(_,i,_) -> i) >> List.sum) |> List.sum
//                 let no_S  = no  |> List.map snd |> List.map (List.map (fun(_,i,_) -> i) >> List.sum) |> List.sum
//                 
//                 eprintfn "lifespansA = (%d in first 16, %d in rest, %d total) [%d]" yes_S no_S (yes_S+no_S) ls'.Length
//                 for l' in ls' do eprintfn "   %A" l'


                 let ls = splitcandidates.ToArray() |> Array.toList
                 let ls' = Lifespans.join_lifespans ls

                 let (yes0,no0) = ls' |> List.partition (fun(i,_) -> i <= howmany16)
                 let (yes,no) = ls' |> List.partition (fun(i,_) -> i <= howmany)

                 let yes0_S = yes0 |> List.map snd |> List.map (List.map (fun(_,i,_) -> i) >> List.sum) |> List.sum
                 let no0_S  = no0  |> List.map snd |> List.map (List.map (fun(_,i,_) -> i) >> List.sum) |> List.sum

                 let yes_S = yes |> List.map snd |> List.map (List.map (fun(_,i,_) -> i) >> List.sum) |> List.sum
                 let no_S  = no  |> List.map snd |> List.map (List.map (fun(_,i,_) -> i) >> List.sum) |> List.sum

                 eprintfn "lifespansA = (%d in first 16, %d in rest, %d total) [%d]" yes0_S no0_S (yes0_S+no0_S) ls'.Length
                 eprintfn "lifespansB = (%d in first 16/32, %d in rest, %d total) [%d]" yes_S no_S (yes_S+no_S) ls'.Length

                 let yes = match fastSPoption
                            with | None -> yes |> List.map (fun (i,spans) ->
                                                                let ofs = bestsize * (i/2) * (if i &&& 1 = 0 then -1 else 1)
                                                                (R64.rsp,ofs,spans))
                                 | Some(deltaSP,fastSP) -> let maxHowmanyFast = 256 / bestsize
                                                           yes |> List.mapi (fun ix (i,spans) ->
                                                                                 let i   = if ix < maxHowmanyFast then i else (i - maxHowmanyFast)
                                                                                 let ofs = bestsize * (i/2) * (if i &&& 1 = 0 then -1 else 1)
                                                                                 let reg = if ix < maxHowmanyFast then fastSP else R64.rsp
                                                                                 (reg,ofs,spans))

                 let map_M (ix:Int32) (m:M) : M =
                           match stackOfsOption(m)
                            with | None -> m
                                 | Some(ofs) -> match yes |> List.tryFind (fun(_,ofs1,spans) -> spans |> List.exists (fun(ofs0,_,(lb,ub)) -> ofs0 = ofs && ix >= lb && ix <= ub))
                                                 with | Some(reg1,ofs1,_) -> M.B_D(reg1,int64(ofs1))
                                                      | None -> m
                 b.Instrs <- b.Instrs |> Array.mapi (fun ix -> Map_M_in_Instr (map_M ix))


    (* promote use of a register other than `%rsp' as the stack-pointer, as this improves instruction encoding *)
    fastSPoption |> Option.iter (fun (delta,sp) ->
                                     for lbl in comp do
                                             let b = tDict.[lbl]
                                             let map_M (m:M) =
                                                       match stackOfsOption(m)
                                                        with | Some(ofs) when ofs >= 128 -> M.B_D(sp,int64(ofs - delta))
                                                             | None | Some _ -> m
                                             b.Instrs <- b.Instrs |> Array.map (Map_M_in_Instr map_M))




    (* insert_save_callee_saved_regs *)
    get_single_entrypoint(comp)
    |> fun l -> let b = tDict.[l]
                let instrs_push_calleesaved = alltobesaved |> Array.map (fun(r) -> R64s(I_R64s.Pushq, {S=r}))
                let instrs_adapt_rsp        = if stacksize = 0 then [||] else [| R64ds(I_R64ds.Subq(Lit.Int32(e_stack_size)),{DS=R64.rsp}) |]
                let instrs_set_sp           = match fastSPoption with None -> [||] | Some(delta,sp) -> [| R64d(I_R64d.Leaq(M.B_D(R64.rsp,int64(delta))), {D=sp}) |]
                b.Instrs <- Array.concat([instrs_push_calleesaved; instrs_adapt_rsp; instrs_set_sp; b.Instrs])
                                          
    get_nonempty_exitpoints tDict comp
    |> Array.iter (fun l ->
                       let b = tDict.[l]
                       let instrs_butlast         = Array.sub b.Instrs 0 (b.Instrs.Length - 1)
                       let instrs_adapt_rsp       = if stacksize = 0 then [||] else [| R64ds(I_R64ds.Addq(Lit.Int32(e_stack_size)),{DS=R64.rsp}) |]
                       let instrs_pop_calleesaved = alltobesaved |> Array.rev |> Array.map (fun(r) -> R64d(I_R64d.Popq, {D=r}))
                       let instrs_last            = Array.sub b.Instrs (b.Instrs.Length - 1) 1
                       b.Instrs <- Array.concat([instrs_butlast; instrs_adapt_rsp; instrs_pop_calleesaved; instrs_last]))
                   // Array.sub b.Instrs 0 (b.Instrs.Length-1)
//                                                       [| R64ds(I_R64ds.Addq(Lit.Int32(e_stack_size)),{DS=R64.rsp}) |]
//                                                       alltobesaved |> Array.rev |> Array.map (fun(r) -> R64d(I_R64d.Popq, {D=r}))
//                                                       instrs_last // Array.sub b.Instrs (b.Instrs.Length-1) 1
//                                                     ])

(***)

let vecRegReassign (how:ReassignXMM) (tDict:Dictionary<String,IBlock>) (comp:String []) =
    match how with
        | ReassignXMM.No -> ()
        | _ -> let inoutDict = elimDC_returnLiveness true tDict comp

               for lbl in comp do
                       let b = tDict.[lbl]
                       let (in0,out0) = inoutDict.[lbl]

                       b.Instrs <- b.Instrs
(*                                   |> Array.toList
                                   |> moveDown_Instrs_like b (function XsXd(_) | M128sXd(_) | M64sXd(_) | M32sXd(_) -> true
                                                                     | _ -> false)
                                   |> Array.ofList
*)                                   
                                   |> swap_copy_operands (in0,out0)
                                   |> OptVecReallocGA.doit how b (in0,out0)

(***)

let scheduling (tDict:Dictionary<String,IBlock>) (comp:String []) =
    let inoutDict = elimDC_returnLiveness true tDict comp

//    eprintfn "scheduling ---"

    let schedinfo = match Basics.scheduling_target()
                     with | AMDk10   -> ListScheduling_AMD_K10.schedParams
                          | IntelCore2 | IntelCoreI7 -> ListScheduling_Intel_Core2_Conroe.schedParams
//                          | IntelCoreI7 -> ListScheduling_Intel_Core2_Conroe.schedParams

    GlobalScheduling.doit schedinfo inoutDict tDict comp

(***)

(* if you don't need to add or remove blocks, operate on  cc: IBlock [] *)

(***)

let rewrite_push_pop (cc:IBlock []) =
                     let is_supported (i:Instr) =
                                      match i
                                       with | I0(I_I0.Enterq(_)) | I0(I_I0.Leaveq) -> false                         (* unsupported: enter/leave *)
                                            | I0(I_I0.Pushfq) | I0(I_I0.Popfq) -> false                             (* unsupported: push/pop flags *)
                                            | M64s(I_M64s.Pushq,_) | M64d(I_M64d.Popq,_) -> false                   (* unsupported: push/pop 64-bit mem->mem *)
                                            | R16s(I_R16s.Pushw,_) | R16d(I_R16d.Popw,_) -> false                   (* unsupported: push/pop 16-bit registers *)
                                            | M16s(I_M16s.Pushw,_) | M16d(I_M16d.Popw,_)  -> false                  (* unsupported: push/pop 16-bit mem->mem *)
                                            | R64s(I_R64s.Pushq,{S=r}) | R64d(I_R64d.Popq,{D=r}) -> r <> R64.rsp    (* unsupported: push/pop stack-pointer *)
                                            | _ -> true
                     let map1 (i:Instr) =
                              match i
                               with | I0(I_I0.Pushq(l))         -> [| R64d(I_R64d.Leaq(M.B_D(R64.rsp,-8L)),{D=R64.rsp});      M64d(I_M64d.Movq(l),{D=M64.S(M.B(R64.rsp))})          |]
                                    | R64s(I_R64s.Pushq,{S=s})  -> [| R64d(I_R64d.Leaq(M.B_D(R64.rsp,-8L)),{D=R64.rsp});      R64sM64d(I_R64sM64d.Movq,{S=s;D=M64.S(M.B(R64.rsp))}) |]
                                    | R64d(I_R64d.Popq,{D=d})   -> [| M64sR64d(I_M64sR64d.Movq,{S=M64.S(M.B(R64.rsp));D=d});  R64d(I_R64d.Leaq(M.B_D(R64.rsp,8L)),{D=R64.rsp});     |]
//                                    | I0(I_I0.Ret(LitInt64(k))) -> [| R64d(I_R64d.Leaq(M.B_D(R64.rsp,k)),{D=R64.rsp});        I0(I_I0.Retq)                                         |]
                                    | x -> [|x|]
                     cc |> Array.iter (fun(b) -> if b.Instrs |> Array.forall is_supported then b.Instrs <- b.Instrs |> Array.map map1 |> Array.concat
                                                                                          else failwithf "rewrite_push_pop: unsupported instruction in %A" b.Instrs)
 
(***)

let optimize2_cc (tDict:Dictionary<String,IBlock>) (comp:String []) =

    let cc = comp |> Array.map (fun l -> tDict.[l])
    


//    let single_instr_peephole_opt (i:Instr) =

    
    (* TBD is there any c in comp that is not in tDict? *)
    
    rewrite_push_pop cc



    

   
    (* eliminate dead code and get liveness information *)

    (* AI *)
    let ((iregsIn,iregsOut),(stackIn,stackOut)) = do_AI tDict comp
    do_AI1 tDict comp ((iregsIn,iregsOut),(stackIn,stackOut)) ai_map_instr1a
    do_AI1 tDict comp ((iregsIn,iregsOut),(stackIn,stackOut)) ai_map_instr1b
    do_AI1 tDict comp ((iregsIn,iregsOut),(stackIn,stackOut)) ai_map_instr_omit_frame_pointer


    if redo_EA() = RedoEA.Yes then
       do_AI1 tDict comp ((iregsIn,iregsOut),(stackIn,stackOut)) ai_map_instr2
       // hoist out scaling of indices to the function entry point
       do_AI3 tDict comp

        
    // assert that all stack accesses can be tracked
    // now all accesses to the stack use (%rsp)+displacement form
    // all accesses to stack should only depend on RSP, no other aliasing is allowed

    let inoutDict = elimDC_returnLiveness true tDict comp
    eliminate_instructions_modifying_RSP tDict comp




//    let inoutDict = elimDC_returnLiveness true tDict comp
//    let inoutDict = elimDC_returnLiveness true tDict comp

    // MEMPARAMS (reduces register pressure slightly)
    // delay insertion until RSP is stable (again)
    let scalingInstrs = inline_mem_params1 tDict comp

    (* elim unneeded stores *)
    let inoutDict = elimDC_returnLiveness false tDict comp

    
    elim_useless_vec_copies tDict inoutDict

    let inoutDict = elimDC_returnLiveness false tDict comp

    // EA
//    let ((iregsIn,iregsOut),(stackIn,stackOut)) = do_AI tDict comp
//
//    let inoutDict = elimDC_returnLiveness false tDict comp
//    let inoutDict = elimDC_returnLiveness false tDict comp

    // re-in-line
    let ((iregsIn,iregsOut),(stackIn,stackOut)) = do_AI tDict comp
    do_AI1 tDict comp ((iregsIn,iregsOut),(stackIn,stackOut)) ai_map_instr1c


    let inoutDict = elimDC_returnLiveness false tDict comp // XXX includes elim_nops?
    

    let removed_callersaved = remove_save_callee_saved_regs tDict comp




    OptPeephole.peephole1(cc)
    OptPeephole.peepholeN(cc)



    // determine which integer registers are globally unused (and could be used for EA temporary values)
    let free_regs0 = get_free_global_intregs(cc)
    let free_regs  = free_regs0 |> ref
 

    (* try allocating second stack pointer which must be a register with 'cheap' encoding *)
    let fastSPoption = match reassign_stack()
                        with | No | Yes -> None
                             | Yes_Xsp -> free_regs |> alloc_free_global_intreg (fun(r) -> if   List.forall ((<>)r) R64.EncodedBest then -1
                                                                                           elif List.forall ((<>)r) R64.CalleeSaved then  1
                                                                                                                                    else  0)

    // assure that register `r' will not be used by EA-phase II, even though the instructions using are not yet inserted...
//    let taken =
//        match fastSPoption with
//            | Some(_,r) -> [r] 
//            | None -> []
    if redo_EA() = RedoEA.Yes then
       eaPhase2 (Option.toList(fastSPoption)) free_regs tDict comp

    // set `delta' of second fast stack-pointer (if any) to the original one
    let fastSPoption = fastSPoption |> Option.map (fun(r) -> (256,r))

    scheduling tDict comp

    reallocStack fastSPoption (reassign_stack()) tDict comp

    

    vecRegReassign (reassign_xmm()) tDict comp

    inline_mem_params2 scalingInstrs tDict comp


  
    scheduling tDict comp



//    vecRegReassign (reassign_xmm_after_sched()) tDict comp


//    if TargetIsAMD then vecRegReassign tDict comp


(***)

let can_be_optimized (tDict:Dictionary<string, IBlock>) (cc:String []) =
    Array.length(Array.filter isFunction cc) = 1 &&
    cc |> Array.forall (fun lbl -> let b = tDict.[lbl]
                                   b.Instrs |> Array.forall (fun(i) -> not(Instr_is_Prefix(i))   &&
                                                                       FixedXs_of_Instr(i)  = [] &&
                                                                       FixedXds_of_Instr(i) = [] &&
                                                                       FixedXd_of_Instr(i)  = [] &&
                                                                       (Instr_is_FunReturn(i) || (let fxd = FixedRd_of_Instr(i) @ FixedRds_of_Instr(i) @ FixedRs_of_Instr(i)
                                                                                                  fxd = [] ||
                                                                                                  fxd = [R.rsp])))) &&
    cc |> Array.forall (fun lbl -> let b = tDict.[lbl]
                                   b.Instrs = [||] ||
                                   (match b.Instrs.[b.Instrs.Length - 1] with
                                        | C(I_C.Callq, _) -> false
                                        | C(I_C.Jmp, {S=s}) -> List.forall (fun(l) -> Array.exists ((=)l) cc) s.Ids
                                        | _ -> true))

let optimize_cc (tDict:Dictionary<string, IBlock>) (cc:String []) =
    if can_be_optimized tDict cc then
//        LogInfo(sprintf "optimizable component : %A" cc)
//        eprintfn "%s: optimizable component : %A" !Global.InFileName cc
        optimize2_cc tDict cc
    else   ()
//        LogInfo(sprintf "NONoptimizable component : %A" cc)

(***)

open Disassembler

(***)

let main (logFileName:String) (inFileName:String) (program0:Ast.Program) =
    program0
    (* register all labels *)
    |> Array.map (fun(name,args,lines) -> lines |> Array.iter (function ALabel(l,_) -> labelGen.Register(l)
                                                                      | Instr(i) -> Id_of_Instr(i) |> List.iter (fun(i) -> labelGen.Register(i.Name))
                                                                      | Directive(d,args) -> labelGen.Register("." + d)
                                                                                             args |> List.fold Fold_Id_in_Item []
                                                                                                  |> List.iter (fun(i) -> labelGen.Register(i.Name)))
                                          (name,args,lines))

    (* replace relative id's "." with absolute ones: this allows arbitrary reordering of some directives like 'type' and 'size' *)
    |> Array.map (fun(name,args,lines0) -> let lines1 = ResizeArray<ILine>()
                                           let maybeAllocLabel = function Some(l) -> l
                                                                        | None -> let lbl = labelGen.Draw()
                                                                                  lines1.Add(ALabel(lbl,1))
                                                                                  lbl
                                           let doLine here line = let add1 f ids = lines1.Add(if List.forall (fun(i:Id) -> i.Name <> ".") ids then line else f(maybeAllocLabel(here))); None
                                                                  match line
                                                                   with | ALabel(l,_) -> lines1.Add(line); Some(l)
                                                                        | Directive(d,args) -> List.fold Fold_Id_in_Item [] args
                                                                                               |> add1 (fun(l) -> Directive(d,List.map (Map_Id_in_Item (fun i -> if i.Name = "." then {Name=l} else i)) args))
                                                                        | Instr(i) -> Id_of_Instr(i)
                                                                                      |> add1 (fun(l) -> Instr(Map_Id_in_Instr (fun i -> if i.Name = "." then {Name=l} else i) i))
                                           (None,lines0) ||> Array.fold doLine |> ignore
                                           let lines2 = lines1.ToArray()

                                           (* if two labels are next to each other (with no instructions or directives in between) then pass on alignment the first to the next *)
                                           (* this allows eliminating either one if it used otherwise *)
                                           let lines = Array.zeroCreate lines2.Length
                                           let mutable align = 1
                                           for i = 0 to lines.Length - 1 do
                                                 lines.[i] <- match lines2.[i]
                                                               with | ALabel(l,a) -> align <- max a align; ALabel(l,align)
                                                                    | (Instr(_) | Directive(_)) as line -> align <- 1; line
                                           (name,args,lines))

    (* split data directives carrying multiple items *)
    |> Array.map(fun(name,args,lines0) -> let lines = new ResizeArray<ILine>(lines0.Length)
                                          lines0 |> Array.iter (function DataDirective(d,args) -> args |> List.iter (fun arg -> lines.Add(Directive(d,[arg])))
                                                                       | line -> lines.Add(line))
                                          (name,args,lines.ToArray()))

    (* find all instructions (typically nops) that are encoded as data: if there is unknown data in text segments then analysis is infeasible *)
    |> disassembleData_in_codeSections
(*
    |> fun sections -> sections |> Array.iter (fun(name,_,lines) -> if name = ".text" then
                                                                       let instrs = lines |> Array.choose (function Instr(i) -> Some(i) | Directive _ | ALabel _ -> None)
//                                                                       InstrSize.PrepareForInstrs08(instrs)
                                                                       let sum = instrs |> InstrSize.sizes_of_Instrs |> Array.sum
                                                                       eprintfn "%s: size0=nbytes %d" !Global.InFileName sum)
                       sections
*)
    (* eliminate nop instructions in "text" code-section *)
    |> Array.map (fun(name,args,lines) -> (name,args,lines |> if name <> ".text" then id else Array.filter (function Instr(i) -> not (Instr_is_nop(i)) | _ -> true)))

                      
    (* move attribute directives out of the way, build up explicit attribute dictionary *)
    |> Array.map (fun(name,args,lines) -> let doAttrLine d = function Iexpr(Id({Name=n}))::args -> let others = if attributes.ContainsKey(n) then attributes.[n] else []
                                                                                                   attributes.[n] <- (d,args)::others
                                                                    | _ -> incr nErrors; eprintfn "missing id for attribute-directive '%s'" d
                                          (name,args,lines |> Array.filter (function AttributeDirective(d,args) -> doAttrLine d args; false | _ -> true)))
    |> if !nErrors = 0 then id else failwithf "%d attribute directives were unknown or not supported" !nErrors
 
    (* get labels that are actually used and eliminate unused ones (which may have been referenced from eliminated debugging-info sections) *)
    |> fun(sections) -> let used = getUsedLabels(sections)
                        sections |> Array.map (fun(name,args,ls) -> (name,args,ls |> Array.filter (function ALabel(x,_) -> isGlobal(x) || used.Contains(x)
                                                                                                          | Directive(_) | Instr(_) -> true)))

    (* extract instruction blocks in text section *)
    |> fun(sections) -> match Array.partition (function (".text",[],_) -> true | _ -> false) sections
                         with | ([|(_,_,tLines)|],others) when tLines.Length > 0 && Array.forall (ILineIsDirective >> not) tLines -> (tLines,others)
                              | _ -> //eprintfn "no text section found %A" sections
                                     exit(1)

    |> fun(tLines,others) -> let (nxyn_sections,others1) = others |> Array.partition (fun(name,_,_) -> name = ".rodata.NXyn")
                             build_voidDict(nxyn_sections)
                             build_konstSizeDict(others1)
                             (tLines,others)

    (* insert additional labels following branching instructions *)
    |> fun(tLines,others) -> let lines = new ResizeArray<ILine>()
                             match tLines.[0] with Instr(_) -> lines.Add(ALabel(labelGen.Draw(),1)) | _ -> ()
                             let isBoundary(i) = i < tLines.Length - 1 && match tLines.[i]
                                                                           with | Instr(n) -> (Instr_is_Jthru(n) || Instr_is_Jfollow(n)) &&
                                                                                              (match tLines.[i+1] with Instr(_) -> true | _ -> false)
                                                                                | _ -> false 
                             tLines |> Array.iteri (fun i line -> lines.Add(line); if isBoundary(i) then lines.Add(ALabel(labelGen.Draw(),1)))
                             (lines.ToArray(),others)
                                  
    (* chunk lines in text segment *)
    |> fun(tLines,others) -> let blocks = new ResizeArray<string * int * array<Instr>>()
                             let addIBlock(l,align,instrs) = if l <> "" then blocks.Add(l,align,(instrs:ResizeArray<_>).ToArray())
                             let newIBlock(l,align) = (l,align,new ResizeArray<Instr>())
                             (newIBlock("",0), tLines) ||> Array.fold (fun s0 -> function Instr(i) -> let (_,_,instrs) = s0
                                                                                                      (instrs:ResizeArray<_>).Add(i)
                                                                                                      s0
                                                                                        | ALabel(lbl,align) -> addIBlock(s0); newIBlock(lbl,align)
                                                                                        | Directive(_) -> failwith "unexpected directive")
                                                       |> addIBlock
                             (blocks.ToArray(),others)
                                  
    (* insert instruction blocks into dictionary *)
    |> fun(tBlocks,others) -> let dynSuccs = new Dictionary<String, HashSet<String>>()
                              let dynPreds = new Dictionary<String, HashSet<String>>()

                              let tBlocks = tBlocks |> Array.mapi (fun i (name,align,instrs) -> dynSuccs.[name] <- new HashSet<String>()
                                                                                                dynPreds.[name] <- new HashSet<String>()
                                                                                                let lexpred = if i < 1 then "" else get1of3(tBlocks.[i-1])
                                                                                                let lexsucc = if i >= tBlocks.Length - 1 then "" else get1of3(tBlocks.[i+1])
                                                                                                (name,align,instrs,lexpred,lexsucc))
                              tBlocks |> Array.iter (fun (name,_,instrs,_,lexsucc) ->
                                                       if instrs <> [||] then
                                                          instrs.[instrs.Length - 1]
                                                          |> targets_of_Instr lexsucc 
                                                          (* TBD assert that _all_ targets are known *)
                                                          |> List.iter (fun x -> dynSuccs.[name].Add(x) |> ignore
                                                                                 if not(dynSuccs.ContainsKey(x)) then externABI.Add(x) |> ignore))

                              tBlocks |> Array.iter (fun (name,_,_,_,_) -> for succ in dynSuccs.[name] do if dynPreds.ContainsKey(succ) then dynPreds.[succ].Add(name) |> ignore)

                              let tDict = new Dictionary<string, IBlock>()

                              tBlocks
                              |> Array.iter (fun (name,alignment,instrs,lexpred,lexsucc) ->
                                               let retvoid =
                                                 if instrs = [||] then false
                                                 else let lastinstr = instrs.[instrs.Length - 1]
                                                      match lastinstr with
                                                          | I0(I_I0.Retq | I_I0.Ret(_)) ->
                                                               let retVoid = ref true
                                                               let visited = new HashSet<string>()
                                                               let rec dfs(name) =
                                                                   if not(visited.Contains(name)) && !retVoid then
                                                                       visited.Add(name) |> ignore
                                                                       if not(dynSuccs.ContainsKey(name)) then eprintfn "not in tDict: %s" name; retVoid := false
                                                                       elif isFunction(name) && not(voidDict.Contains(name)) then
                                                                           if !dodump then eprintfn "function %s, but not returning void" name
                                                                           retVoid := false
                                                                       else
                                                                           for succ in dynPreds.[name] do dfs(succ)
                                                               dfs(name)
                                                               !retVoid
                                                          | _ -> false
                                               tDict.[name] <- { Name        = name;
                                                                 Alignment   = alignment;
                                                                 Instrs      = instrs;
                                                                 LexPred     = lexpred;
                                                                 LexSucc     = lexsucc;
                                                                 DynPreds    = ArrayMisc.array_of_collection(dynPreds.[name]);
                                                                 DynSuccs    = ArrayMisc.array_of_collection(dynSuccs.[name]);
                                                                 ReturnsVoid = retvoid })
                              (tDict, others)
    (* *)
    
    |> fun(tDict,others) -> (* connected components *)
                            let ccs = get_ccs(tDict)

                            if !dodump then eprintfn "n ccs = %d, = %A" ccs.Length ccs
                            (* don't reduce comps any further, functions may be referenced indirectly *)

                            (* topologically sort blocks to minimize the number of data-flow analysis iterations (per comp) *)
                            let ccs = Array.map (tsort_cc tDict) ccs

                            
                            ccs |> Array.iter (optimize_cc tDict)

                            (tDict,others)

    (* convert dictionary back to lines suitable for output *)
    |> fun(tDict,others) -> let tBlocks = ArrayMisc.array_of_collection(tDict.Values)

                            (* get blocks that have no lexical predecessor/successor and assert there's exactly one of each *)
                            let firsts = tBlocks |> Array.choose (fun(b) -> if b.LexPred = "" then Some(b.Name) else None)
                            let lasts  = tBlocks |> Array.filter (fun(b) -> b.LexSucc = "")
                            if firsts.Length <> 1 || lasts.Length  <> 1 then failwith "firsts || lasts" 

                            let addAttsToLines lines l = let (found,ls) = attributes.TryGetValue(l)
                                                         if found then ls |> List.iter (fun(d,args) -> (lines:ResizeArray<ILine>).Add(Directive(d,Iexpr(Iexpr.Named(l))::args)))
//                            failwith "before output!"

                            let tLines = new ResizeArray<ILine>()

//                            let size1 = ref 0
                            let rec loop(l) = if l <> "" then let b = tDict.[l]
                                                              addAttsToLines tLines l
                                                              tLines.Add(ALabel(l,b.Alignment))

//                                                              size1 := !size1 + (b.Instrs |> InstrSize.sizes_of_Instrs |> Array.sum) // PrepareForInstrs08(b.Instrs)
//                                                              ninstrs1 := !ninstrs1 + b.Instrs.Length
                                                              b.Instrs |> Array.iter (fun(i) -> tLines.Add(Instr(i)))
                                                              b.LexSucc |> loop
                            firsts.[0] |> loop

//                            eprintfn "ninstrs %d => %d" !ninstrs0 !ninstrs1
//                            eprintfn "%s: size1=nbytes %d" !Global.InFileName !size1

//                            LogInfo(sprintf "ninstrs %d => %d" !ninstrs0 !ninstrs1)

                            others |> Array.filter (function (".data",[],[||]) -> false | _ -> true)
                                   |> Array.map (fun(name,args,lines0) -> let lines = new ResizeArray<ILine>()
                                                                          lines0 |> Array.iter (fun l -> match l with ALabel(lbl,_) -> addAttsToLines lines lbl
                                                                                                                    | Directive _ | Instr _ -> ()
                                                                                                         lines.Add(l))
                                                                          (name,args,lines.ToArray()))
                                   |> Array.append [|(".text",[],tLines.ToArray())|]

(***)
(* main gets (r:IO.TextReader) (w:IO.TextReader) *)

(* BOILER *)
match Environment.GetCommandLineArgs() |> Array.toList
 with | (_::logFile::inFile::outFile::options) -> InFileName  := inFile
                                                  OutFileName := outFile
                                                  LogFileName := logFile
                                                  RawOptions  := options
                                                  AppName     := "nxas_opt"

                                                  DumpOptions()
                                                  
                                                  use r = new System.IO.StreamReader(inFile)
                                                  let program0 = Program_of_JSONvalue(JSON.Value.Read(r))
                                                  r.Close()
                                                  let program = main logFile inFile program0
                                                  use w = new System.IO.StreamWriter(outFile)
                                                  JSONvalue_of_Program(program).Write(w)
                                                  w.Close()
      | _ -> eprintfn "usage: nxas_opt <logFile> <input.JSON> <output.JSON> [options...]"; exit 1
